/****************************************************************************
**  KOJAK                                                                  **
**  http://www.fz-juelich.de/zam/kojak/                                    **
*****************************************************************************
**  Copyright (c) 2003-2004                                                ** 
**  University of Tennessee, Innovative Computing Laboratory               **
**                                                                         ** 
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#include "dlg_prog.h"
#include "elg_error.h"

#include <string.h>
#include <stdlib.h>

#define MAX_LINE_LENGTH 8192

struct dlg_prog
{
  char*  name;
  long   nmods;
  long   nfuncs;
  long   ncsites;
  struct dlg_mod**   modv;
  struct dlg_func**  funcv;
  struct dlg_csite** csitev;
};

struct dlg_prog*  
dlg_prog_create(const char* path)
{
  FILE* file;
  char  line[MAX_LINE_LENGTH];
  char* name;
  int   i;
  struct dlg_prog* prog;

  file = fopen(path, "r");
  if ( file == NULL )
    elg_error();

  prog = malloc(sizeof(struct dlg_prog));
  if ( prog == NULL )
    elg_error();

  if ( fgets(line, MAX_LINE_LENGTH, file) == NULL )
    elg_error();
    
  if ( strcmp(strtok(line, " "), "PROGRAM") != 0 )
    elg_error_msg("invalid program-specification string");
  
  name = strtok(NULL, " ");
  prog->name = malloc(strlen(name) + 1);
  if ( prog->name == NULL )
    elg_error();
  memcpy(prog->name, name, strlen(name) + 1);
  
  prog->nmods = atol(strtok(NULL, " "));
  if ( prog->nmods > 0 )
    {
      prog->modv  = calloc(prog->nmods, sizeof(struct dlg_mod*));
      if ( prog->modv == NULL )
	elg_error();
    }
  else
    {
      prog->modv  = NULL;
    }

  prog->nfuncs = atol(strtok(NULL, " "));
  if ( prog->nfuncs > 0 )
    {
      prog->funcv  = calloc(prog->nfuncs, sizeof(struct dlg_func*));
      if ( prog->funcv == NULL )
	elg_error();
    }
  else
    {
      prog->funcv = NULL;
    }
  
  prog->ncsites = atol(strtok(NULL, " "));
  if ( prog->ncsites > 0 )
    {
      prog->csitev  = calloc(prog->ncsites, sizeof(struct dlg_csite*));
      if ( prog->csitev == NULL )
	elg_error();
    }
  else
    {
      prog->csitev = NULL;
    }
  
  for ( i = 0; i < prog->nmods; i++)
    {
      struct dlg_mod* mod;

      if ( fgets(line, MAX_LINE_LENGTH, file) == NULL )
	elg_error();
      mod = dlg_mod_create(line);
      prog->modv[dlg_mod_id(mod)] = mod;
    }

  for ( i = 0; i < prog->nfuncs; i++)
    {
      struct dlg_func* func;

      if ( fgets(line, MAX_LINE_LENGTH, file) == NULL )
	elg_error();
      func = dlg_func_create(line);
      prog->funcv[dlg_func_id(func)] = func;
    }

  for ( i = 0; i < prog->ncsites; i++)
    {
      struct dlg_csite* csite;

      if ( fgets(line, MAX_LINE_LENGTH, file) == NULL )
	elg_error();
      csite = dlg_csite_create(line);
      prog->csitev[dlg_csite_id(csite)] = csite;
    }

  close(file);

  return prog;
}

void              
dlg_prog_delete(struct dlg_prog* prog)
{
  int i;

  free(prog->name);

  for ( i = 0; i < prog->nmods; i++)
    dlg_mod_delete(prog->modv[i]);
  for ( i = 0; i < prog->nfuncs; i++)
    dlg_func_delete(prog->funcv[i]);
  for ( i = 0; i < prog->ncsites; i++)
    dlg_csite_delete(prog->csitev[i]);

  free(prog->modv);
  free(prog->funcv);
  free(prog->csitev);

  free(prog);
}

const char*      
dlg_prog_name(struct dlg_prog* prog)
{
  return prog->name;
}

long             
dlg_prog_num_mods(struct dlg_prog* prog)
{
  return prog->nmods;
}

long             
dlg_prog_num_funcs(struct dlg_prog* prog)
{
  return prog->nfuncs;
}

long             
dlg_prog_num_csites(struct dlg_prog* prog)
{
  return prog->ncsites;
}

struct dlg_mod*   
dlg_prog_mod(struct dlg_prog* prog, long i)
{
  return prog->modv[i];
}

struct dlg_func*  
dlg_prog_func(struct dlg_prog* prog, long i)
{
  return prog->funcv[i];
}

struct dlg_csite* 
dlg_prog_csite(struct dlg_prog* prog, long i)
{
  return prog->csitev[i];
}

void             
dlg_prog_fprintf(FILE* stream, struct dlg_prog* prog)
{
  int i;
  
  fprintf(stream, "PROGRAM %s %d %d %d\n", 
	  prog->name, 
	  prog->nmods,  
	  prog->nfuncs,  
	  prog->ncsites);
  
  for ( i = 0; i < prog->nmods; i++)
    dlg_mod_fprintf(stream, prog->modv[i]);
  for ( i = 0; i < prog->nfuncs; i++)
    dlg_func_fprintf(stream, prog->funcv[i]);
  for ( i = 0; i < prog->ncsites; i++)
    dlg_csite_fprintf(stream, prog->csitev[i]);
}
