/****************************************************************************
**  KOJAK                                                                  **
**  http://www.fz-juelich.de/zam/kojak/                                    **
*****************************************************************************
**  Copyright (c) 1998-2004                                                **
**  Forschungszentrum Juelich, Zentralinstitut fuer Angewandte Mathematik  **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <limits.h>
#include <stdio.h>

#include "elg_readcb.h"
#include "elg_error.h"

/*
 * Utility module
 */

static const char* mpi_colls[] = {
  "MPI_Allgather", "MPI_Allgatherv", "MPI_Allreduce",
  "MPI_Alltoall", "MPI_Alltoallv", "MPI_Alltoallw",
  "MPI_Barrier", "MPI_Bcast", "MPI_Exscan",
  "MPI_Gather", "MPI_Gatherv", "MPI_Reduce", "MPI_Reduce_scatter",
  "MPI_Scan", "MPI_Scatter", "MPI_Scatterv"
};

int mycmp(const void *v1, const void *v2) {
  return strcasecmp(*((const char**)v1), *((const char**)v2));
}

int is_mpi_collective(const char* str) {
  return (bsearch(&str, mpi_colls, sizeof(mpi_colls)/sizeof(char *),
                  sizeof(char *), mycmp) !=0);
}

void *x_malloc(size_t size) {
  void *r = malloc(size);
  if ( r == NULL ) elg_error_msg("Out of memory");
  return r;
}

void *x_realloc(void *ptr, size_t size) {
  void *r = realloc(ptr, size);
  if ( r == NULL ) elg_error_msg("Out of memory");
  return r;
}

static int printdecls = 0;


/*
 * Region Stack Module
 */

#define STACK_SIZ 2048

typedef struct {
  elg_ui4 reg;
  elg_d8  tim;
} StackNode;

typedef struct {
  StackNode* vals;
  StackNode** idxs;
} RegStack;

static RegStack mystack;

void stk_init(int totallocs) {
  int i;
  StackNode *n;

  mystack.vals = x_malloc(totallocs * STACK_SIZ * sizeof(StackNode));
  mystack.idxs = x_malloc(totallocs * sizeof(StackNode *));
  for (i=0; i<totallocs; i++) {
    n = mystack.idxs[i] = mystack.vals + (i*STACK_SIZ);
    n->reg = -1;
    n->tim = -1.0;
  }
}

void stk_push(elg_ui4 l, elg_ui4 r, elg_d8 t) {
  StackNode *n = ++mystack.idxs[l];
  if ( n >= mystack.vals + ((l+1)*STACK_SIZ) )
    elg_error_msg("Region stack overflow l=%d r=%d t=%g", l, r, t);
  n->reg = r;
  n->tim = t;
}

StackNode* stk_pop(elg_ui4 l) {
  if ( mystack.idxs[l] <= mystack.vals + (l*STACK_SIZ) )
    elg_error_msg("Region stack underflow");
  return mystack.idxs[l]--;
}

/*
 * String Table Module
 */

static int numstring = 0;
static char** stringtab = 0;

void str_add(int strid, char* str) {
  if ( (strid+1) > numstring ) {
    numstring += 100;
    stringtab = x_realloc(stringtab, numstring * sizeof(char *));
  }
  stringtab[strid] = str;
}

/*
 * File Table Module
 */

static int numfile = 0;
static char** filetab = 0;

void file_add(int fid, const char*str) {
  int i;

  if ( (fid+1) > numfile ) {
    numfile += 20;
    filetab = x_realloc(filetab, numfile * sizeof(char *));
    for (i=(numfile-20); i<numfile; i++) filetab[i] = 0;
  }
  filetab[fid] = strdup(str);
}

void file_write() {
  int i;

  for (i=0; i<numfile; i++) {
    if ( filetab[i] ) printf("file %d \"%s\"\n", i, filetab[i]);
  }
}

/*
 * SCL Table Module
 */

typedef struct {
  int fid;
  int line;
} SCL;

static int numscl = 0;
static int sclid = 0;
static SCL* scltab = 0;

int scl_add(int fid, int line) {
  if ( fid == ELG_NO_ID || line == ELG_NO_LNO ) return -1;

  if ( (sclid+1) > numscl ) {
    numscl += 20;
    scltab = x_realloc(scltab, numscl * sizeof(SCL));
  }
  scltab[sclid].fid  = fid;
  scltab[sclid].line = line;
  return sclid++;
}

void scl_write() {
  int i;

  for (i=0; i<sclid; i++) {
    printf("scl %d file %d line %d\n", i, scltab[i].fid, scltab[i].line);
  }
}

/*
 * Activity Table Module
 */

typedef struct {
  elg_ui4 id;
  const char* name;
} Activity;

static int numact = 0;
static Activity* acttab = 0;
static elg_ui4 maxact = 0;

static int act_elg = -1;
static int act_omp = -1;

void act_add(elg_ui4 id, const char* s) {
  int i;

  /*search table*/
  for (i=0; i<numact; i++) {
    if ( acttab[i].id == id ) {
      /*already added*/
      return;
    } else if ( acttab[i].id == ELG_NO_ID ) {
      /*empty slot; add*/
      acttab[i].id = id;
      acttab[i].name = s;
      if ( id > maxact ) maxact = id;
      if ( strcmp(s, "OMP") == 0 ) act_omp = id;
      return;
    }
  }

  /*table full; resize and add*/
  numact += 10;
  acttab = x_realloc(acttab, numact * sizeof(Activity));
  for (i=numact-10; i<numact; i++) acttab[i].id = ELG_NO_ID;
  acttab[numact-10].id = id;
  acttab[numact-10].name = s;
  if ( id > maxact ) maxact = id;
  if ( strcmp(s, "OMP") == 0 ) act_omp = id;
}

void act_write() {
  int i;

  act_elg = maxact + 1;
  act_add(act_elg, "ELG");

  for (i=0; i<numact; i++) {
    if ( acttab[i].id == ELG_NO_ID ) return;
    printf("group %d \"%s\"\n", acttab[i].id, acttab[i].name);
  }
}

/*
 * Callsite Table Module
 */

typedef struct {
  elg_ui4 csid;
  elg_ui4 erid;
  int sclid;
} Csite;

static int numcsite = 0;
static Csite* csitetab = 0;

void csite_add(elg_ui4 csid, elg_ui4 erid, int sclid) {
  int i;

  if ( (csid + 1) > numcsite ) {
    /*table full; resize*/
    numcsite += 20;
    csitetab = x_realloc(csitetab, numcsite * sizeof(Csite));
    for (i=(numcsite-20); i<numcsite; i++) csitetab[i].erid = ELG_NO_ID;
  }
  csitetab[csid].erid  = erid;
  csitetab[csid].sclid = sclid;
}

void csite_write() {
  int i;

  for (i=0; i<numcsite; i++) {
    if ( csitetab[i].erid != ELG_NO_ID )
      printf("callsite %d region %d scl %d\n", i,
             csitetab[i].erid, csitetab[i].sclid);
  }
}

/*
 * State Table Module
 */

typedef struct {
  elg_ui4 act;
  elg_ui4 id;
  const char* name;
  int sclid;
  elg_ui1 type;
} State;

static int numstate = 0;
static int capstate = 0;
static State* statetab = 0;
static elg_ui4 maxstate = 0;

static int state_flush = -1;
static int state_off = -1;
static int state_pregion = -1;

void state_add(elg_ui4 act, elg_ui4 id, const char* s, elg_ui1 t, int sclid) {
  if (numstate == capstate) {
    /*table full; resize*/
    capstate += 100;
    statetab = x_realloc(statetab, capstate * sizeof(State));
  }
  statetab[numstate].act   = act;
  statetab[numstate].id    = id;
  statetab[numstate].name  = s;
  statetab[numstate].sclid = sclid;
  statetab[numstate].type  = t;
  if ( id > maxstate ) maxstate = id;
  numstate++;
}

static char*rtypenames[] = {
  "UNKNOWN", "FUNCTION", "LOOP", "USER_REGION",
  "?", "?", "?", "?", "?", "?", "?",
  "OMP_PARALLEL", "OMP_LOOP", "OMP_SECTIONS", "OMP_SECTION",
  "OMP_WORKSHARE", "OMP_SINGLE", "OMP_MASTER", "OMP_CRITICAL",
  "OMP_ATOMIC", "OMP_BARRIER", "OMP_IBARRIER", "OMP_FLUSH",
  "OMP_CRITICAL_SBLOCK", "OMP_SINGLE_SBLOCK"
};

void state_write() {
  int i;

  state_flush = maxstate + 1;
  state_add(act_elg, state_flush, "flush", ELG_UNKNOWN, -1);
  state_off = state_flush + 1;
  state_add(act_elg, state_off, "off", ELG_UNKNOWN, -1);
  if ( act_omp != -1 ) {
    state_pregion = state_flush + 2;
    state_add(act_omp, state_pregion, "parallel region", ELG_OMP_PARALLEL, -1);
  }

  for (i=0; i<numstate; i++) {
    printf("region %d \"%s\" group %d", 
           statetab[i].id, statetab[i].name, statetab[i].act);
    if ( statetab[i].sclid != -1 )
      printf(" scl %d", statetab[i].sclid);
    printf(" %s\n", rtypenames[statetab[i].type]);
  }
}

/*
 * GlobalOp Table Module
 */

typedef struct {
  elg_ui4 id;
  const char* name;
} Glop;

static int numglop = 0;
static int capglop = 0;
static Glop* gloptab = 0;

void glop_add(elg_ui4 id, const char* s) {
  if (numglop == capglop) {
    /*table full; resize*/
    capglop += 10;
    gloptab = x_realloc(gloptab, capglop * sizeof(Glop));
  }
  gloptab[numglop].id   = id;
  gloptab[numglop].name = s;
  numglop++;
}

int glopcmp(const void *v1, const void *v2) {
  return ((Glop*)v1)->id - ((Glop*)v2)->id;
}

int is_glop(elg_ui4 id) {
  Glop key;
  key.id = id;
  if (numglob)
    return (bsearch(&key, gloptab, numglop, sizeof(Glop), glopcmp) != 0);
  else
    return 0;
}

void glop_write() {
  int i;

  for (i=0; i<numglop; i++) {
    printf("globalop %d \"%s\"\n", gloptab[i].id, gloptab[i].name);
  }
}

/*
 * Groups Module
 */

typedef struct grp {
  char* name;
  int id;
  int num_member;
  struct grp* member;
} Group;

static Group machine[1];
const char *group_descr[] = { "machine", "node", "process", "thread" };

static unsigned int grpid = 100000;

unsigned int group_print(Group *g) {
  int i;

  if ( g->member ) {
    unsigned int* array = x_malloc(g->num_member * sizeof(unsigned int));
    for (i=0; i<g->num_member; ++i) {
      array[i] = group_print(g->member + i);
    }
    grpid++;
    printf("locationgroup G%d \"%s\" members", (grpid-100000), g->name);
    for (i=0; i<g->num_member; ++i) {
      if ( array[i] > 100000 )
        printf(" G%d", (array[i]-100000));
      else
        printf(" %d", array[i]);
    }
    printf("\n");
    
    return grpid;
  } else {
    return g->id;
  }
}

static int totallocs = 0;
static int *locmap = 0;
static int *locremap = 0;

typedef struct {
  Group* mach;
  Group* node;
  Group* proc;
  Group* thrd;
  char lstr[16];
} Location;

static int numlocs = 0;
static Location* loctab = 0;

void machine_write() {
  int n, p, numnode, numproc;
  Group *mach, *node;

  /*fix real number of nodes and procs*/
  mach = machine;
  numnode = 0;
  for (n=0; n<mach->num_member; n++) {
    node = mach->member + n;
    if ( node->id == -1 ) break;
    numnode++;
    numproc = 0;
    for (p=0; p<node->num_member; p++) {
      if ( node->member[p].id == -1 ) break;
      numproc++;
    }
    node->num_member = numproc;
  }
  mach->num_member = numnode;

  /*dump machine*/
  if (printdecls) {
    group_print(machine);
  }
}

#define MAX_THREADS 256

void loc_add(elg_ui4 lid, elg_ui4 mid, elg_ui4 nid, elg_ui4 pid, elg_ui4 tid) {
  int i;
  Group *node = &(machine[mid].member[nid]);
  Group *proc = 0;
  Group *thrd = 0;
  Group *old = 0;
  elg_ui4 cpuc = node->num_member;

  /*more processes than CPUs found -> get more space for proc's*/
  if ( pid >= cpuc ) {
    old = node->member;
    node->member = x_realloc(node->member, (pid+1) * sizeof(Group));
    node->num_member = pid+1;
    for (i=cpuc; i<(pid+1); ++i) {
      node->member[i].id = -1;
    }

    /*if realloc moved block -> fix old pointers into block*/
    if ( old != node->member ) {
      for (i=0; i<lid; ++i) {
        if ( loctab[i].node == node ) {
          /*we changed this node -> fix proc pointer here*/
          int p = loctab[i].proc - old;
          loctab[i].proc = node->member + p;
        }
      }
    }
  }

  for (i=0; i<node->num_member; i++) {
    proc = node->member + i;
    if ( proc->id == -1 ) {
      /*not found; add*/
      proc->id = pid;
      proc->name = x_malloc(16 * sizeof(char));
      sprintf(proc->name, "process %d", pid);
      proc->num_member = tid+1;
      if ( act_omp == -1 ) {
        /* plain MPI; no thread*/
        proc->member = 0;
        thrd = 0;
      } else {
        /*OMP: threads*/
        proc->member = x_malloc(MAX_THREADS * sizeof(Group));
        for (i=0; i<MAX_THREADS; ++i) {
          proc->member[i].id = -1;
        }
        thrd = proc->member + tid;
      }
      break;
    } else if ( proc->id == pid ) {
      /*found; update num threads*/
      if ( (tid+1) > proc->num_member )
        proc->num_member = tid+1;
      thrd = proc->member + tid;
      break;
    }
  }
  if ( thrd ) {
    thrd->id = lid;
    thrd->name = 0;
    thrd->member = 0;
    thrd->num_member = 0;
  }

  /*update loctab*/
  if ( (lid+1) > numlocs ) {
    numlocs += 32;
    loctab = x_realloc(loctab, numlocs * sizeof(Location));
  }
  loctab[lid].mach = machine + mid;
  loctab[lid].node = node;
  loctab[lid].proc = proc;
  loctab[lid].thrd = thrd;
  if ( lid+1 > totallocs ) totallocs = lid+1;
}

void cpuname_write() {
  int i, t;
  Group *proc;

  locmap = x_malloc(totallocs * sizeof(int));
  locremap = x_malloc(totallocs * sizeof(int));
  if ( act_omp == -1 ) {
    /*plain MPI: no reordering, default names OK*/
    for (i=0; i<totallocs; i++) {
      locmap[i] = i;
      locremap[i] = i;
      sprintf(loctab[i].lstr, "%d", loctab[i].proc->id);
    }
  } else {
    /*reorder threads correctly*/
    t = 0;
    while ( t < totallocs ) {
      proc = loctab[t].proc;
      for (i=0; i<proc->num_member; i++) {
        locmap[proc->member[i].id] = t++;
      }
    }

    /*create better naming*/
    for (i=0; i<totallocs; i++) {
      loctab[i].thrd->name = x_malloc(24 * sizeof(char));
      if ( loctab[i].proc->member[0].id == i ) {
        /*master thread*/
        sprintf(loctab[i].thrd->name, "process %d", locmap[i]);
        sprintf(loctab[i].lstr, "%d:%d:0", loctab[i].node->id,
                                           loctab[i].proc->id);
      } else {
        /*slave thread*/
        int tid = (loctab[i].thrd - loctab[i].proc->member);
        sprintf(loctab[i].thrd->name, "thread %d:%d",
                locmap[loctab[i].proc->member[0].id], tid);
        sprintf(loctab[i].lstr, "%d:%d:%d", loctab[i].node->id,
                                            loctab[i].proc->id, tid);
      }
    }
  
    /*write new name in sorted order*/
    for (i=0; i<totallocs; i++) {
      for (t=0; t<totallocs; t++) {
        if ( i == locmap[t] ) {
          if (printdecls) {
            printf("location %d %d:%d:%d:%d \"%s\"\n", i,
                   loctab[t].mach->id, loctab[t].node->id,
                   loctab[t].proc->id, loctab[t].thrd->id,
                   loctab[t].thrd->name);
	  }
          locremap[i] = t;
          break;
        }
      }
    }
  }
}

/*
 * Communicator Table Module
 */

typedef struct {
  int numbyte;
  elg_ui1* bits;
} Comm;

static int numcomm = 0;
static int maxcomm = 0;
static Comm* commtab = 0;

void comm_add(elg_ui4 cid, elg_ui4 grpc, elg_ui1 *v) {
  int i;

  if ( (cid+1) > numcomm ) {
    numcomm += 20;
    commtab = x_realloc(commtab, numcomm * sizeof(Comm));
    for (i=(numcomm-20); i<numcomm; i++) commtab[i].numbyte = 0;
  }
  commtab[cid].numbyte = grpc;
  commtab[cid].bits    = v;

  if ( cid > maxcomm ) maxcomm = cid;
}

unsigned int get_comm_loc(int l) {
  if ( act_omp == -1 ) {
    /*plain MPI: location==comm member id*/
    return l;
  } else {
    int i;
    for (i=0; i<totallocs; i++) {
      if ( loctab[i].proc->id == l && loctab[i].proc->member[0].id == i )
        return locmap[i];
    }
  }
  return -1;
}

void comm_write() {
  int i, j;
  int n, p, t;

  /*MPI Communicators*/
  for (i=0; i<numcomm; i++) {
    if ( commtab[i].numbyte ) {
      char *name = x_malloc(16 * sizeof(char));
      unsigned int *array;
      int num = 0;

      for (j=0; j<(commtab[i].numbyte*8); j++) {
        if ( (1 << (j % 8)) & commtab[i].bits[j/8] ) num++;
      }
      array = x_malloc(num * sizeof(unsigned int));
      num = 0;
      for (j=0; j<(commtab[i].numbyte*8); j++) {
        if ( (1 << (j % 8)) & commtab[i].bits[j/8] )
          array[num++] = get_comm_loc(j);
      }

      sprintf(name, "comm %d", i);
      if (printdecls) {
        printf("communicator %d members", i);
        for (j=0; j<num; j++) printf(" %d", array[j]);
        printf("\n");
      }
    }
  }

  /*OpenMP Thread Groups*/
  if ( act_omp != -1 ) {
    for (n=0; n<machine[0].num_member; n++) {
      Group *node = machine[0].member + n;
      for (p=0; p<node->num_member; p++) {
        Group *proc = node->member + p;
        char *name = x_malloc(16 * sizeof(char));
        unsigned int *array = x_malloc(proc->num_member * sizeof(int));
        for (t=0; t<proc->num_member; t++) {
          array[t] = proc->member[t].id;
        }
        sprintf(name, "tgroup %d", proc->id);
        if (printdecls) {
          printf("threadgroup %d members", proc->id);
          for (j=0; j<proc->num_member; j++) printf(" %d", array[j]);
          printf("\n");
	}
      }
    }
  }
}

/*
 * Metrics
 */

typedef struct {
  const char *name;
  const char *descr;
  elg_ui1 type;
  elg_ui1 mode;
  elg_ui1 iv;
} Metric;

static int nummet = 0;
static Metric* mettab = 0;

void metric_add(elg_ui4 metid, const char *metname, const char *metdescr,
                elg_ui1 metdtype, elg_ui1 metmode, elg_ui1 metiv) {
  int i;

  if ( (metid+1) > nummet ) {
    nummet += 5;
    mettab = x_realloc(mettab, nummet * sizeof(Metric));
    for (i=(nummet-5); i<nummet; i++) mettab[i].name = 0;
  }
  mettab[metid].name  = metname;
  mettab[metid].descr = metdescr;
  mettab[metid].type  = metdtype;
  mettab[metid].mode  = metmode;
  mettab[metid].iv    = metiv;
}

static char *type2str[] = { "INT", "FLOAT" };
static char *mode2str[] = { "COUNTER", "RATE", "SAMPLE" };
static char *iv2str[]   = { "START", "LAST", "NEXT" };

void metric_write() {
  int i;

  for (i=0; i<nummet; i++) {
    if ( mettab[i].name ) {
      printf("metric %d %s", i, mettab[i].name);
      if ( mettab[i].descr )
        printf(" [%s]",  mettab[i].descr);
      else
        printf(" -");
      printf(" %s %s %s\n", type2str[mettab[i].type],
                            mode2str[mettab[i].mode], iv2str[mettab[i].iv]);
    }
  }
}

/*
 * EPILOG read callbacks
 */

typedef struct {
  off_t done;
  int   numcont;
  char* strpos;
} MYDATA;

void elg_readcb_ALL(elg_ui1 t, elg_ui1 length, void* userdata) {
  MYDATA* md = (MYDATA*) userdata;
  md->done += (length + 2);
}

void elg_readcb_STRING(elg_ui4 strid, elg_ui1 cntc, const char* str,
                       void* userdata) {
  MYDATA* md = (MYDATA*) userdata;
  if ( cntc ) {
    md->numcont = cntc;
    md->strpos  = malloc((250 + md->numcont*255) * sizeof(char));
    str_add(strid, md->strpos);
    strncpy(md->strpos, str, 250);
    md->strpos += 250;
  } else {
    str_add(strid, strdup(str));
  }
}

void elg_readcb_STRING_CNT(const char* str, void* userdata) {
  MYDATA* md = (MYDATA*) userdata;
  md->numcont--;
  if ( md->numcont ) {
    strncpy(md->strpos, str, 255);
    md->strpos += 255;
  } else {
    strcpy(md->strpos, str);
  }
}

void elg_readcb_MACHINE(elg_ui4 mid, elg_ui4 nodec, elg_ui4 mnid,
                        void* userdata) {
  int i;

  if ( mnid == ELG_NO_ID ) {
    machine[mid].name = x_malloc(16 * sizeof(char));
    sprintf(machine[mid].name, "machine %d", mid);
  } else {
    machine[mid].name = stringtab[mnid];
  }
  machine[mid].id   = mid;
  machine[mid].num_member = nodec;
  machine[mid].member     = x_malloc(nodec * sizeof(Group));
  for (i=0; i<nodec; ++i) {
    machine[mid].member[i].id = -1;
  }
}

void elg_readcb_NODE(elg_ui4 nid, elg_ui4 mid, elg_ui4 cpuc, elg_ui4 nnid, 
                     elg_d8 cr, void* userdata) {
  int i;

  elg_ui4 nodec = machine[mid].num_member;
  Group *node;

  if ( nid >= nodec ) {
    machine[mid].member = x_realloc(machine[mid].member,
                                    (nid+1) * sizeof(Group));
    machine[mid].num_member = nid+1;
    for (i=nodec; i<(nid+1); ++i) {
      machine[mid].member[i].id = -1;
    }
  }
  node = &(machine[mid].member[nid]);
  if ( nnid == ELG_NO_ID ) {
    node->name = x_malloc(16 * sizeof(char));
    sprintf(node->name, "node %d", mid);
  } else {
    node->name = stringtab[nnid];
  }
  node->id   = nid;
  node->num_member = cpuc;
  node->member     = x_malloc(cpuc * sizeof(Group));
  for (i=0; i<cpuc; ++i) {
    node->member[i].id = -1;
  }
}

void elg_readcb_PROCESS(elg_ui4 pid, elg_ui4 pnid, void* userdata) {
  /* NOT YET USED */
}

void elg_readcb_THREAD(elg_ui4 tid, elg_ui4 pid, elg_ui4 tnid, void* userdata) {
  /* NOT YET USED */
}

void elg_readcb_LOCATION(elg_ui4 lid, elg_ui4 mid, elg_ui4 nid, elg_ui4 pid,
                         elg_ui4 tid, void* userdata) {
  loc_add(lid, mid, nid, pid, tid);
}

void elg_readcb_FILE(elg_ui4 fid, elg_ui4 fnid, void* userdata) {
  file_add(fid, stringtab[fnid]);
}

void elg_readcb_REGION(elg_ui4 rid, elg_ui4 rnid, elg_ui4 fid,
                       elg_ui4 begln, elg_ui4 endln, elg_ui4 rdid,
                       elg_ui1 rtype, void* userdata) {
  act_add(rdid, stringtab[rdid]);
  state_add(rdid, rid, stringtab[rnid], rtype, scl_add(fid, begln));
  if ( (rtype == ELG_FUNCTION && is_mpi_collective(stringtab[rnid])) ||
       (rtype == ELG_OMP_BARRIER) ||
       (rtype == ELG_OMP_IBARRIER) ) {
    glop_add(rid, stringtab[rnid]);
  }
}

void elg_readcb_CALL_SITE(elg_ui4 csid, elg_ui4 fid, elg_ui4 lno,
                          elg_ui4 erid, elg_ui4 lrid, void* userdata) {
  csite_add(csid, erid, scl_add(fid, lno));
}

void elg_readcb_METRIC(elg_ui4 metid, elg_ui4 metnid, elg_ui4 metdid,
		      elg_ui1 metdtype, elg_ui1 metmode, elg_ui1 metiv,
                      void* userdata) {
  metric_add(metid, stringtab[metnid],
             (metdid==ELG_NO_ID ? 0 : stringtab[metdid]),
             metdtype, metmode, metiv);
}

void elg_readcb_MPI_COMM(elg_ui4 cid, elg_ui4 grpc, elg_ui1 grpv[],
                         void* userdata) {
  int i;

  elg_ui1* comm = x_malloc(grpc * sizeof(elg_ui1));
  for (i=0; i<grpc; i++) comm[i] = grpv[i];
  comm_add(cid, grpc, comm);
}

void elg_readcb_CART_TOPOLOGY(elg_ui4 topid, elg_ui1 ndims, elg_ui4 dimv[], 
			      elg_ui1 periodv, void* userdata)
{
  /* NOT YET USED */
}

void elg_readcb_CART_CORDS(elg_ui4 topid, elg_ui4 lid, elg_ui1 ndims, 
			   elg_ui4 coordv[], void* userdata )
{
  /* NOT YET USED */
}

void elg_readcb_OFFSET(elg_d8 ltime, elg_d8 offset, void* userdata) {
  /* NOT USED IN GLOBAL EPILOG TRACES */
}

void elg_readcb_NUM_EVENTS(elg_ui4 eventc, void* userdata) {
  /* NOT YET USED */
}

void elg_readcb_LAST_DEF(void* userdata) {
  cpuname_write();
  machine_write();
  comm_write();
  if (printdecls) {
    file_write();
    scl_write();
    act_write();
    state_write();
    csite_write();
    metric_write();
    glop_write();
  }
  stk_init(totallocs);
}

void elg_readcb_ENTER(elg_ui4 lid, elg_d8 time, elg_ui4 rid,
                      elg_ui1 metc, elg_ui8 metv[], void* userdata) {
  lid = locmap[lid];
  stk_push(lid, rid, time);
}

void elg_readcb_ENTER_CS(elg_ui4 lid, elg_d8 time, elg_ui4 csid,
                      elg_ui1 metc, elg_ui8 metv[], void* userdata) {
  lid = locmap[lid];
  stk_push(lid, csitetab[csid].erid, time);
}

void elg_readcb_EXIT(elg_ui4 lid, elg_d8 time, elg_ui1 metc, elg_ui8 metv[],
                     void* userdata) {
  StackNode *n;

  lid = locmap[lid];
  n = stk_pop(lid);
}

void elg_readcb_MPI_SEND(elg_ui4 lid, elg_d8 time, elg_ui4 did,
                         elg_ui4 cid, elg_ui4 tag, elg_ui4 sent,
                         void* userdata) {
  lid = locmap[lid];
  did = locmap[did];
}

void elg_readcb_MPI_RECV(elg_ui4 lid, elg_d8 time,
                         elg_ui4 sid, elg_ui4 cid, elg_ui4 tag,
                         void* userdata) {
  lid = locmap[lid];
  sid = locmap[sid];
}

void elg_readcb_MPI_COLLEXIT(elg_ui4 lid, elg_d8 time,
                             elg_ui1 metc, elg_ui8 metv[], elg_ui4 rlid,
                             elg_ui4 cid, elg_ui4 sent, elg_ui4 recvd,
                             void* userdata) {
  StackNode *n;

  lid = locmap[lid];
  n = stk_pop(lid);
  if ( rlid != ELG_NO_ID ) rlid = locmap[rlid];
}

void elg_readcb_OMP_FORK(elg_ui4 lid, elg_d8 time, void* userdata) {
  lid = locmap[lid];
}

void elg_readcb_OMP_JOIN(elg_ui4 lid, elg_d8 time, void* userdata) {
  lid = locmap[lid];
}

void elg_readcb_OMP_ALOCK(elg_ui4 lid, elg_d8 time, elg_ui4 lkid,
                          void* userdata) {
  lid = locmap[lid];
}

void elg_readcb_OMP_RLOCK(elg_ui4 lid, elg_d8 time, elg_ui4 lkid,
                          void* userdata) {
  lid = locmap[lid];
}

void elg_readcb_OMP_COLLEXIT(elg_ui4 lid, elg_d8 time,
                             elg_ui1 metc, elg_ui8 metv[], void* userdata) {
  StackNode *n;

  lid = locmap[lid];
  n = stk_pop(lid);
}

void elg_readcb_ENTER_DUMP(elg_ui4 lid, elg_d8 time,
                           elg_ui1 metc, elg_ui8 metv[], void* userdata) {
  lid = locmap[lid];
}

void elg_readcb_EXIT_DUMP(elg_ui4 lid, elg_d8 time,
                          elg_ui1 metc, elg_ui8 metv[], void* userdata) {
  lid = locmap[lid];
}

void elg_readcb_LOG_OFF(elg_ui4 lid, elg_d8 time,
                        elg_ui1 metc, elg_ui8 metv[], void* userdata) {
  lid = locmap[lid];
}

void elg_readcb_LOG_ON(elg_ui4 lid, elg_d8 time, elg_ui1 metc, elg_ui8 metv[],
                       void* userdata) {
  lid = locmap[lid];
}

/*
 * M A I N
 */

int main(int argc, char **argv) {
  char*   infile;
  int     ch;
  int     eflag = 0;
  struct stat statbuf;
  elg_ui4 numrec = 0;
  ElgRCB* handle = NULL;
  MYDATA  md = { 0, 0, 0 };

  while ((ch = getopt(argc, argv, "dh")) != EOF) {
    switch (ch) {
    case 'd': printdecls = 1;
              break;
    case 'h':
    case '?': eflag = 1;
              break;
    }
  }

  if ((argc-optind) == 1) {
    infile  = argv[optind];
  } else {
    eflag = 1;
  }

  if ( eflag ) {
    fprintf(stderr, "Usage:   %s [-d] <infile>.elg\n", argv[0]);
    fprintf(stderr, "Options: -d print declaration event info\n");
    fflush(stderr);
    exit(EXIT_FAILURE);
  }

  handle = elg_read_open(infile);
  if ( ! handle ) exit(EXIT_FAILURE);
  stat(infile, &statbuf);

  while ( elg_read_next(handle, (void*) &md) ) {
    numrec++;
    if ( numrec % 1000 == 0 ) {
      printf("done: %3.0f%%\r", ((double) md.done / statbuf.st_size) * 100.0);
      fflush(stdout);
    }
  }

  elg_read_close(handle);

  return 0;
}
