/****************************************************************************
**  KOJAK                                                                  **
**  http://www.fz-juelich.de/zam/kojak/                                    **
*****************************************************************************
**  Copyright (c) 2003-2004                                                **
**  University of Tennessee                                                **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

/****************************************************************************
*                                                                           *
* elg_dpcl                                                                  *
*                                                                           *
* Automatic EPILOG instrumenter based on DPCL                               *
*                                                                           *
****************************************************************************/

#include <cstdlib>    // for atoi() call
#include <dpcl.h>
#include <fstream>
   using std::ofstream;
#include <iostream>
   using std::cerr;
   using std::cout;
#include <libgen.h>   // for basename() call


#include "Executable.h"
#include "elg_error.h"
#include "elg_comp_dpcl.h"

/****************************************************************************
* global variables
****************************************************************************/

#define ELG_PROBE_MODULE "elg_comp_dpcl.c"

PoeAppl pappl;
Process proc;

ProbeExp init_pfun;
ProbeExp start_pfun;
ProbeExp enter_pfun;
ProbeExp exit_pfun;
ProbeExp before_pfun;
ProbeExp after_pfun;

string mainfun = "main";

Executable* exec = NULL;

// poe application ?
bool ispoe  = true;

// default poe name
char* poe_name = "/usr/bin/poe";  

// default host name
char* host_name = "localhost";

// default program name
char* prog_name = "a.out";

int nprocs = 0; 

// argument counter
int a = 1; 

// status
AisStatus sts;

/****************************************************************************
* termination of processes
****************************************************************************/

void terminate_application()
{
  if ( ispoe )
    pappl.bdestroy();
  else
    proc.bdestroy();
}

/****************************************************************************
* call-back functions
****************************************************************************/

void
data_cb(GCBSysType sys, GCBTagType tag, GCBObjType obj, GCBMsgType msg)
{
  cout << (char*) msg; 
}

void
stdout_cb(GCBSysType sys, GCBTagType tag, GCBObjType obj, GCBMsgType msg)
{
  cout << (char*) msg; 
}

void
stderr_cb(GCBSysType sys, GCBTagType tag, GCBObjType obj, GCBMsgType msg)
{
  cerr << (char*) msg; 
}

void
exit_cb(GCBSysType sys, GCBTagType tag, GCBObjType obj, GCBMsgType msg)
{
  if ( ispoe )
    {
      nprocs--;
      if ( nprocs == 0 )
	Ais_end_main_loop();
    }
  else
    Ais_end_main_loop();    
}

int sigint_handler(int s)
{
  cerr << "\nELG_DPCL terminated.\n";
  terminate_application();
}

/****************************************************************************
* parse command line 
****************************************************************************/

void parse_cmdl(int argc, char *argv[])
{
  bool error = false;

  while ( a < argc && argv[a][0] == '-' ) 
    {
      // is mpi application?
      if ( strcmp(argv[a], "-poe") == 0 )
	{
	  ispoe = true;
	}
      // is serial application?
      else if ( strcmp(argv[a], "-nopoe") == 0 )
	{
	  ispoe = false;
	}
      // numer of processes
      else if ( strcmp(argv[a], "-procs") == 0 ) 
	{
	  ispoe = true;

	  if ( (a+1) < argc )
	    {
	      nprocs = atoi(argv[++a]);
	    }
	  else
	    {
	      elg_warning("missing value for option -procs");
	      error = true;
	    }
	  
	  if ( nprocs < 1 )
	    {
	      elg_warning("invalid value for option -procs");
	      error = true;
	    }
	}
      // name of main function
      else if ( strcmp(argv[a], "-main") == 0 ) 
	{
	  if ( (a+1) < argc ) 
	    {
	      mainfun = argv[++a];
	    }
	  else
	    {
	      elg_warning("missing value for option -main");
	      error = true;
	    }
	}
      // host name
      else if ( strcmp(argv[a], "-host") == 0 ) 
	{
	  if ( (a+1) < argc ) 
	    {
	      host_name = argv[++a];
	    }
	  else
	    {
	      elg_warning("missing value for option -host");
	      error = true;
	    }
	}
      else if ( strcmp(argv[a], "-help") == 0 ) 
	{
	  error = true;
	  break;
	}
      else 
	{
	  elg_warning("unknown option");
	  error = true;
	}
      ++a;
    }
 
  if ( a >= argc )
    {
      elg_warning("missing executable");
      error = true;
    }
  else
    {
      prog_name = argv[a++];
    }

  if ( error )
    {
      cerr << "Usage: " << argv[0] << " [-poe|-nopoe] [-procs <n>]\n"
	   << "       [-host <name>] [-main <name>] <executable>\n";
      cerr << "   or: " << argv[0] << " [-help]\n";
      exit(1);
    }
}

/****************************************************************************
* inititalize DPCL
****************************************************************************/

void init_dpcl()
{
  cerr << "Intitializing DPCL...           ";

  Ais_initialize();
  
  Ais_add_signal(SIGINT, sigint_handler);
  
  sts = Ais_override_default_callback(AIS_PROC_TERMINATE_MSG, &exit_cb, NULL, NULL, NULL);
  if (sts.status() != ASC_success) 
    elg_error_msg("call-back override failed: %s", sts.status_name());

  cerr << "done.\n";
}

void create_application(int argc, char *argv[])
{
  cerr << "Creating application...         ";
  
  // arguments
  vector<char*> args(argc - a  + 3);
  args[0] = poe_name;
  args[1] = prog_name;
  for ( int i = 0; i < argc - a; i++ ) 
    args[i + 2] = argv[i + a];
  args[argc - a  + 2] = NULL;
  
  if ( ispoe )
    {
      // set environment variable MP_PROCS if -procs has been
      // specified

      if ( nprocs != 0 )
	{
	  char tmp[1024];
	  sprintf(tmp, "MP_PROCS=%d", nprocs);
	  putenv(tmp);
	}    

      sts = pappl.bcreate(host_name, 
			  poe_name,
			  &args[0],
			  environ, 
			  stdout_cb,
			  (GCBTagType) 0,
			  stderr_cb,
			  (GCBTagType) 0		    
			  );

      // if number of processes wasn't specified using -procs option,
      // nprocs is assigned the correct value here
      nprocs = pappl.get_count();

      if (sts.status() != ASC_success)
	{
	  terminate_application();
	  elg_error_msg("PoeAppl::bcreate failed: %s", sts.status_name());
	}

      proc = pappl.get_process(0);
    }
  else
    {
      sts = proc.bcreate(host_name, 
			 prog_name,
			 &argv[a], 
			 environ, 
			 stdout_cb,
			 (GCBTagType) 0,
			 stderr_cb,
			 (GCBTagType) 0		    
			 );
      if (sts.status() != ASC_success)
	{
	  terminate_application();
	  elg_error_msg("Process::bcreate failed: %s", sts.status_name());
	}
    }
  
  cerr << "done.\n";
}

/****************************************************************************
* locate probe functions
****************************************************************************/

void locate_probe_functions()
{
  cerr << "Locating probe functions...     ";

  int npfuns = 0; 

  SourceObj prog = proc.get_program_object();
  for (int c = 0 ; c < prog.child_count(); c++)
    {
      SourceObj modobj = prog.child(c);
      
      vector<char> mod_name(modobj.module_name_length());
      modobj.module_name(&mod_name[0], mod_name.size());
      	  
      if ( strcmp(&mod_name[0],  ELG_PROBE_MODULE ) == 0 )
	{
	  for ( int d = 0; d < modobj.child_count(); d++ ) 
	    {
		SourceObj funcobj = modobj.child(d);
		
		vector<char> func_name(funcobj.get_demangled_name_length());
		funcobj.get_demangled_name(&func_name[0], func_name.size());
		
		if ( strcmp(&func_name[0], "dlg_init" ) == 0 )
		    init_pfun = funcobj.reference();
		else if ( strcmp(&func_name[0], "dlg_start" ) == 0 ) 
		  start_pfun = funcobj.reference();
		else if ( strcmp(&func_name[0], "dlg_enter" ) == 0 ) 
		  enter_pfun = funcobj.reference();
		else if ( strcmp(&func_name[0], "dlg_exit" ) == 0 ) 
		  exit_pfun = funcobj.reference();
		else if ( strcmp(&func_name[0], "dlg_before" ) == 0 ) 
		  before_pfun = funcobj.reference();
		else if ( strcmp(&func_name[0], "dlg_after" ) == 0 ) 
		  after_pfun = funcobj.reference();
		else
		  continue;

		npfuns++;
	    }
	}
    }
  // verify that all probe functions have been found
  if ( npfuns < 6 ) 
    {
      terminate_application();
      elg_error_msg("Missing probe function");
    }

  cerr << "done.\n";
}

/****************************************************************************
* install & activate probe
****************************************************************************/

void install_probe(ProbeExp& pexp, InstPoint& point)
{
  AisStatus sts;
  GCBTagType tag = (GCBTagType) 0;
  ProbeHandle phandle;
  GCBFuncType cb[1] = { data_cb };

  if ( ispoe )
    {
      sts = pappl.binstall_probe(1, &pexp, &point, cb, &tag, &phandle);
      if (sts.status() != ASC_success) 
	{
	  terminate_application();
	  elg_error_msg("PoeAppl::binstall_probe failed: %s", sts.status_name());
	}
      sts = pappl.bactivate_probe(1, &phandle);
      if (sts.status() != ASC_success) 
	{
	  terminate_application();
	  elg_error_msg("PoeAppl::bactivate_probe failed: %s", sts.status_name());
	}
    }
  else
    {
      sts = proc.binstall_probe(1, &pexp, &point, cb, &tag, &phandle);
      if (sts.status() != ASC_success) 
	{
	  terminate_application();
	  elg_error_msg("Process::binstall_probe failed: %s", sts.status_name());
	}
      sts = proc.bactivate_probe(1, &phandle);
      if (sts.status() != ASC_success) 
	{
	  terminate_application();
	  elg_error_msg("Process:bactivate_probe failed: %s", sts.status_name());
	}
    }
}

/****************************************************************************
* instrument user functions
****************************************************************************/

void instrument_functions()
{
  cerr << "Instrumenting user functions... ";

  // cerr << endl;

  for ( int i = 0; i < exec->get_num_funcs(); i++ )
    {
      Function* func = exec->get_func(i);
      ProbeExp exp(func->id);

      if ( !func->isuser || func->name.find("@") != string::npos )
	continue;

      if ( func->name == mainfun )
	{
	  ProbeExp start_call = start_pfun.call(1, &exp);
	  install_probe(start_call, func->entry);

	  // cerr << "INSTALL START " << func->name << endl;
	}
      else 
	{
	  ProbeExp exit_call  = exit_pfun.call(1, &exp);	  
	  install_probe(exit_call, func->exit);	  

	  // cerr << "INSTALL EXIT " << func->name << endl;
	}
    }
  cerr << "done.\n";
}

/****************************************************************************
* instrument call sites
****************************************************************************/

void instrument_call_sites()
{
  cerr << "Instrumenting call sites...     ";

  // cerr << endl;

  for ( int i = 0; i < exec->get_num_csites(); i++ )
    {
      Callsite* csite = exec->get_csite(i);
      ProbeExp exp(csite->id);

      if ( !csite->isuser )
	continue;
      
      if ( csite->callee->isuser || csite->callee->ismpicoll )
	{
	  ProbeExp before_call = before_pfun.call(1, &exp);

	  install_probe(before_call, csite->before);
	  
	  // cerr << "INSTALL BEFORE " << csite->callee->name << endl;
	}
      else
	{
	  ProbeExp before_call = before_pfun.call(1, &exp);
	  ProbeExp after_call  = after_pfun.call(1, &exp);
	  
	  install_probe(before_call, csite->before);	  
	  install_probe(after_call, csite->after);	  

	  // cerr << "INSTALL BEFORE " << csite->callee->name << endl;
	  // cerr << "INSTALL AFTER  " << csite->callee->name << endl;
	}
    }
  cerr << "done.\n";
}

/****************************************************************************
* initialize probe module
****************************************************************************/

void init_probes()
{
  cerr << "Initializing probe module...    ";

  AisStatus sts;

  ProbeExp init_call = init_pfun.call(0, NULL);

  if ( ispoe )
    {
      sts = pappl.bexecute(init_call, &data_cb, (GCBTagType) 0);
      if (sts.status() != ASC_success)
	{
	  terminate_application();
	  elg_error_msg("PoeAppl::bexecute failed: %s", sts.status_name());
	}
    }
  else
    {
      sts = proc.bexecute(init_call, &data_cb, (GCBTagType) 0);
      if (sts.status() != ASC_success)
	{
	  terminate_application();
	  elg_error_msg("Process::bexecute failed: %s", sts.status_name());
	}
    }
  cerr << "done.\n";
}

/****************************************************************************
* start application
****************************************************************************/

void start_application()
{
  cerr << "Starting application...\n\n";

  if ( ispoe )
    {
      sts = pappl.bstart();
      if (sts.status() != ASC_success) 
	{
	  terminate_application();
	  elg_error_msg("PoeAppl::bstart failed %s", sts.status_name());
	}
    }
  else
    {
      sts = proc.bstart();
      if (sts.status() != ASC_success) 
	{
	  terminate_application();
	  elg_error_msg("Process::bstart failed %s", sts.status_name());
	}
    }
}

/****************************************************************************
* main
****************************************************************************/

int main(int argc, char* argv[])
{
  parse_cmdl(argc, argv);
  init_dpcl();
  create_application(argc, argv);
                                           
  cerr << "Parsing application...          ";
  exec = new Executable(proc, basename(prog_name));
  cerr << "done.\n";
  
  std::ofstream out(DLG_APPL_DESCR_FILE);
  out << *exec;
  //cerr << *exec;
  out.close();
  
  locate_probe_functions();
  instrument_functions();
  instrument_call_sites();
  init_probes();       
  remove(DLG_APPL_DESCR_FILE);
  start_application();
  
  Ais_main_loop();
  cerr << "\nELG_DPCL end.\n";
  return 0;
}

