/****************************************************************************
**  KOJAK                                                                  **
**  http://www.fz-juelich.de/zam/kojak/                                    **
*****************************************************************************
**  Copyright (c) 1998-2004                                                **
**  Forschungszentrum Juelich, Zentralinstitut fuer Angewandte Mathematik  **
**                                                                         **
**  Copyright (c) 2003-2004                                                **
**  University of Tennessee, Innovative Computing Laboratory               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#include "elg_gen.h"
#include "elg_impl.h"
#include "elg_error.h"
#include "elg_metric.h"
#include "elg_pform.h"
#include "elg_trc.h"

#include <unistd.h>
#include <string.h>
#include <stdlib.h>

/*
 *-----------------------------------------------------------------------------
 * Macro functions
 *-----------------------------------------------------------------------------
 */

#define ELGGEN_ALLOC(gen, bytes)                                    \
  if ((gen->buf->pos - gen->buf->mem) > (gen->buf->size - (bytes))) \
      ElgGen_flush(gen)
 
#define ELGGEN_COPY_NUM(gen, var)           \
  memcpy(gen->buf->pos, &var, sizeof(var)); \
  gen->buf->pos += sizeof(var)       

#define ELGGEN_COPY_STR(gen, str)           \
  strcpy((char*) gen->buf->pos, str);       \
  gen->buf->pos += strlen(str) + 1     

#define ELGGEN_COPY_STRN(gen, str, n)       \
  strncpy((char*) gen->buf->pos, str, n);   \
  gen->buf->pos += n     
                         
/*
 *-----------------------------------------------------------------------------
 * ElgGen
 *-----------------------------------------------------------------------------
 */

/* Data types */

typedef struct 
{
  buffer_t  mem;
  buffer_t  pos;
  size_t    size;
} ElgBuf;

struct ElgGen_struct
{
  FILE*             file;
  char              name[PATH_MAX];
  ElgBuf*           buf;
};

ElgGen* ElgGen_open(const char* path_name, 
		    size_t buffer_size)
{
  ElgGen* gen;

  elg_ui1 major_vnr  = ELG_MAJOR_VNR;
  elg_ui1 minor_vnr  = ELG_MINOR_VNR;
  elg_ui1 byte_order = ELG_BYTE_ORDER;

  /* allocate ElgGen record */

  gen = malloc(sizeof(ElgGen));
  if (gen == NULL) 
    elg_error();
    
  /* open file */

  gen->file = fopen(path_name, "w");
  if (gen->file == NULL)
    elg_error_msg("Cannot open trace file %s", path_name);
  else
    elg_cntl_msg("Opened trace file %s for generation", path_name);

  /* store file name */
  strcpy(gen->name, path_name);

  /* check buffer_size */

  if (buffer_size < ELG_MIN_BUFSIZE)
    buffer_size = ELG_MIN_BUFSIZE;

  /* allocate buffer record */

  gen->buf = malloc(sizeof(ElgBuf));  
  if (gen->buf == NULL) 
    elg_error();

  /* allocate buffer */

  gen->buf->mem = malloc(buffer_size);  
  if (gen->buf->mem == NULL) 
    elg_error();
  
  /* initalize buffer */
  
  gen->buf->pos  = gen->buf->mem;
  /* subtraction leaves space for size of ELG_ENTER_DUMP record */
  gen->buf->size = buffer_size - ( 2 * sizeof(elg_ui1) + sizeof(elg_d8) + ELG_METRIC_MAXNUM * sizeof(elg_ui8) );  

  /* write file header */

  ELGGEN_ALLOC(gen, strlen(ELG_HEADER) + 1 + 
	       sizeof(major_vnr) + sizeof(minor_vnr) + sizeof(byte_order)); 

  ELGGEN_COPY_STR(gen, ELG_HEADER);
  ELGGEN_COPY_NUM(gen, major_vnr);
  ELGGEN_COPY_NUM(gen, minor_vnr);
  ELGGEN_COPY_NUM(gen, byte_order);
  
  /* return */
  return gen;
}

void ElgGen_flush(ElgGen* gen)
{
  int fd            = fileno(gen->file);
  size_t bytes;    

  /* write ELG_ENTER_DUMP record in full buffer */
  elg_enter_dump();

  /* flush buffer */
  bytes = gen->buf->pos - gen->buf->mem;
  if (write(fd, gen->buf->mem, bytes) != bytes)
    elg_error();
  
  /* reset buffer */
  gen->buf->pos  = gen->buf->mem;

  /* write ELG_EXIT_DUMP record in fresh buffer */
  elg_exit_dump();
  
  elg_cntl_msg("Flushed trace file %s", gen->name);
}

void ElgGen_close(ElgGen* gen)
{
  FILE* file   = gen->file;
  int fd       = fileno(gen->file);
  size_t bytes = gen->buf->pos - gen->buf->mem;

 /* flush buffer */
  if (write(fd, gen->buf->mem, bytes) != bytes)
    elg_error();

  /* close file */
  if (fclose(file) == EOF)
    elg_error();
  else
    elg_cntl_msg("Closed trace file %s", gen->name);    

  /* free buffer memory */
  free(gen->buf->mem); 

  /* free buffer record */
  free(gen->buf); 
  
  /* free gen record */
  free(gen); 
}

char* ElgGen_get_name(ElgGen* gen)
{
  return gen->name;
}


/* -- Writing trace records -- */


/* - Definition records - */


void ElgGen_write_STRING(ElgGen* gen,
			 elg_ui4 strid,
			 elg_ui1 cntc,
			 const char* str) 
{
  elg_ui1 length;
  elg_ui1 type    = ELG_STRING;
  
  /* Does string exceed record size ? */
  if (!cntc) 
    {
      length = sizeof(strid) + sizeof(cntc) + strlen(str) + 1;
      
      ELGGEN_ALLOC(gen, sizeof(length) + sizeof(type) + length);
      
      ELGGEN_COPY_NUM(gen, length);
      ELGGEN_COPY_NUM(gen, type);
      ELGGEN_COPY_NUM(gen, strid);
      ELGGEN_COPY_NUM(gen, cntc);
      ELGGEN_COPY_STR(gen, str);
    }
  else
    {
      length = 255;

      ELGGEN_ALLOC(gen, sizeof(length) + sizeof(type) + 255);
      
      ELGGEN_COPY_NUM(gen, length);
      ELGGEN_COPY_NUM(gen, type);
      ELGGEN_COPY_NUM(gen, strid);
      ELGGEN_COPY_NUM(gen, cntc);
      ELGGEN_COPY_STRN(gen, str, 255 - (sizeof(strid) + sizeof(cntc)));
    }
}

void ElgGen_write_STRING_CNT(ElgGen* gen,
			     const char* str) 
{
  elg_ui1 length;
  elg_ui1 type      = ELG_STRING_CNT;
  int     stringlen = strlen(str) + 1;

  /* Does string exceed record size ? */
  if (stringlen <= 255) 
    {
      length = stringlen;
      
      ELGGEN_ALLOC(gen, sizeof(length) + sizeof(type) + length);
      
      ELGGEN_COPY_NUM(gen, length);
      ELGGEN_COPY_NUM(gen, type);
      ELGGEN_COPY_STR(gen, str);
    }
  else
    {
      length = 255;

      ELGGEN_ALLOC(gen, sizeof(length) + sizeof(type) + 255);
      
      ELGGEN_COPY_NUM(gen, length);
      ELGGEN_COPY_NUM(gen, type);
      ELGGEN_COPY_STRN(gen, str, 255);
    }  
}

void ElgGen_write_MACHINE(ElgGen* gen,
			  elg_ui4 mid,
			  elg_ui4 nodec,
			  elg_ui4 mnid) 
{
  elg_ui1 length = sizeof(mid) + sizeof(nodec) + sizeof(mnid);
  elg_ui1 type   = ELG_MACHINE;
  
  ELGGEN_ALLOC(gen, sizeof(length) + sizeof(type) + length);

  ELGGEN_COPY_NUM(gen, length);
  ELGGEN_COPY_NUM(gen, type);
  ELGGEN_COPY_NUM(gen, mid);
  ELGGEN_COPY_NUM(gen, nodec);
  ELGGEN_COPY_NUM(gen, mnid);
}

void ElgGen_write_NODE(ElgGen* gen,
		       elg_ui4 nid,
		       elg_ui4 mid,
		       elg_ui4 cpuc,
		       elg_ui4 nnid,
		       elg_d8 cr) 
{
  elg_ui1 length = sizeof(nid) + sizeof(mid) + sizeof(cpuc) + sizeof(nnid) + sizeof(cr);
  elg_ui1 type   = ELG_NODE;
  
  ELGGEN_ALLOC(gen, sizeof(length) + sizeof(type) + length);

  ELGGEN_COPY_NUM(gen, length);
  ELGGEN_COPY_NUM(gen, type);
  ELGGEN_COPY_NUM(gen, nid);
  ELGGEN_COPY_NUM(gen, mid);
  ELGGEN_COPY_NUM(gen, cpuc);
  ELGGEN_COPY_NUM(gen, nnid);
  ELGGEN_COPY_NUM(gen, cr);
}

void ElgGen_write_PROCESS(ElgGen* gen,
			  elg_ui4 pid,
			  elg_ui4 pnid) 
{
  elg_ui1 length = sizeof(pid) + sizeof(pnid);
  elg_ui1 type   = ELG_PROCESS;
  
  ELGGEN_ALLOC(gen, sizeof(length) + sizeof(type) + length);

  ELGGEN_COPY_NUM(gen, length);
  ELGGEN_COPY_NUM(gen, type);
  ELGGEN_COPY_NUM(gen, pid);
  ELGGEN_COPY_NUM(gen, pnid);
}

void ElgGen_write_THREAD(ElgGen* gen,
			 elg_ui4 tid,
			 elg_ui4 pid,
			 elg_ui4 tnid) 
{
  elg_ui1 length = sizeof(tid) + sizeof(pid) + sizeof(tnid);
  elg_ui1 type   = ELG_THREAD;
  
  ELGGEN_ALLOC(gen, sizeof(length) + sizeof(type) + length);

  ELGGEN_COPY_NUM(gen, length);
  ELGGEN_COPY_NUM(gen, type);
  ELGGEN_COPY_NUM(gen, tid);
  ELGGEN_COPY_NUM(gen, pid);
  ELGGEN_COPY_NUM(gen, tnid);
}

void ElgGen_write_LOCATION(ElgGen* gen,
			   elg_ui4 lid,
			   elg_ui4 mid,
			   elg_ui4 nid,
			   elg_ui4 pid,
			   elg_ui4 tid) 
{
  elg_ui1 length = 
    sizeof(lid) + 
    sizeof(mid) + 
    sizeof(nid) + 
    sizeof(pid) + 
    sizeof(tid);

  elg_ui1 type   = ELG_LOCATION;
  
  ELGGEN_ALLOC(gen, sizeof(length) + sizeof(type) + length);

  ELGGEN_COPY_NUM(gen, length);
  ELGGEN_COPY_NUM(gen, type);
  ELGGEN_COPY_NUM(gen, lid);
  ELGGEN_COPY_NUM(gen, mid);
  ELGGEN_COPY_NUM(gen, nid);
  ELGGEN_COPY_NUM(gen, pid);
  ELGGEN_COPY_NUM(gen, tid);
}

void ElgGen_write_FILE(ElgGen* gen,
		       elg_ui4 fid,
		       elg_ui4 fnid) 
{
  elg_ui1 length = sizeof(fid) + sizeof(fnid);
  elg_ui1 type   = ELG_FILE;
  
  ELGGEN_ALLOC(gen, sizeof(length) + sizeof(type) + length);

  ELGGEN_COPY_NUM(gen, length);
  ELGGEN_COPY_NUM(gen, type);
  ELGGEN_COPY_NUM(gen, fid);
  ELGGEN_COPY_NUM(gen, fnid);
}

void ElgGen_write_REGION(ElgGen* gen,
			 elg_ui4 rid,
			 elg_ui4 rnid,
			 elg_ui4 fid,
			 elg_ui4 begln,
			 elg_ui4 endln,
			 elg_ui4 rdid,
			 elg_ui1 rtype) 
{
  elg_ui1 length = 
    sizeof(rid) + 
    sizeof(rnid) + 
    sizeof(fid) + 
    sizeof(begln) + 
    sizeof(endln) + 
    sizeof(rdid) + 
    sizeof(rtype);

  elg_ui1 type   = ELG_REGION;
  
  ELGGEN_ALLOC(gen, sizeof(length) + sizeof(type) + length);

  ELGGEN_COPY_NUM(gen, length);
  ELGGEN_COPY_NUM(gen, type);
  ELGGEN_COPY_NUM(gen, rid);
  ELGGEN_COPY_NUM(gen, rnid);
  ELGGEN_COPY_NUM(gen, fid);
  ELGGEN_COPY_NUM(gen, begln);
  ELGGEN_COPY_NUM(gen, endln);
  ELGGEN_COPY_NUM(gen, rdid);
  ELGGEN_COPY_NUM(gen, rtype);
}

void ElgGen_write_CALL_SITE( ElgGen* gen,
			     elg_ui4 csid,
			     elg_ui4 fid,
			     elg_ui4 lno,
			     elg_ui4 erid,
			     elg_ui4 lrid )
{
  elg_ui1 length = 
    sizeof(csid) + 
    sizeof(fid) + 
    sizeof(lno) + 
    sizeof(erid) + 
    sizeof(lrid); 

  elg_ui1 type   = ELG_CALL_SITE;
  
  ELGGEN_ALLOC(gen, sizeof(length) + sizeof(type) + length);

  ELGGEN_COPY_NUM(gen, length);
  ELGGEN_COPY_NUM(gen, type);
  ELGGEN_COPY_NUM(gen, csid);
  ELGGEN_COPY_NUM(gen, fid);
  ELGGEN_COPY_NUM(gen, lno);
  ELGGEN_COPY_NUM(gen, erid);
  ELGGEN_COPY_NUM(gen, lrid);
}

void ElgGen_write_METRIC( ElgGen* gen,
			  elg_ui4 metid,
			  elg_ui4 metnid,
			  elg_ui4 metdid,
			  elg_ui1 metdtype,
			  elg_ui1 metmode,
			  elg_ui1 metiv )
{
  elg_ui1 length = 
    sizeof(metid) + 
    sizeof(metnid) + 
    sizeof(metdid) + 
    sizeof(metdtype) + 
    sizeof(metmode) + 
    sizeof(metiv); 

  elg_ui1 type   = ELG_METRIC;
  
  ELGGEN_ALLOC(gen, sizeof(length) + sizeof(type) + length);

  ELGGEN_COPY_NUM(gen, length);
  ELGGEN_COPY_NUM(gen, type);
  ELGGEN_COPY_NUM(gen, metid);
  ELGGEN_COPY_NUM(gen, metnid);
  ELGGEN_COPY_NUM(gen, metdid);
  ELGGEN_COPY_NUM(gen, metdtype);
  ELGGEN_COPY_NUM(gen, metmode);
  ELGGEN_COPY_NUM(gen, metiv);
}

void ElgGen_write_MPI_COMM(ElgGen* gen,
			   elg_ui4 cid,
			   elg_ui4 grpc,
			   elg_ui1 grpv[]) 
{
  int i;
  elg_ui1 length = sizeof(cid) + sizeof(grpc) + grpc * sizeof(elg_ui1);
  elg_ui1 type   = ELG_MPI_COMM;
  
  ELGGEN_ALLOC(gen, sizeof(length) + sizeof(type) + length);

  ELGGEN_COPY_NUM(gen, length);
  ELGGEN_COPY_NUM(gen, type);
  ELGGEN_COPY_NUM(gen, cid);
  ELGGEN_COPY_NUM(gen, grpc);

  for (i = 0; i < grpc; i++)
    {
      ELGGEN_COPY_NUM(gen, grpv[i]);
    }
}


void ElgGen_write_CART_TOPOLOGY ( ElgGen* gen,
				  elg_ui4 topid,
				  elg_ui1 ndims,
				  elg_ui4 dims[],
				  elg_ui1 periods[] )
{
  int i;

  elg_ui1 length = sizeof(topid) +
                   sizeof(ndims) +
                   ndims * sizeof(elg_ui4) +
                   ndims * sizeof(elg_ui1);

  elg_ui1 type   = ELG_CART_TOPOLOGY;


  ELGGEN_ALLOC(gen, sizeof(length) + sizeof(type) + length);

  ELGGEN_COPY_NUM(gen, length);
  ELGGEN_COPY_NUM(gen, type);
  ELGGEN_COPY_NUM(gen, topid);
  ELGGEN_COPY_NUM(gen, ndims);

  for ( i = 0 ; i < ndims ; i++ )
    {
      ELGGEN_COPY_NUM(gen, dims[i]);
    }

  for ( i = 0 ; i < ndims ; i++ )
    {
      ELGGEN_COPY_NUM(gen, periods[i]);
    }    
}

void ElgGen_write_CART_COORDS   ( ElgGen* gen,
				  elg_ui4 topid,
				  elg_ui4 lid,
				  elg_ui1 ndims,
				  elg_ui4 coords[] )
{
  int i;
  elg_ui1 length = sizeof(topid) +
                   sizeof(lid) +
                   sizeof(ndims) + 
                   ndims * sizeof(elg_ui4);

  elg_ui1 type = ELG_CART_COORDS;

  ELGGEN_ALLOC(gen, sizeof(length) + sizeof(type) + length);

  ELGGEN_COPY_NUM(gen, length);
  ELGGEN_COPY_NUM(gen, type);
  ELGGEN_COPY_NUM(gen, topid);
  ELGGEN_COPY_NUM(gen, lid);
  ELGGEN_COPY_NUM(gen, ndims);

  for ( i = 0 ; i < ndims ; i++ ) 
    {
      ELGGEN_COPY_NUM(gen, coords[i]);     
    }  
}


void ElgGen_write_OFFSET(ElgGen* gen,
			 elg_d8  ltime,
			 elg_d8  offset) 
{
  elg_ui1 length = sizeof(ltime) + sizeof(offset);
  elg_ui1 type   = ELG_OFFSET;

  ELGGEN_ALLOC(gen, sizeof(length) + sizeof(type) + length);

  ELGGEN_COPY_NUM(gen, length);
  ELGGEN_COPY_NUM(gen, type);
  ELGGEN_COPY_NUM(gen, ltime);
  ELGGEN_COPY_NUM(gen, offset);
}

void ElgGen_write_NUM_EVENTS( ElgGen* gen,
			      elg_ui4 eventc )
{
  elg_ui1 length = sizeof(eventc);
  elg_ui1 type   = ELG_NUM_EVENTS;

  ELGGEN_ALLOC(gen, sizeof(length) + sizeof(type) + length);

  ELGGEN_COPY_NUM(gen, length);
  ELGGEN_COPY_NUM(gen, type);
  ELGGEN_COPY_NUM(gen, eventc);
}

void ElgGen_write_LAST_DEF(ElgGen* gen) 
{
  elg_ui1 length = 0;
  elg_ui1 type   = ELG_LAST_DEF;

  ELGGEN_ALLOC(gen, sizeof(length) + sizeof(type));

  ELGGEN_COPY_NUM(gen, length);
  ELGGEN_COPY_NUM(gen, type);
}

/* - Event records - */

void ElgGen_write_ENTER( ElgGen* gen,
			 elg_ui4 rid,
			 elg_ui1 metc,
			 elg_ui8 metv[] )
{
  int i;
  elg_ui1 type   = ELG_ENTER;
  elg_d8  time   = elg_pform_wtime();
  elg_ui1 length = 
    sizeof(time) + 
    sizeof(rid) +
    (metc                ? metc * sizeof(elg_ui8) : 0);
   
  ELGGEN_ALLOC(gen, sizeof(length) + sizeof(type) + length);

  ELGGEN_COPY_NUM(gen, length);
  ELGGEN_COPY_NUM(gen, type);
  ELGGEN_COPY_NUM(gen, time);  
  ELGGEN_COPY_NUM(gen, rid);

  if (metc) 
    for (i = 0; i < metc; i++)
      {
	ELGGEN_COPY_NUM(gen, metv[i]);
      }
}

void ElgGen_write_ENTER_CS( ElgGen* gen,
			    elg_ui4 csid,
			    elg_ui1 metc,
			    elg_ui8 metv[] )
{
  int i;
  elg_ui1 type   = ELG_ENTER_CS;
  elg_d8  time   = elg_pform_wtime();
  elg_ui1 length = 
    sizeof(time) + 
    sizeof(csid) +
    (metc ? metc * sizeof(elg_ui8) : 0);
   
  ELGGEN_ALLOC(gen, sizeof(length) + sizeof(type) + length);

  ELGGEN_COPY_NUM(gen, length);
  ELGGEN_COPY_NUM(gen, type);
  ELGGEN_COPY_NUM(gen, time);  
  ELGGEN_COPY_NUM(gen, csid);

  if (metc) 
    for (i = 0; i < metc; i++)
      {
	ELGGEN_COPY_NUM(gen, metv[i]);
      }
}

void ElgGen_write_EXIT(ElgGen* gen,
		       elg_ui1 metc,
		       elg_ui8 metv[])
{
  int i;
  elg_ui1 type   = ELG_EXIT;
  elg_d8  time   = elg_pform_wtime();
  elg_ui1 length = 
    sizeof(time) + 
    (metc ? metc * sizeof(elg_ui8) : 0);
   
  ELGGEN_ALLOC(gen, sizeof(length) + sizeof(type) + length);

  ELGGEN_COPY_NUM(gen, length);
  ELGGEN_COPY_NUM(gen, type);
  ELGGEN_COPY_NUM(gen, time);  

  if (metc) 
    for (i = 0; i < metc; i++)
      {
	ELGGEN_COPY_NUM(gen, metv[i]);
      }
}

void ElgGen_write_MPI_SEND(ElgGen* gen,
			   elg_ui4 dpid,
			   elg_ui4 cid,
			   elg_ui4 tag,      
			   elg_ui4 sent)
{
  elg_ui1 type   = ELG_MPI_SEND;
  elg_d8  time   = elg_pform_wtime();
  elg_ui1 length = 
    sizeof(time) + 
    sizeof(dpid) + 
    sizeof(cid)  + 
    sizeof(tag)  + 
    sizeof(sent);

  ELGGEN_ALLOC(gen, sizeof(length) + sizeof(type) + length);

  ELGGEN_COPY_NUM(gen, length);
  ELGGEN_COPY_NUM(gen, type);
  ELGGEN_COPY_NUM(gen, time);
  ELGGEN_COPY_NUM(gen, dpid);
  ELGGEN_COPY_NUM(gen, cid);
  ELGGEN_COPY_NUM(gen, tag);
  ELGGEN_COPY_NUM(gen, sent);
}

void ElgGen_write_MPI_RECV(ElgGen* gen,
			   elg_ui4 spid,
			   elg_ui4 cid,
			   elg_ui4 tag)
{
  elg_ui1 type   = ELG_MPI_RECV;
  elg_d8  time   = elg_pform_wtime();
  elg_ui1 length = 
    sizeof(time) + 
    sizeof(spid) + 
    sizeof(cid)  + 
    sizeof(tag);

  ELGGEN_ALLOC(gen, sizeof(length) + sizeof(type) + length);

  ELGGEN_COPY_NUM(gen, length);
  ELGGEN_COPY_NUM(gen, type);
  ELGGEN_COPY_NUM(gen, time);
  ELGGEN_COPY_NUM(gen, spid);
  ELGGEN_COPY_NUM(gen, cid);
  ELGGEN_COPY_NUM(gen, tag);
}

void ElgGen_write_MPI_COLLEXIT(ElgGen* gen,
			       elg_ui1 metc,
			       elg_ui8 metv[], 
			       elg_ui4 rpid,
			       elg_ui4 cid,
			       elg_ui4 sent,  
			       elg_ui4 recvd)
{
  int i;
  elg_ui1 type   = ELG_MPI_COLLEXIT;
  elg_d8  time   = elg_pform_wtime();
  elg_ui1 length = 
    sizeof(time) + 
    sizeof(rpid) + 
    sizeof(cid) + 
    sizeof(sent) + 
    sizeof(recvd) + 
    (metc ? metc * sizeof(elg_ui8) : 0);

  ELGGEN_ALLOC(gen, sizeof(length) + sizeof(type) + length);

  ELGGEN_COPY_NUM(gen, length);
  ELGGEN_COPY_NUM(gen, type);
  ELGGEN_COPY_NUM(gen, time);
  if (metc) 
    for (i = 0; i < metc; i++)
      {
	ELGGEN_COPY_NUM(gen, metv[i]);
      }
  ELGGEN_COPY_NUM(gen, rpid);
  ELGGEN_COPY_NUM(gen, cid);
  ELGGEN_COPY_NUM(gen, sent);
  ELGGEN_COPY_NUM(gen, recvd);
}

void ElgGen_write_OMP_FORK(ElgGen* gen)
{
  elg_ui1 type   = ELG_OMP_FORK;
  elg_d8  time   = elg_pform_wtime();
  elg_ui1 length = sizeof(time); 

  ELGGEN_ALLOC(gen, sizeof(length) + sizeof(type) + length);

  ELGGEN_COPY_NUM(gen, length);
  ELGGEN_COPY_NUM(gen, type);
  ELGGEN_COPY_NUM(gen, time);
}

void ElgGen_write_OMP_JOIN(ElgGen* gen)
{
  elg_ui1 type   = ELG_OMP_JOIN;
  elg_d8  time   = elg_pform_wtime();
  elg_ui1 length = sizeof(time); 

  ELGGEN_ALLOC(gen, sizeof(length) + sizeof(type) + length);

  ELGGEN_COPY_NUM(gen, length);
  ELGGEN_COPY_NUM(gen, type);
  ELGGEN_COPY_NUM(gen, time);
}

void ElgGen_write_OMP_ALOCK(ElgGen* gen,
			    elg_ui4 lkid)
{
  elg_ui1 type   = ELG_OMP_ALOCK;
  elg_d8  time   = elg_pform_wtime();
  elg_ui1 length = 
    sizeof(time) + 
    sizeof(lkid); 
   
  ELGGEN_ALLOC(gen, sizeof(length) + sizeof(type) + length);

  ELGGEN_COPY_NUM(gen, length);
  ELGGEN_COPY_NUM(gen, type);
  ELGGEN_COPY_NUM(gen, time);
  ELGGEN_COPY_NUM(gen, lkid);
}

void ElgGen_write_OMP_RLOCK(ElgGen* gen,
			    elg_ui4 lkid)
{
  elg_ui1 type   = ELG_OMP_RLOCK;
  elg_d8  time   = elg_pform_wtime();
  elg_ui1 length = 
    sizeof(time) + 
    sizeof(lkid); 
   
  ELGGEN_ALLOC(gen, sizeof(length) + sizeof(type) + length);

  ELGGEN_COPY_NUM(gen, length);
  ELGGEN_COPY_NUM(gen, type);
  ELGGEN_COPY_NUM(gen, time);
  ELGGEN_COPY_NUM(gen, lkid);
}

void ElgGen_write_OMP_COLLEXIT(ElgGen* gen,
			       elg_ui1 metc,
			       elg_ui8 metv[])
{
  int i;
  elg_ui1 type   = ELG_OMP_COLLEXIT;
  elg_d8  time   = elg_pform_wtime();
  elg_ui1 length = 
    sizeof(time) + 
    (metc ? metc * sizeof(elg_ui8) : 0);
   
  ELGGEN_ALLOC(gen, sizeof(length) + sizeof(type) + length);

  ELGGEN_COPY_NUM(gen, length);
  ELGGEN_COPY_NUM(gen, type);
  ELGGEN_COPY_NUM(gen, time);  

  if (metc) 
    for (i = 0; i < metc; i++)
      {
	ELGGEN_COPY_NUM(gen, metv[i]);
      }
}

/* Tracing event records */

void ElgGen_write_ENTER_DUMP(ElgGen* gen,
			     elg_ui1 metc,
			     elg_ui8 metv[])
{
  int i;
  elg_ui1 type   = ELG_ENTER_DUMP;
  elg_d8  time   = elg_pform_wtime();
  elg_ui1 length = sizeof(time) +
    (metc ? metc * sizeof(elg_ui8) : 0); 
  
  ELGGEN_COPY_NUM(gen, length);
  ELGGEN_COPY_NUM(gen, type);
  ELGGEN_COPY_NUM(gen, time);

  if (metc) 
    for (i = 0; i < metc; i++)
      {
	ELGGEN_COPY_NUM(gen, metv[i]);
      }
}

void ElgGen_write_EXIT_DUMP(ElgGen* gen,
			    elg_ui1 metc,
			    elg_ui8 metv[])
{
  int i;
  elg_ui1 type   = ELG_EXIT_DUMP;
  elg_d8  time   = elg_pform_wtime();
  elg_ui1 length = sizeof(time) +
    (metc ? metc * sizeof(elg_ui8) : 0); 
  
  ELGGEN_COPY_NUM(gen, length);
  ELGGEN_COPY_NUM(gen, type);
  ELGGEN_COPY_NUM(gen, time);

  if (metc) 
    for (i = 0; i < metc; i++)
      {
	ELGGEN_COPY_NUM(gen, metv[i]);
      }
}

void ElgGen_write_LOG_OFF(ElgGen* gen,
			  elg_ui1 metc,
			  elg_ui8 metv[])
{
  int i;
  elg_ui1 type   = ELG_LOG_OFF;
  elg_d8  time   = elg_pform_wtime();
  elg_ui1 length = sizeof(time) +
    (metc ? metc * sizeof(elg_ui8) : 0); 
  
  ELGGEN_ALLOC(gen, sizeof(length) + sizeof(type) + length);

  ELGGEN_COPY_NUM(gen, length);
  ELGGEN_COPY_NUM(gen, type);
  ELGGEN_COPY_NUM(gen, time);

  if (metc) 
    for (i = 0; i < metc; i++)
      {
	ELGGEN_COPY_NUM(gen, metv[i]);
      }
}

void ElgGen_write_LOG_ON(ElgGen* gen,
			 elg_ui1 metc,
			 elg_ui8 metv[])
{
  int i;
  elg_ui1 type   = ELG_LOG_ON;
  elg_d8  time   = elg_pform_wtime();
  elg_ui1 length = sizeof(time) +
    (metc ? metc * sizeof(elg_ui8) : 0); 
  
  ELGGEN_ALLOC(gen, sizeof(length) + sizeof(type) + length);

  ELGGEN_COPY_NUM(gen, length);
  ELGGEN_COPY_NUM(gen, type);
  ELGGEN_COPY_NUM(gen, time);

  if (metc) 
    for (i = 0; i < metc; i++)
      {
	ELGGEN_COPY_NUM(gen, metv[i]);
      }
}
