/****************************************************************************
**  KOJAK                                                                  **
**  http://www.fz-juelich.de/zam/kojak/                                    **
*****************************************************************************
**  Copyright (c) 1998-2004                                                **
**  Forschungszentrum Juelich, Zentralinstitut fuer Angewandte Mathematik  **
**                                                                         **
**  Copyright (c) 2003-2004                                                **
**  University of Tennessee                                                **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#include "elg_mrgproc.h"
#include "elg_rw.h"
#include "elg_env.h"
#include "elg_error.h"
#include <set>
#include <map>
#include <string>
#include <vector>
#include <algorithm>
#include <cstddef>

#define ELG_MAX_STRING 4096

using namespace std;
#include <iostream>

/*
 *---------------------------------------------------------------------------------
 * Declaration of types and functions
 *---------------------------------------------------------------------------------
 */
  
struct Infile;
static double get_time_stamp(ElgRec* rec);
static void   set_time_stamp(ElgRec* rec, double time);
static void   merge_def_recs(int numnodes);
static void   merge_event_recs();

/* region */

struct Region
{
  elg_ui4 rnid;
  elg_ui4 fid;
  elg_ui4 begln;
  elg_ui4 endln;
};

bool 
operator>(const Region& lhs, const Region& rhs)
{
  if (lhs.rnid > rhs.rnid)
    return true;
  else if (lhs.rnid == rhs.rnid)
    if (lhs.fid > rhs.fid)
      return true;
    else if (lhs.fid == rhs.fid)
      if (lhs.begln > rhs.begln)
	return true;
      else if (lhs.begln == rhs.begln)
	if (lhs.endln > rhs.endln)
	  return true;
	else
	  return false;
      else
	return false;
    else
      return false;
  else
    return false;
}

bool 
operator==(const Region& lhs, const Region& rhs)
{
  return (lhs.rnid  == rhs.rnid  &&
	  lhs.fid   == rhs.fid   &&
	  lhs.begln == rhs.begln &&
	  lhs.endln == rhs.endln);
}

bool 
operator<(const Region& lhs, const Region& rhs)
{
  return (!(lhs == rhs || lhs > rhs));  
}

/* call site */

struct Csite
{
  elg_ui4 fid;
  elg_ui4 lno;
  elg_ui4 erid;
  elg_ui4 lrid;
};

bool 
operator>(const Csite& lhs, const Csite& rhs)
{
  if (lhs.fid > rhs.fid)
    return true;
  else if (lhs.fid == rhs.fid)
    if (lhs.lno > rhs.lno)
      return true;
    else if (lhs.lno == rhs.lno)
      if (lhs.lrid > rhs.lrid)
	return true;
      else if (lhs.erid == rhs.erid)
	if (lhs.lrid > rhs.lrid)
	  return true;
	else
	  return false;
      else
	return false;
    else
      return false;
  else
    return false;
}

bool 
operator==(const Csite& lhs, const Csite& rhs)
{
  return (lhs.fid  == rhs.fid  &&
	  lhs.lno  == rhs.lno  &&
	  lhs.erid == rhs.erid &&
	  lhs.lrid == rhs.lrid);
}

bool 
operator<(const Csite& lhs, const Csite& rhs)
{
  return (!(lhs == rhs || lhs > rhs));  
}

/*
 *---------------------------------------------------------------------------------
 * Global variables 
 *---------------------------------------------------------------------------------
 */

ElgOut* of;

vector<Infile*> infilev;

/*-- strings --*/

/* string |-> global string id */
map<string, elg_ui4> str2gstrid;

/*-- location related data --*/

/* physical machine id |-> enumerated machine id */
map<elg_ui4, elg_ui4> pmid2emid;      
/* physical node id |-> enumerated node id */
map<elg_ui4, elg_ui4> pnid2enid;      
/* set of defined machines */
set<elg_ui4> machines;
/* set of defined nodes (machine, node) */
set<pair<elg_ui4, elg_ui4> > nodes;
/* set of named processes */
set<elg_ui4> processes;
/* set of named threads (process, thread) */
set<pair<elg_ui4, elg_ui4> > threads;
/* process id |-> global location id of (-,-,p,0) */
map<elg_ui4, elg_ui4> pid2glid;
/* set of defined metrics */
set<elg_ui4> metrics;
/* set of defined topologis */
set<elg_ui4> tops;


/*-- performance counter related data --*/

/* number of performance metrics */
elg_ui1 metc = 0; 

/*-- source-code related data --*/

/* global file name id |-> global file id */
map<elg_ui4, elg_ui4> gfnid2gfid;

/* Region |-> global region id */
map<Region, elg_ui4> region2grid; 

/* Csite |-> global call-site id */
map<Csite, elg_ui4> csite2gcsid; 

/* mpi group |-> list of communicator ids in the order of creation */
map<vector<elg_ui1>, vector<elg_ui4> > mpigrp2gcids;

/*-- unique ids --*/
int curstrid  = 0; /* string */
int curmid    = 0; /* machine */
int curnid    = 0; /* node */
int curlid    = 0; /* location */
int currid    = 0; /* region */
int curcsid   = 0; /* call site */
int curfid    = 0; /* file */
int curcid    = 0; /* communicator */

/*-- number of event records --*/
long eventc = 0;

/*
 *---------------------------------------------------------------------------------
 * struct Infile
 *---------------------------------------------------------------------------------
 */

struct Infile
{
  Infile(char* file_name);
  ~Infile()                                         { ElgIn_close(inf); }

  void rewind()                                     { ElgIn_rewind(inf); }

  ElgRec* next_def_rec();
  ElgRec* next_event_rec();
  
  double corrected_time(double time);
  
  ElgIn* inf;
    
  /* location id in case of locationless event records */
  elg_ui4 glid; 

  /* local |-> global */
  map<elg_ui4, elg_ui4> lstrid2gstrid;   // string
  map<elg_ui4, elg_ui4> llid2glid;       // location
  map<elg_ui4, elg_ui4> lfid2gfid;       // file
  map<elg_ui4, elg_ui4> lrid2grid;       // region
  map<elg_ui4, elg_ui4> lcsid2gcsid;     // call site
  map<elg_ui4, elg_ui4> lcid2gcid;       // communicator
  map<elg_ui4, elg_ui4> gcid2topid;      // topologies      

  // cartcoordsv temporarily stores the ELG_CART_COORDS records
  // which are finally written in ELG_LOCATION
  map<elg_ui4, ElgRec*> lid2cart;           

  /* chronological offsets to global time and corresponding local times */
  double offset[2];
  double ltime[2]; 
  int osidx; 
};

Infile::Infile(char* file_name) : 
  inf(ElgIn_open(file_name)), 
  osidx(0)
{
  if (!inf)
    elg_error_msg("Cannot open file \"%s\"", file_name);

  /* default values for post-mortem clock synchronization */
  this->offset[0] = 0.0;
  this->offset[1] = 0.0;
  this->ltime[0]  = 0.0;
  this->ltime[1]  = 1.0;

  lstrid2gstrid[ELG_NO_ID] = ELG_NO_ID;  // predefine (global) empty string
  lfid2gfid[ELG_NO_ID]     = ELG_NO_ID;  // predefine (global) empty file
}

double
Infile::corrected_time(double time)
{
  return time + (((offset[1] - offset[0]) / (ltime[1] - ltime[0])) * 
		 (time - ltime[0])) + offset[0];
}

ElgRec* 
Infile::next_def_rec()
{
  ElgRec* rec;
 
  while((rec = ElgIn_read_record(inf)))
    if (!ElgRec_is_event(rec))
      return rec;
    else
      eventc++;
  return NULL;
}

ElgRec* 
Infile::next_event_rec()
{
  ElgRec* rec;

  while((rec = ElgIn_read_record(inf)) != NULL)
    {
      elg_ui1 rec_type = ElgRec_get_type(rec);
      
      if (ElgRec_is_event(rec))
	{
	  int pos = 0;
	  
	  /* reset byte offset for reading */
	  ElgRec_seek(rec, 0);
	  
	  /* correct location id */
	  elg_ui4 llid, glid;

	  llid = ElgRec_read_ui4(rec);
	  glid = llid2glid[llid];	      
	  ElgRec_write_ui4(rec, glid , 0);
	  pos += sizeof(glid);
	  
	  /* correct time stamp */
	  elg_d8  oldtime, newtime; 

	  oldtime = ElgRec_read_d8(rec);
	  newtime = corrected_time(oldtime); 	  
	  ElgRec_write_d8(rec, newtime, pos);
	  pos += sizeof(newtime);
	  
	  switch(rec_type)
	    {
	    case ELG_ENTER:  
	      {
		elg_ui4 lrid, grid;

		/* correct region id */
		lrid = ElgRec_read_ui4(rec);
		grid = lrid2grid[lrid];
		ElgRec_write_ui4(rec, grid, pos);
		break;
	      }
	    case ELG_ENTER_CS:  
	      {
		elg_ui4 lcsid, gcsid;

		/* correct call-site id */
		lcsid = ElgRec_read_ui4(rec);
		gcsid = lcsid2gcsid[lcsid];
		ElgRec_write_ui4(rec, gcsid, pos);
		break;
	      }
	    case ELG_MPI_SEND:
	    case ELG_MPI_RECV:
	      {
		elg_ui4 lcid, gcid, spid, gslid;

		/* Correct src/dest location id  */
		spid  = ElgRec_read_ui4(rec);
		gslid = ( spid == ELG_NO_ID ) ? ELG_NO_ID : pid2glid[spid];
		ElgRec_write_ui4(rec, gslid, pos);
		pos += sizeof(gslid);

		/* Correct communicator id */
		lcid = ElgRec_read_ui4(rec);
		gcid = lcid2gcid[lcid];
		ElgRec_write_ui4(rec, gcid, pos);

		break;
	      }
	    case ELG_MPI_COLLEXIT:
	      {
		elg_ui4 lcid, gcid, rpid, grlid;

		/* Jump over performance counter */
		pos += metc * sizeof(elg_ui8);
		ElgRec_seek(rec, pos);

		/* Correct root location id  */
		rpid  = ElgRec_read_ui4(rec);
		grlid = ( rpid == ELG_NO_ID ) ? ELG_NO_ID : pid2glid[rpid];
		ElgRec_write_ui4(rec, grlid, pos);
		pos += sizeof(grlid);

		/* Correct communicator id */
		lcid = ElgRec_read_ui4(rec);
		gcid = lcid2gcid[lcid];
		ElgRec_write_ui4(rec, gcid, pos);

		break;
	      }
	    }
	  return rec;
	}
      free(rec);
    }
  return NULL;
}
  
/*
 *---------------------------------------------------------------------------------
 * Definition of local functions
 *---------------------------------------------------------------------------------
 */

static double get_time_stamp(ElgRec* rec)
{
  ElgRec_seek(rec, sizeof(elg_ui4));
  
  return ElgRec_read_d8(rec); 
}

static void set_time_stamp(ElgRec* rec, double time)
{
  ElgRec_write_d8(rec, time, sizeof(elg_ui4));
}


static void merge_def_recs(int num_nodes)
{
  ElgRec* rec;
  elg_ui4 topid;
  
  topid = 0;

  for (int i = 0; i < infilev.size(); i++)
    {
      /*
	Temporary table used for counting how many times a
	communicator is created for a certain group.  
      */

      map<vector<elg_ui1>, int> nummpigrps;
      
      while((rec = infilev[i]->next_def_rec()))
	{
	  switch(ElgRec_get_type(rec))
	    {
	    case ELG_STRING :
	      {
		elg_ui4 lstrid;
		int cntc = 0;
		char str[ELG_MAX_STRING];
		char* strptr = str;
		vector<ElgRec*> cntvec;

		lstrid = ElgRec_read_ui4(rec);
		cntc = ElgRec_read_ui1(rec);
		sprintf(strptr, "%.250s", ElgRec_read_string(rec));
		strptr += 250;

		/* Are there any continuation rercords? */       
		while(cntc)
		  {
		    ElgRec* cntrec;
		    
		    cntrec = infilev[i]->next_def_rec();
		    sprintf(strptr, "%.255s", ElgRec_read_string(cntrec));
		    cntvec.push_back(cntrec);
		    strptr += 255;
		    cntc--;
		  }

		map<string, elg_ui4>::const_iterator it = str2gstrid.find(str);
		
		/* Is string not yet defined ? */
		if (it == str2gstrid.end())
		  {
		    elg_ui4 gstrid = curstrid++;
		    
		    str2gstrid[str] = gstrid;
		    infilev[i]->lstrid2gstrid[lstrid] = gstrid;

		    ElgRec_write_ui4(rec, gstrid, 0);
		    
		    /* write out ELG_STRING(_CNT) records */
		    ElgOut_write_record(of, rec);
		    for (int j=0; j < cntvec.size(); j++)
		      {
			ElgOut_write_record(of, cntvec[j]);
			ElgRec_free(cntvec[j]);
		      }
		  }
		else
		  {
		    infilev[i]->lstrid2gstrid[lstrid] =  str2gstrid[str];
		  }
		break;
	      }
	    case ELG_MACHINE :               
	      {
		elg_ui4 pmid, emid, lmnid, gmnid; 

		pmid = ElgRec_read_ui4(rec);
		ElgRec_seek(rec, 2 * sizeof(elg_ui4));
		lmnid = ElgRec_read_ui4(rec);

		set<elg_ui4>::const_iterator it = machines.find(pmid);
		
		/* Is machine not yet defined ? */
		if (it == machines.end())
		  {
		    machines.insert(pmid);
		    emid = curmid++;
		    pmid2emid[pmid] = emid;
		    gmnid = infilev[i]->lstrid2gstrid[lmnid];
		    ElgRec_write_ui4(rec, emid, 0);
		    if ( num_nodes != -1 )
		      {
		        ElgRec_write_ui4(rec, num_nodes, sizeof(elg_ui4));
		      }
		    ElgRec_write_ui4(rec, gmnid, 2 * sizeof(elg_ui4));
		    ElgOut_write_record(of, rec);
		  }
		break;
	      }
	    case ELG_NODE :                  
	      {
		elg_ui4 pnid, enid, pmid, emid, lnnid, gnnid;
		
		pnid = ElgRec_read_ui4(rec);
		pmid = ElgRec_read_ui4(rec);
		ElgRec_seek(rec, 3 * sizeof(elg_ui4));
		lnnid = ElgRec_read_ui4(rec);

		set<pair<elg_ui4, elg_ui4> >::const_iterator it = nodes.find(make_pair(pnid, pmid));
		
		/* Is node not yet defined ? */
		if (it == nodes.end())
		  {
		    nodes.insert(make_pair(pnid, pmid));
		    emid = pmid2emid[pmid];
		    enid = curnid++;
		    pnid2enid[pnid] = enid;
		    gnnid = infilev[i]->lstrid2gstrid[lnnid];
		    ElgRec_write_ui4(rec, enid, 0);
		    ElgRec_write_ui4(rec, emid, sizeof(elg_ui4));
		    ElgRec_write_ui4(rec, gnnid, 3 * sizeof(elg_ui4));
		    ElgOut_write_record(of, rec);
		  }
		break;
	      }
	    case ELG_PROCESS :               
	      {
		elg_ui4 pid, lpnid, gpnid; 

		pid    = ElgRec_read_ui4(rec);
		lpnid = ElgRec_read_ui4(rec);

		set<elg_ui4>::const_iterator it = processes.find(pid);
		
		/* Is process not yet defined ? */
		if (it == processes.end())
		  {
		    processes.insert(pid);
		    gpnid = infilev[i]->lstrid2gstrid[lpnid];
		    ElgRec_write_ui4(rec, gpnid, 2 * sizeof(elg_ui4));
		    ElgOut_write_record(of, rec);
		  }
		break;
	      }
	    case ELG_THREAD :                
	      {
		elg_ui4 tid, pid, ltnid, gtnid;
		
		tid    = ElgRec_read_ui4(rec);
		pid    = ElgRec_read_ui4(rec);
		ltnid = ElgRec_read_ui4(rec);

		set<pair<elg_ui4, elg_ui4> >::const_iterator it = 
		  threads.find(make_pair(tid, pid));
		
		/* Is thread not yet defined ? */
		if (it == threads.end())
		  {
		    threads.insert(make_pair(tid, pid));
		    gtnid = infilev[i]->lstrid2gstrid[ltnid];
		    ElgRec_write_ui4(rec, gtnid, 2 * sizeof(elg_ui4));
		    ElgOut_write_record(of, rec);
		  }
		break;
	      }
	    case ELG_LOCATION :              
	      {	
       		/* 
		   It is assumed that each location is defined only
		   once, so it is not necessary to check whether it is
		   already defined.  
		*/

	        elg_ui4 pmid, emid, pnid, enid, llid, glid, pid, tid;
		ElgRec* cartrec; 
		
		llid = ElgRec_read_ui4(rec);
		pmid = ElgRec_read_ui4(rec);
		pnid = ElgRec_read_ui4(rec);
		pid  = ElgRec_read_ui4(rec);
		tid  = ElgRec_read_ui4(rec);
		glid = curlid++;

		infilev[i]->glid = glid;

		emid = pmid2emid[pmid];
		enid = pnid2enid[pnid];
		if (tid == 0)
		  pid2glid[pid] = glid;
		infilev[i]->llid2glid[llid] = glid;
		ElgRec_write_ui4(rec, glid, 0);
		ElgRec_write_ui4(rec, emid, sizeof(elg_ui4));
		ElgRec_write_ui4(rec, enid, 2 * sizeof(elg_ui4));	
		ElgOut_write_record(of, rec);		

		// write the global location ids in the 
		// ELG_CART_COORDS definition record		
		
		if ( infilev[i]->lid2cart.size() != 0 ) 
		  {		    
		    if( infilev[i]->lid2cart.find(llid) != 
			infilev[i]->lid2cart.end() )
		      {	
			cartrec = infilev[i]->lid2cart[llid];
			infilev[i]->lid2cart.erase(llid);
			ElgRec_write_ui4(cartrec, glid, sizeof(elg_ui4));
			ElgOut_write_record(of, cartrec);
			// free the records in infilev[i]->cartv
			ElgRec_free(cartrec);
		      }		    		  
		  }		  

		break;
	      }
	    case ELG_FILE :                  
	      {
		elg_ui4 lfid, lfnid, gfid, gfnid; 
		string fname;

		lfid  = ElgRec_read_ui4(rec);
		lfnid = ElgRec_read_ui4(rec);
		gfnid = infilev[i]->lstrid2gstrid[lfnid];

		map<elg_ui4, elg_ui4>::const_iterator it = gfnid2gfid.find(gfnid);
		
		/* Is file not yet defined ? */
		if (it == gfnid2gfid.end())
		  {
		    gfid = curfid++;
		    gfnid2gfid[gfnid] = gfid;
		    infilev[i]->lfid2gfid[lfid] = gfid;
		    ElgRec_write_ui4(rec, gfid, 0);
		    ElgRec_write_ui4(rec, gfnid, sizeof(elg_ui4));
		    ElgOut_write_record(of, rec);
		  }
		else
		  {
		    infilev[i]->lfid2gfid[lfid] = gfnid2gfid[gfnid];
		  }
		break;
	      }
	    case ELG_REGION :                
	      {
		elg_ui4 lrid, lrnid, lfid, lrdid;
		elg_ui4 grid, grnid, gfid, grdid; 
		string rname;
		Region region;

		lrid  = ElgRec_read_ui4(rec);
		lrnid = ElgRec_read_ui4(rec);
		lfid  = ElgRec_read_ui4(rec);
		region.begln = ElgRec_read_ui4(rec);
		region.endln = ElgRec_read_ui4(rec);
		lrdid = ElgRec_read_ui4(rec);
		
		region.rnid = grnid = infilev[i]->lstrid2gstrid[lrnid];
		region.fid  = gfid  = infilev[i]->lfid2gfid[lfid];
		grdid = infilev[i]->lstrid2gstrid[lrdid];

		map<Region, elg_ui4>::const_iterator it = 
		  region2grid.find(region);
		
		/* Is region not yet defined ? */
		if (it == region2grid.end())
		  {
		    grid = currid++;
		    region2grid[region] = grid;
		    infilev[i]->lrid2grid[lrid] = grid;
		    ElgRec_write_ui4(rec, grid, 0);
		    ElgRec_write_ui4(rec, grnid,     sizeof(elg_ui4));
		    ElgRec_write_ui4(rec, gfid,  2 * sizeof(elg_ui4));
		    ElgRec_write_ui4(rec, grdid, 5 * sizeof(elg_ui4));
		    ElgOut_write_record(of, rec);
		  }
		else
		  {
		    infilev[i]->lrid2grid[lrid] = 
		      region2grid[region];
		  }
		break;
	      }
	    case ELG_CALL_SITE :                
	      {
		elg_ui4 lcsid, lfid, lerid, llrid;
		elg_ui4 gcsid, gfid, gerid, glrid;

		Csite csite;

		lcsid     = ElgRec_read_ui4(rec);
		lfid      = ElgRec_read_ui4(rec);
		csite.lno = ElgRec_read_ui4(rec);
		lerid     = ElgRec_read_ui4(rec);
		llrid     = ElgRec_read_ui4(rec);

		csite.fid  = gfid  = infilev[i]->lfid2gfid[lfid];
		csite.erid = gerid = infilev[i]->lrid2grid[lerid];

		if ( llrid == ELG_NO_ID )
		  csite.lrid = glrid = ELG_NO_ID;
		else
		  csite.lrid = glrid = infilev[i]->lrid2grid[llrid];

		map<Csite, elg_ui4>::const_iterator it = 
		  csite2gcsid.find(csite);
		
		/* Is csite not yet defined ? */
		if (it == csite2gcsid.end())
		  {
		    gcsid = curcsid++;
		    csite2gcsid[csite] = gcsid;
		    infilev[i]->lcsid2gcsid[lcsid] = gcsid;
		    ElgRec_write_ui4(rec, gcsid, 0);
		    ElgRec_write_ui4(rec, gfid,      sizeof(elg_ui4));
		    ElgRec_write_ui4(rec, gerid, 3 * sizeof(elg_ui4));
		    ElgRec_write_ui4(rec, glrid, 4 * sizeof(elg_ui4));
		    ElgOut_write_record(of, rec);
		  }
		else
		  {
		    infilev[i]->lcsid2gcsid[lcsid] = csite2gcsid.find(csite)->second;
		    // infilev[i]->lcsid2gcsid[lcsid] = csite2gcsid[csite];
		  }
		break;
	      }
	    case ELG_METRIC :                
	      {
		/* 
		   It is assumed that each metric carries its global
		   identifier.

		   Also, it is assumed that the metric identifiers are
		   alread enumerated from 0 to n-1.
		*/
		
		elg_ui4 metid = ElgRec_read_ui4(rec);

		set<elg_ui4>::const_iterator it = metrics.find(metid);
		
		/* Is metric not yet defined ? */
		if (it == metrics.end())
		  {
		    elg_ui4 lmetnid, lmetdid;
		    elg_ui4 gmetnid, gmetdid; 
		    
		    metrics.insert(metid);

		    lmetnid = ElgRec_read_ui4(rec);
		    lmetdid = ElgRec_read_ui4(rec);
		    
		    gmetnid = infilev[i]->lstrid2gstrid[lmetnid];
		    gmetdid = infilev[i]->lstrid2gstrid[lmetdid];
		    
		    ElgRec_write_ui4(rec, gmetnid, sizeof(elg_ui4));
		    ElgRec_write_ui4(rec, gmetdid, 2 * sizeof(elg_ui4));
		    ElgOut_write_record(of, rec);

		    // increase metric counter by one
		    metc++;
		  }
		break;
	      }
	    case ELG_MPI_COMM :            
	      {
		elg_ui4 lcid, gcid, grpc;
		lcid = ElgRec_read_ui4(rec);
		grpc = ElgRec_read_ui4(rec);
		vector<elg_ui1> mpigrp;
	    
		for (int j = 0; j < grpc; j++)
		  mpigrp.push_back(ElgRec_read_ui1(rec));
		
		if (nummpigrps.find(mpigrp) == nummpigrps.end())
		  nummpigrps[mpigrp] = 1;
		else
		  nummpigrps[mpigrp]++;
		
		/* Is communicator not yet defined ? */ 
		if (mpigrp2gcids[mpigrp].size() < nummpigrps[mpigrp])
		  {
		    gcid = curcid++;
		    mpigrp2gcids[mpigrp].push_back(gcid);
		    infilev[i]->lcid2gcid[lcid] = gcid;
		    ElgRec_write_ui4(rec, gcid, 0);
		    ElgOut_write_record(of, rec);
		  }
		else
		  {
		    infilev[i]->lcid2gcid[lcid] = 
		      mpigrp2gcids[mpigrp][nummpigrps[mpigrp] - 1];
		  }
		break;
	      }
	    case ELG_CART_TOPOLOGY :
	      {
		elg_ui4 lcid, gcid;	

		lcid = ElgRec_read_ui4(rec);	
		gcid =  infilev[i]->lcid2gcid[lcid];

		if ( tops.find(gcid) == tops.end() ) 
		  {
		    tops.insert(gcid);
		    infilev[i]->gcid2topid[gcid] = ++topid;		   
		    ElgRec_write_ui4(rec, topid, 0);
		    ElgOut_write_record(of, rec);
		  }		
		else 
		  {
		    infilev[i]->gcid2topid[gcid] = topid;
		  }
		break;
	      }
	      // ELG_CART_COORDS is temporarily created here,
	      // but is eventually written in ELG_LOCATION
	    case ELG_CART_COORDS :
	      {
		elg_ui4 lcid, gcid, topid, lid;		
		
					
		lcid = ElgRec_read_ui4(rec);
		lid = ElgRec_read_ui4(rec);
		gcid = infilev[i]->lcid2gcid[lcid];	
		topid = infilev[i]->gcid2topid[gcid];		
		
		ElgRec_write_ui4(rec, topid, 0);		
	
		// push the rec Record in cartcoordsv
		infilev[i]->lid2cart[lid] = rec;			
		break;
	      }
	    case ELG_OFFSET :               
	      {
		elg_d8 ltime, offset;
		
		ltime  = ElgRec_read_d8(rec);
		offset = ElgRec_read_d8(rec);
			
		infilev[i]->ltime[infilev[i]->osidx]  = ltime;
		infilev[i]->offset[infilev[i]->osidx] = offset;
		infilev[i]->osidx++;
		break;
	      }
	    }

	  if ( ElgRec_get_type(rec) != ELG_CART_COORDS )
	    ElgRec_free(rec);
	}    
    }  

  /* write ELG_NUM_EVENTS record */
  elg_cntl_msg("Counted %d events", eventc);
  rec = ElgRec_create_NUM_EVENTS(eventc);
  ElgOut_write_record(of, rec);
  free(rec);

  /* write ELG_LAST_DEF record */
  rec = ElgRec_create_LAST_DEF();
  ElgOut_write_record(of, rec);
  free(rec);
}

static void merge_event_recs()
{
  vector<ElgRec*> recv(infilev.size());
  double start_time, timestamp;
  bool is_first=true;
  
  /* rewind */
  for_each(infilev.begin(), infilev.end(), mem_fun(&Infile::rewind)); 
 
  /* init */
  for (int i = 0; i < recv.size(); i++)
      recv[i] = infilev[i]->next_event_rec();

  while(1)
    {
      ElgRec* min_rec = NULL;
      int min_pos;      
      
      for (int i = 0; i < infilev.size() ; i++)
	{
	  if (recv[i] == NULL)
	      continue;
	  /* select first available record */
	  else if (min_rec == NULL)
	    {
	      min_rec = recv[i];	      
	      min_pos = i;
	    }
	  /* compare timestamps: looking for oldest (minimum time stamp) record */     
	  else
	    {
	      double first  = get_time_stamp(min_rec);
	      double second = get_time_stamp(recv[i]);
	      
	      if (first > second)
		{
		  min_rec = recv[i];
		  min_pos = i;
		}
	    }
	}
      /* finish if all records are processed */
      if (min_rec == NULL)
	return;

      /* replace oldest record by new record */
      recv[min_pos] = infilev[min_pos]->next_event_rec();
      
      /* adjust time stamps in a way that the first event has time stamp 0.0 */
      timestamp = get_time_stamp(min_rec);
      
       if (is_first)
	{
	  start_time = timestamp;
	  is_first = false;
	}
      set_time_stamp(min_rec, timestamp - start_time);

      /* write record to outfile */
      ElgOut_write_record(of, min_rec);

      ElgRec_free(min_rec);
    }
}

/*
 *---------------------------------------------------------------------------------
 * Definition of extern functions
 *---------------------------------------------------------------------------------
 */

void elg_mrgproc(int infilec, char* fprefix, int num_nodes)
{
  int i;

  /* open infiles */
  for (i = 0; i < infilec; i++)
    {
      char* infile_name = (char*) calloc(strlen(fprefix) + 9, sizeof(char));
      sprintf(infile_name, "%s.%d.elg", fprefix, i);

      Infile* infile = new Infile(infile_name);
      infilev.push_back(infile);

      free(infile_name);
    }

  /* open outfile */

  char* outfile_name = (char*) calloc(strlen(fprefix) + 9, sizeof(char));
  sprintf(outfile_name, "%s.elg", fprefix);

  if (! (of = ElgOut_open(outfile_name)) )
      elg_error_msg("Cannot open file \"%s\"", outfile_name);

  free(outfile_name);

  /* merge definition records (pass 1) */
  merge_def_recs(num_nodes);

  /* merge event records (pass 2) */
  merge_event_recs();

  /* close outfile */
  ElgOut_close(of);  

  /* close infiles */
  for (i = 0; i < infilev.size(); i++)
    delete infilev[i];

  /* delete infiles */
  for (i = 0; i < infilev.size(); i++)
    {
      char* infile_name = (char*) calloc(strlen(fprefix) + 9, sizeof(char));
      sprintf(infile_name, "%s.%d.elg", fprefix, i);
      
      int result = remove(infile_name);  
      if (result == 0)
	elg_cntl_msg("Removed trace file %s", infile_name);
      else
	elg_error_msg("Cannot remove trace file %s", infile_name);
      free(infile_name);
    }
}  

// Usage: elg_merge <#files> <prefix> [<numnodes>]

int main(int argc, char** argv)
{
  char*  fprefix;
  char*  rest;
  int infilec = 0;
  int num_nodes = -1;

  /* ckeck # of arguments */
  if (argc < 2 || argc > 4)
    elg_error_msg("Wrong # of arguments\nUsage: elg_merge <#files> "
	          "<prefix> [<numnodes>]");
    
  infilec = strtol(argv[1], &rest, 10);
  if (rest == argv[1])
    elg_error_msg("Integer expected as 1st argument\nUsage: elg_merge "
	          "<#files> <prefix> [<numnodes>]");

  if (argc > 2)
    fprefix = argv[2];
  else
    fprefix = elg_env_fprefix();

  if (argc > 3)
    {
      num_nodes = strtol(argv[3], &rest, 10);
      if (rest == argv[3])
        elg_error_msg("Integer expected as 1st argument\nUsage: elg_merge "
	              "<#files> <prefix> [<numnodes>]");
    }

  /* merge */
  elg_mrgproc(infilec, fprefix, num_nodes);

  exit(0);
}
















