/****************************************************************************
**  KOJAK                                                                  **
**  http://www.fz-juelich.de/zam/kojak/                                    **
*****************************************************************************
**  Copyright (C) 1998-2003                                                **
**  Forschungszentrum Juelich, Zentralinstitut fuer Angewandte Mathematik  **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#include "elg_pomp2_ctc.h"

/*
 * Initialize CTC field descriptor
 */
int ctc_init(ctc_field* f, char* str) {
  int len = 0;
  char* curr = str;

  /* -- check parameters -- */
  if ( f == 0 || str == 0 ) return 0;

  /* -- determine length of ctc string -- */
  while ( isdigit(*curr) ) {
    len = len*10 + *curr - '0';
    ++curr;
  }

  /* -- initialialize ctc_string if successful -- */
  if ( len > 0 && *curr == '*' ) {
    f->str = malloc(len * sizeof(char));
    strcpy(f->str, curr+1);
    f->attr = "";
    f->val  = "";
    f->curr = f->str;
    return 1;
  } else {
    return 0;
  }
}

/*
 * Get next CTC field
 */
int ctc_get_next_field(ctc_field* f) {
  char* curr;

  /* -- check parameters -- */
  if ( f == 0 ) return 0;

  /* -- no more fields? -- */
  curr = f->curr;
  if ( *curr == '*' ) return 0;

  /* -- find next attr=value pair */
  f->attr = curr;
  while ( *curr && *curr != '=' ) ++curr;
  if ( *curr == '=' ) {
    *curr = '\0';
    ++curr;
    f->val = curr;
    while ( *curr && *curr != '*' ) ++curr;
    if ( *curr == '*' ) {
      *curr = '\0';
      ++curr;
      f->curr = curr;
      return 1;
    }
  }
  return 0;
}

/*
 * Determine CTC Field Type
 */

char *ctc_field_type_2_str[] = {
  "escl", "group", "hascopyin", "hascopypriv", "hasfpriv", "haslpriv",
  "hasnowait", "hasord", "hasred", "name", "rtype", "sscl", "schedule",
  "unknown"
};

enum ctc_field_t ctc_field_type(char *str) {
  switch ( str[0] ) {
  case 'e': if ( strcmp(str, "escl") == 0 ) return CTC_ESCL;
            break;
  case 'g': if ( strcmp(str, "group") == 0 ) return CTC_GROUP;
            break;
  case 'h': switch ( str[3] ) {
        case 'c': if ( strcmp(str, "hascopyin") == 0 ) return CTC_HASCOPYIN;
                  if ( strcmp(str, "hascopypriv") == 0 ) return CTC_HASCOPYPRIV;
                  break;
        case 'f': if ( strcmp(str, "hasfpriv") == 0 ) return CTC_HASFPRIV;
                  break;
        case 'l': if ( strcmp(str, "haslpriv") == 0 ) return CTC_HASLPRIV;
                  break;
        case 'n': if ( strcmp(str, "hasnowait") == 0 ) return CTC_HASNOWAIT;
                  break;
        case 'o': if ( strcmp(str, "hasord") == 0 ) return CTC_HASORD;
                  break;
        case 'r': if ( strcmp(str, "hasred") == 0 ) return CTC_HASRED;
                  break;
        }
        break;
  case 'n': if ( strcmp(str, "name") == 0 ) return CTC_NAME;
            break;
  case 'r': if ( strcmp(str, "rtype") == 0 ) return CTC_RTYPE;
            break;
  case 's': if ( strcmp(str, "sscl") == 0 ) return CTC_SSCL;
            if ( strcmp(str, "schedule") == 0 ) return CTC_SCHEDULE;
            break;
  }
  return CTC_UNKNOWN;
}

/*
 * Determine CTC Region Type
 */

void conv_ctc_region_type(char *str, elg_ui1* rtype, char** rname) {
  switch ( str[0] ) {
  case 'a': if ( strcmp(str, "api") == 0 ) {
	      *rtype = ELG_FUNCTION;
	      return;
	    }
            if ( strcmp(str, "atomic") == 0 ) {
	      *rtype = ELG_OMP_ATOMIC;
	      *rname = "!$omp atomic";
	      return;
	    }
            break;
  case 'b': if ( strcmp(str, "barrier") == 0 ) {
	      *rtype = ELG_OMP_BARRIER;
              *rname = "!$omp barrier";
	      return;
	    }
            break;
  case 'c': if ( strcmp(str, "critical") == 0 ) {
	      *rtype = ELG_OMP_CRITICAL;
	      *rname = "!$omp critical";
	      return;
	    }
            break;
  case 'e': if ( strcmp(str, "ebarrier") == 0 ) {
	      *rtype = ELG_OMP_BARRIER;
              *rname = "!$omp barrier";
	      return;
	    }
            break;
  case 'f': if ( strcmp(str, "flush") == 0 ) {
	      *rtype = ELG_OMP_FLUSH;
	      *rname = "!$omp flush";
	      return;
	    }
            if ( strcmp(str, "func") == 0 ) {
	      *rtype = ELG_FUNCTION;
	      return;
	    }
            break;
  case 'i': if ( strcmp(str, "ibarrier") == 0 ) {
	      *rtype = ELG_OMP_IBARRIER;
              *rname = "!$omp ibarrier";
	      return;
	    }
            break;
  case 'l': if ( strcmp(str, "loop") == 0 ) {
	      *rtype = ELG_OMP_LOOP;
	      *rname = "!$omp do/for";
	      return;
	    }
            break;
  case 'm': if ( strcmp(str, "master") == 0 ) {
	      *rtype = ELG_OMP_MASTER;
	      *rname = "!$omp master";
	      return;
	    }
            break;
  case 'o': if ( strcmp(str, "ordered") == 0 ) {
	      /*NOT USED BY EPILOG YET*/
              *rtype = ELG_UNKNOWN;
	      return;
	    }
            break;
  case 'p': if ( strcmp(str, "pregion") == 0 ) {
	      *rtype = ELG_OMP_PARALLEL;
	      *rname = "!$omp parallel";
	      return;
	    }
            break;
  case 's': if ( strcmp(str, "section") == 0 ) {
	      *rtype = ELG_OMP_SECTIONS;
	      *rname = "!$omp sections";
	      return;
	    }
            if ( strcmp(str, "single") == 0 ) {
	      *rtype = ELG_OMP_SINGLE;
	      *rname = "!$omp single";
	      return;
	    }
            break;
  case 'u': if ( strcmp(str, "user") == 0 ) {
	      *rtype = ELG_USER_REGION;
	      return;
	    }
            break;
  case 'w': if ( strcmp(str, "wshare") == 0 ) {
	      *rtype = ELG_OMP_WORKSHARE;
	      *rname = "!$omp workshare";
	      return;
	    }
            break;
  }
  *rtype = ELG_UNKNOWN;
}
