/****************************************************************************
**  KOJAK                                                                  **
**  http://www.fz-juelich.de/zam/kojak/                                    **
*****************************************************************************
**  Copyright (c) 1998-2004                                                **
**  Forschungszentrum Juelich, Zentralinstitut fuer Angewandte Mathematik  **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "pomp_lib.h"

#include "elg_pomp.h"
#include "elg_omplock.h"
#include "elg_ompreg.h"
#include "elg_trc.h"

/*
 * Global variables
 */

int pomp_tracing = 0;

/*
 *----------------------------------------------------------------
 * EPILOG implementation of the C pomp function library
 *----------------------------------------------------------------
 */
   
void POMP_Finalize() {
  static int pomp_finalize_called = 0;

  if ( ! pomp_finalize_called ) {
    pomp_finalize_called = 1;
    elg_close();
  }
}

void POMP_Init() {
  static int pomp_init_called = 0;
  static struct ElgRegDescr rd_data_table[1000];
  int i;
  elg_ui1 rtype;
  char* rname;
  const char* rdesc;

  if ( ! pomp_init_called ) {
    pomp_init_called = 1;

    elg_open();
    atexit(POMP_Finalize);

    /* register wrapper functions for OpenMP API */
#ifdef ELG_OMP
    elg_omp_register();
#endif

    for(i = 0; i < POMP_MAX_ID; ++i) {
      if ( pomp_rd_table[i] ) {
        struct ElgRegDescr* data = &rd_data_table[i];
        struct ompregdescr* r    = pomp_rd_table[i];      
        r->data = data;
	rdesc   = "OMP";

        /* -- register file --*/
        data->fid   = elg_def_file(r->file_name);
        data->begln = r->begin_first_line;
        data->endln = r->end_last_line;
	data->sbrid = ELG_NO_ID;

        if (strcmp(r->name, "region") == 0)  {
          rtype = ELG_USER_REGION;
          rname = r->sub_name;
	  rdesc   = "USR";
#ifdef ELG_OMP
        } else if (strcmp(r->name, "atomic") == 0) {
          rtype = ELG_OMP_ATOMIC;
          rname = "!$omp atomic";
        } else if (strcmp(r->name, "barrier") == 0) {
          rtype = ELG_OMP_BARRIER;
          rname = "!$omp barrier";
        } else if (strcmp(r->name, "critical") == 0) {
          rtype = ELG_OMP_CRITICAL;
          rname = "!$omp critical";
          data->sbrid = elg_def_region("!$omp critical sblock",
			               data->fid, r->begin_last_line+1,
                                       r->end_first_line-1, "OMP",
         			       ELG_OMP_CRITICAL_SBLOCK);
	  if ( r->sub_name[0] ) {
            rname = malloc(24 + strlen(r->sub_name));
	    sprintf(rname, "!$omp critical %s", r->sub_name);
	    data->brid = elg_lock_init(r->sub_name);
	  } else {
	    data->brid = elg_lock_init("default");
          }
        } else if (strcmp(r->name, "do") == 0) {
          rtype = ELG_OMP_LOOP;
          rname = "!$omp do";
        } else if (strcmp(r->name, "flush") == 0) {
          rtype = ELG_OMP_FLUSH;
          rname = "!$omp flush";
        } else if (strcmp(r->name, "for") == 0) {
          rtype = ELG_OMP_LOOP;
          rname = "!$omp for";
        } else if (strcmp(r->name, "function") == 0)  {
          rtype = ELG_FUNCTION;
          rname = r->sub_name;
	  rdesc   = "USR";
        } else if (strcmp(r->name, "master") == 0) {
          rtype = ELG_OMP_MASTER;
          rname = "!$omp master";
        } else if (strcmp(r->name, "parallel") == 0)  {
          rtype = ELG_OMP_PARALLEL;
          rname = "!$omp parallel";
        } else if (strcmp(r->name, "paralleldo") == 0)  {
          rtype = ELG_OMP_PARALLEL;
          rname = "!$omp parallel";
          data->sbrid = elg_def_region("!$omp do",  data->fid, data->begln,
                                       data->endln, "OMP", ELG_OMP_LOOP);
        } else if (strcmp(r->name, "parallelfor") == 0)  {
          rtype = ELG_OMP_PARALLEL;
          rname = "!$omp parallel";
          data->sbrid = elg_def_region("!$omp for",  data->fid, data->begln,
                                       data->endln, "OMP", ELG_OMP_LOOP);
        } else if (strcmp(r->name, "parallelsections") == 0)  {
          rtype = ELG_OMP_PARALLEL;
          rname = "!$omp parallel";
          data->sbrid = elg_def_region("!$omp sections",
			               data->fid, data->begln,
                                       data->endln, "OMP", ELG_OMP_SECTIONS);
        } else if (strcmp(r->name, "parallelworkshare") == 0)  {
          rtype = ELG_OMP_PARALLEL;
          rname = "!$omp parallel";
          data->sbrid = elg_def_region("!$omp workshare",
			               data->fid, data->begln,
                                       data->endln, "OMP", ELG_OMP_WORKSHARE);
        } else if (strcmp(r->name, "sections") == 0) {
          rtype = ELG_OMP_SECTIONS;
          rname = "!$omp sections";
          data->sbrid = elg_def_region("!$omp section",
			               data->fid, r->begin_last_line,
                                       r->end_first_line, "OMP",
         			       ELG_OMP_SECTION);
        } else if (strcmp(r->name, "section") == 0) {
          /* NOT DEFINED BY POMP YET */
          /* rtype = ELG_OMP_SECTION; */
          /* rname = "!$omp section"; */
        } else if (strcmp(r->name, "single") == 0) {
          rtype = ELG_OMP_SINGLE;
          rname = "!$omp single";
          data->sbrid = elg_def_region("!$omp single sblock",
			               data->fid, r->begin_last_line+1,
                                       r->end_first_line-1, "OMP",
         			       ELG_OMP_SINGLE_SBLOCK);
        } else if (strcmp(r->name, "workshare") == 0) {
          rtype = ELG_OMP_WORKSHARE;
          rname = "!$omp workshare";
#endif
        } else {
          rtype = ELG_UNKNOWN;
          rname = r->name;
        }

        /* -- register region -- */
        data->rid = elg_def_region(rname,  data->fid, data->begln,
                                   data->endln, rdesc, rtype);

#ifdef ELG_OMP
        if (rtype == ELG_OMP_PARALLEL ||
            rtype == ELG_OMP_LOOP     ||
            rtype == ELG_OMP_SECTIONS ||
            rtype == ELG_OMP_SINGLE   ||
            rtype == ELG_OMP_WORKSHARE)
          /* -- register implicit barrier -- */
          data->brid = elg_def_region("!$omp ibarrier",  
                                      data->fid, data->endln, data->endln,
                                      "OMP", ELG_OMP_IBARRIER);
        else
#endif
          data->brid = ELG_NO_ID;
      }
    }
    pomp_tracing = 1;
  }
}

void POMP_Off() {
  pomp_tracing = 0;
}

void POMP_On() {
  pomp_tracing = 1;
}

void POMP_Begin(struct ompregdescr* r) {
  GUARDED_ENTER(rid);
}

void POMP_End(struct ompregdescr* r) {
  GUARDED_EXIT();
}
