/****************************************************************************
**  KOJAK                                                                  **
**  http://www.fz-juelich.de/zam/kojak/                                    **
*****************************************************************************
**  Copyright (c) 1998-2004                                                **
**  Forschungszentrum Juelich, Zentralinstitut fuer Angewandte Mathematik  **
**                                                                         **
**  Copyright (c) 2003-2004                                                **
**  University of Tennessee                                                **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#include <string.h>
#include <stdlib.h>
#include "elg_error.h"
#include "elg_rw.h"
#include "elg_readcb.h"

struct ElgRCB_struct
{
  ElgIn*   in;
  elg_ui1  metc;
  elg_ui4  single_lid;
  elg_ui8* metv;
  elg_ui1  activated[ELG_MAX_EVENT];
};

ElgRCB* elg_read_open(const char* path)
{
  int i;

  ElgRCB* handle = calloc(1, sizeof(ElgRCB));
  if (handle == NULL)
  {
    elg_warning("Can't allocate RCB handle");
    return NULL;
  }
  if ( (handle->in = ElgIn_open(path)) == NULL )
  {
    elg_warning("Can't open file %s", path);
    return NULL;
  }
  for (i=0; i<ELG_ALL_EVENTS; i++) handle->activated[i] = 1;
  return handle;
}

void elg_read_close( ElgRCB* handle )
{
  ElgIn_close(handle->in);
  if (handle->metc) free(handle->metv); 
  free(handle);
}

void elg_read_activate( ElgRCB* handle, elg_ui1 type )
{
  int i;

  if (type == ELG_ALL_EVENTS)
    for (i=0; i<ELG_ALL_EVENTS; i++) handle->activated[i] = 1;
  else
    handle->activated[type] = 1;
}

void elg_read_deactivate( ElgRCB* handle, elg_ui1 type )
{
  int i;

  if (type == ELG_ALL_EVENTS)
    for (i=0; i<ELG_ALL_EVENTS; i++) handle->activated[i] = 0;
  else
    handle->activated[type] = 0;
}

int elg_read_next( ElgRCB* handle, void* userdata )
{
  int i;
  elg_ui1 type, length;
  
  ElgRec* rec = ElgIn_read_record(handle->in);
  
  if ( rec == NULL )
    return 0;

  type   = ElgRec_get_type(rec);
  length = ElgRec_get_length(rec);

  elg_readcb_ALL(type, length, userdata);

  if (handle->activated[type])
  {
    switch(type)
    {
      /* -- Definition records -- */
      
    case ELG_STRING :
      {
	elg_ui4 strid = ElgRec_read_ui4(rec);
	elg_ui1 cntc  = ElgRec_read_ui1(rec);
	char* str = ElgRec_read_string(rec);	    

	elg_readcb_STRING(strid, cntc, str, userdata);
	break;
      }
    case ELG_STRING_CNT :           
      {
	char* str = ElgRec_read_string(rec);	    

	elg_readcb_STRING_CNT(str, userdata);
	break;
      }
    case ELG_MACHINE :               
      {
	elg_ui4 mid = ElgRec_read_ui4(rec);
	elg_ui4 nodec = ElgRec_read_ui4(rec);
	elg_ui4 mnid = ElgRec_read_ui4(rec);
	    
	elg_readcb_MACHINE(mid, nodec, mnid, userdata);
	break;
      }
    case ELG_NODE :                  
      {
	elg_ui4 nid = ElgRec_read_ui4(rec);
	elg_ui4 mid = ElgRec_read_ui4(rec);
	elg_ui4 cpuc = ElgRec_read_ui4(rec);
	elg_ui4 nnid = ElgRec_read_ui4(rec); 
	elg_d8  cr = ElgRec_read_d8(rec);
	    
	elg_readcb_NODE(nid, mid, cpuc, nnid, cr, userdata);
	break;
      }
    case ELG_PROCESS :               
      {
	elg_ui4 pid = ElgRec_read_ui4(rec);
	elg_ui4 pnid = ElgRec_read_ui4(rec);

	elg_readcb_PROCESS(pid, pnid, userdata);
	break;
      }
    case ELG_THREAD :                
      {
	elg_ui4 tid = ElgRec_read_ui4(rec);
	elg_ui4 pid = ElgRec_read_ui4(rec);
	elg_ui4 tnid = ElgRec_read_ui4(rec);

	elg_readcb_THREAD(tid, pid, tnid, userdata);
	break;
      }
    case ELG_LOCATION :              
      {
	elg_ui4 lid = ElgRec_read_ui4(rec);
	elg_ui4 mid = ElgRec_read_ui4(rec);
	elg_ui4 nid = ElgRec_read_ui4(rec);
	elg_ui4 pid = ElgRec_read_ui4(rec);
	elg_ui4 tid = ElgRec_read_ui4(rec);
	  
	handle->single_lid = lid;
	    
	elg_readcb_LOCATION(lid, mid, nid, pid, tid, userdata);
	break;
      }
    case ELG_FILE :                  
      {
	elg_ui4 fid = ElgRec_read_ui4(rec);
	elg_ui4 fnid = ElgRec_read_ui4(rec);

	elg_readcb_FILE(fid, fnid, userdata);
	break;
      }
    case ELG_REGION :                
      {
	elg_ui4 rid = ElgRec_read_ui4(rec);
	elg_ui4 rnid = ElgRec_read_ui4(rec);
	elg_ui4 fid = ElgRec_read_ui4(rec);
	elg_ui4 begln = ElgRec_read_ui4(rec);
	elg_ui4 endln = ElgRec_read_ui4(rec);
	elg_ui4 rdid = ElgRec_read_ui4(rec);
	elg_ui1 rtype = ElgRec_read_ui1(rec); 

	elg_readcb_REGION(rid, rnid, fid, begln, endln, rdid, rtype, userdata);
	break;
      }
    case ELG_CALL_SITE :                
      {
	elg_ui4 csid = ElgRec_read_ui4(rec); 
	elg_ui4 fid = ElgRec_read_ui4(rec); 
	elg_ui4 lno = ElgRec_read_ui4(rec); 
	elg_ui4 erid = ElgRec_read_ui4(rec); 
	elg_ui4 lrid = ElgRec_read_ui4(rec);  

	elg_readcb_CALL_SITE(csid, fid, lno, erid, lrid, userdata);
	break;
      }
    case ELG_METRIC :                
      {
	elg_ui4 metid = ElgRec_read_ui4(rec);
	elg_ui4 metnid = ElgRec_read_ui4(rec);
	elg_ui4 metdid = ElgRec_read_ui4(rec);
	elg_ui1 metdtype = ElgRec_read_ui1(rec);
	elg_ui1 metmode = ElgRec_read_ui1(rec);
	elg_ui1 metiv = ElgRec_read_ui1(rec);

	elg_readcb_METRIC(metid, metnid, metdid, metdtype, metmode,
                          metiv, userdata);
	handle->metc++;
	handle->metv = realloc(handle->metv, handle->metc*sizeof(elg_ui8));
	if (handle->metv == NULL)
	  elg_error_msg("Can't allocate metric data buffer");
	break;	     
      }
    case ELG_MPI_COMM :            
      {
	elg_ui4 cid = ElgRec_read_ui4(rec);
	elg_ui4 grpc = ElgRec_read_ui4(rec);
	elg_ui1* grpv = calloc(grpc, sizeof(elg_ui1));
	if (grpv == NULL)
	  elg_error_msg("Can't allocate communicator data buffer");
	   	    
	for (i = 0; i < grpc; i++)
	  grpv[i] = ElgRec_read_ui1(rec);

	elg_readcb_MPI_COMM(cid, grpc, grpv, userdata);
	free(grpv);
	break;
      }
    case ELG_CART_TOPOLOGY :
      {
	elg_ui4  topid = ElgRec_read_ui4(rec);
	elg_ui1  ndims = ElgRec_read_ui1(rec);
	elg_ui4* dimv;
	elg_ui1* periodv;

	dimv = calloc(ndims, sizeof(elg_ui4));
	if (dimv == NULL)
	  elg_error_msg("Can't allocate dimension vector data buffer");
	for (i = 0; i < ndims ; i++)
	  dimv[i] = ElgRec_read_ui4(rec);

	periodv = calloc(ndims, sizeof(elg_ui1));
	if (periodv == NULL)
	  elg_error_msg("Can't allocate period vector data buffer");	      
	for (i = 0; i < ndims; i++)
	  periodv[i] = ElgRec_read_ui1(rec);

	elg_readcb_CART_TOPOLOGY(topid, ndims, dimv, periodv, userdata);
	free(dimv);
	free(periodv);
	break;
      }  
    case ELG_CART_COORDS :
      {
	elg_ui4 topid = ElgRec_read_ui4(rec);
	elg_ui4 lid   = ElgRec_read_ui4(rec);
	elg_ui4 ndims = ElgRec_read_ui1(rec);
	elg_ui4* coordv = calloc(ndims, sizeof(elg_ui4));
	if (coordv == NULL)
	  elg_error_msg("Can't allocate coordinate vector data buffer");
	for (i = 0; i < ndims; i++)
	  coordv[i] = ElgRec_read_ui4(rec); 	    

	elg_readcb_CART_COORDS(topid, lid, ndims, coordv, userdata);
	free(coordv);
	break;
      }
    case ELG_OFFSET :               
      {
	elg_d8  ltime = ElgRec_read_d8(rec);
	elg_d8  offset = ElgRec_read_d8(rec);

	elg_readcb_OFFSET(ltime, offset, userdata);
	break;
      }
    case ELG_NUM_EVENTS :               
      {
	elg_ui4 eventc = ElgRec_read_ui4(rec);

	elg_readcb_NUM_EVENTS(eventc, userdata);
	break;
      }
    case ELG_LAST_DEF :             
      {
	elg_readcb_LAST_DEF(userdata);
	break;
      }

      /* -- Event records -- */

    case ELG_ENTER :               
      {
	int i;
	elg_ui4 lid;
	elg_d8 time;
	elg_ui4 rid;

	if ( length == sizeof(lid) + sizeof(time) + sizeof(rid)
                                   + (handle->metc * sizeof(elg_ui8)) )  
	  lid = ElgRec_read_ui4(rec);
	else
	  lid = handle->single_lid;
	time = ElgRec_read_d8(rec);
	rid = ElgRec_read_ui4(rec);
	for (i = 0; i < handle->metc; i++)
	  handle->metv[i] = ElgRec_read_ui8(rec);
	    
	elg_readcb_ENTER(lid, time, rid, handle->metc, handle->metv, userdata);
	break;
      }
    case ELG_ENTER_CS :               
      {
	int i;
	elg_ui4 lid;
	elg_d8 time;
	elg_ui4 csid;

	if ( length == sizeof(lid) + sizeof(time) + sizeof(csid)
                                   + (handle->metc * sizeof(elg_ui8)) )  
	  lid = ElgRec_read_ui4(rec);
	else
	  lid = handle->single_lid;
	time = ElgRec_read_d8(rec);
	csid = ElgRec_read_ui4(rec);
	for (i = 0; i < handle->metc; i++)
	  handle->metv[i] = ElgRec_read_ui8(rec);
	    
	elg_readcb_ENTER_CS(lid, time, csid, handle->metc, handle->metv,
                            userdata);
	break;
      }
    case ELG_EXIT :                
      {
	int i;
	elg_ui4 lid;
	elg_d8 time;

	if ( length == sizeof(lid) + sizeof(time)
                                   + (handle->metc * sizeof(elg_ui8)) )  
	  lid = ElgRec_read_ui4(rec);
	else
	  lid = handle->single_lid;
	time = ElgRec_read_d8(rec);
	for (i = 0; i < handle->metc; i++)
	  handle->metv[i] = ElgRec_read_ui8(rec);
	    
	elg_readcb_EXIT(lid, time, handle->metc, handle->metv, userdata);
	break;
      }
    case ELG_MPI_SEND :            
      {
	elg_ui4 lid; 
	elg_d8  time;
	elg_ui4 dlid;
	elg_ui4 cid;
	elg_ui4 tag;      
	elg_ui4 sent;

	if ( length == sizeof(lid) + sizeof(time) + sizeof(dlid)
                                   + sizeof(cid) + sizeof(tag) + sizeof(sent) )
	  lid = ElgRec_read_ui4(rec);
	else
	  lid = handle->single_lid;
	time = ElgRec_read_d8(rec);

	dlid = ElgRec_read_ui4(rec);
	cid = ElgRec_read_ui4(rec);
	tag = ElgRec_read_ui4(rec);      
	sent = ElgRec_read_ui4(rec);
	    
	elg_readcb_MPI_SEND(lid, time, dlid, cid, tag, sent, userdata);
	break;
      }
    case ELG_MPI_RECV :            
      {
	elg_ui4 lid; 
	elg_d8  time;
	elg_ui4 slid;
	elg_ui4 cid;
	elg_ui4 tag;      

	if ( length == sizeof(lid) + sizeof(time) + sizeof(slid) + sizeof(cid)
                                   + sizeof(tag) )  
	  lid = ElgRec_read_ui4(rec);
	else
	  lid = handle->single_lid;
	time = ElgRec_read_d8(rec);

	slid = ElgRec_read_ui4(rec);
	cid = ElgRec_read_ui4(rec);
	tag = ElgRec_read_ui4(rec);      
	    
	elg_readcb_MPI_RECV(lid, time, slid, cid, tag, userdata);
	break;
      }
    case ELG_MPI_COLLEXIT :            
      {
	int i;
	elg_ui4 lid;
	elg_d8 time;
	elg_ui4 rlid;
	elg_ui4 cid;
	elg_ui4 sent;  
	elg_ui4 recvd;

	if ( length == sizeof(lid) + sizeof(time)
                                   + (handle->metc * sizeof(elg_ui8)) + 
	     sizeof(rlid) + sizeof(cid) + sizeof(sent) + sizeof(recvd) )  
	  lid = ElgRec_read_ui4(rec);
	else
	  lid = handle->single_lid;
	time = ElgRec_read_d8(rec);
	for (i = 0; i < handle->metc; i++)
	  handle->metv[i] = ElgRec_read_ui8(rec);
	    
	rlid = ElgRec_read_ui4(rec);
	cid = ElgRec_read_ui4(rec);
	sent = ElgRec_read_ui4(rec);  
	recvd = ElgRec_read_ui4(rec);

	elg_readcb_MPI_COLLEXIT(lid, time, handle->metc, handle->metv,
                                rlid, cid, sent, recvd, userdata);
	break;
      }
    case ELG_OMP_FORK :
      {
	elg_ui4 lid; 
	elg_d8  time;

	if ( length == sizeof(lid) + sizeof(time) )  
	  lid = ElgRec_read_ui4(rec);
	else
	  lid = handle->single_lid;
	time = ElgRec_read_d8(rec);

	elg_readcb_OMP_FORK(lid, time, userdata);
	break;
      }
    case ELG_OMP_JOIN : 
      {
	elg_ui4 lid; 
	elg_d8  time;

	if ( length == sizeof(lid) + sizeof(time) )  
	  lid = ElgRec_read_ui4(rec);
	else
	  lid = handle->single_lid;
	time = ElgRec_read_d8(rec);

	elg_readcb_OMP_JOIN(lid, time, userdata);
	break;
      }
    case ELG_OMP_ALOCK :          
      {
	elg_ui4 lid, lkid; 
	elg_d8  time;

	if ( length == sizeof(lid) + sizeof(time) + sizeof(lkid))  
	  lid = ElgRec_read_ui4(rec);
	else
	  lid = handle->single_lid;
	time = ElgRec_read_d8(rec);
	lkid = ElgRec_read_ui4(rec);
 
	elg_readcb_OMP_ALOCK(lid, time, lkid, userdata);
	break;
      }
    case ELG_OMP_RLOCK :      
      {
	elg_ui4 lid, lkid; 
	elg_d8  time;

	if ( length == sizeof(lid) + sizeof(time) + sizeof(lkid))  
	  lid = ElgRec_read_ui4(rec);
	else
	  lid = handle->single_lid;
	time = ElgRec_read_d8(rec);
	lkid = ElgRec_read_ui4(rec);

	elg_readcb_OMP_RLOCK(lid, time, lkid, userdata);
	break;
      }
    case ELG_OMP_COLLEXIT :                
      {
	int i;
	elg_ui4 lid;
	elg_d8 time;

	if ( length == sizeof(lid) + sizeof(time)
                                   + (handle->metc * sizeof(elg_ui8)) )  
	  lid = ElgRec_read_ui4(rec);
	else
	  lid = handle->single_lid;
	time = ElgRec_read_d8(rec);
	for (i = 0; i < handle->metc; i++)
	  handle->metv[i] = ElgRec_read_ui8(rec);
	    
	elg_readcb_OMP_COLLEXIT(lid, time, handle->metc, handle->metv,
                                userdata);
	break;
      }
    case ELG_LOG_OFF :             
      {
	int i;
	elg_ui4 lid;
	elg_d8 time;

	if ( length == sizeof(lid) + sizeof(time)
                                   + (handle->metc * sizeof(elg_ui8)) )  
	  lid = ElgRec_read_ui4(rec);
	else
	  lid = handle->single_lid;
	time = ElgRec_read_d8(rec);
	for (i = 0; i < handle->metc; i++)
	  handle->metv[i] = ElgRec_read_ui8(rec);
	    
	elg_readcb_LOG_OFF(lid, time, handle->metc, handle->metv, userdata);
	break;
      }
    case ELG_LOG_ON :              
      {
	int i;
	elg_ui4 lid;
	elg_d8 time;

	if ( length == sizeof(lid) + sizeof(time)
                                   + (handle->metc * sizeof(elg_ui8)) )  
	  lid = ElgRec_read_ui4(rec);
	else
	  lid = handle->single_lid;
	time = ElgRec_read_d8(rec);
	for (i = 0; i < handle->metc; i++)
	  handle->metv[i] = ElgRec_read_ui8(rec);
	    
	elg_readcb_LOG_ON(lid, time, handle->metc, handle->metv, userdata);
	break;
      }
    case ELG_ENTER_DUMP :          
      {
	int i;
	elg_ui4 lid;
	elg_d8 time;

	if ( length == sizeof(lid) + sizeof(time)
                                   + (handle->metc * sizeof(elg_ui8)) )  
	  lid = ElgRec_read_ui4(rec);
	else
	  lid = handle->single_lid;
	time = ElgRec_read_d8(rec);
	for (i = 0; i < handle->metc; i++)
	  handle->metv[i] = ElgRec_read_ui8(rec);
	    
	elg_readcb_ENTER_DUMP(lid, time, handle->metc, handle->metv, userdata);
	break;
      }
    case ELG_EXIT_DUMP :           
      {
	int i;
	elg_ui4 lid;
	elg_d8 time;

	if ( length == sizeof(lid) + sizeof(time)
                                   + (handle->metc * sizeof(elg_ui8)) )  
	  lid = ElgRec_read_ui4(rec);
	else
	  lid = handle->single_lid;
	time = ElgRec_read_d8(rec);
	for (i = 0; i < handle->metc; i++)
	  handle->metv[i] = ElgRec_read_ui8(rec);
	    
	elg_readcb_EXIT_DUMP(lid, time, handle->metc, handle->metv, userdata);
	break;
      }	 
    default:
      elg_warning("unknown record type");	  
    }
  }
  ElgRec_free(rec);
  return 1;
}
