/****************************************************************************
**  KOJAK                                                                  **
**  http://www.fz-juelich.de/zam/kojak/                                    **
*****************************************************************************
**  Copyright (c) 1998-2004                                                **
**  Forschungszentrum Juelich, Zentralinstitut fuer Angewandte Mathematik  **
**                                                                         **
**  Copyright (c) 2003-2004                                                **
**  University of Tennessee, Innovative Computing Laboratory               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#include "elg_rw.h"
#include "elg_error.h"
#include "elg_impl.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/*
 *-----------------------------------------------------------------------------
 * Helper functions
 *-----------------------------------------------------------------------------
 */

static void elg_swap(void* buffer, size_t length)
{
  /* length must be an even number */

  char tmp;
  int i;

  if (length > 1)
    for (i = 0; i < length / 2; i++)
      {
	tmp = ((char*) buffer)[i];
	((char*) buffer)[i] = ((char*) buffer)[length - 1 - i];
	((char*) buffer)[length - 1 - i] = tmp;
      }  
}

/*
 *-----------------------------------------------------------------------------
 * Macro functions
 *-----------------------------------------------------------------------------
 */

#define ELGREC_GET_NUM(rec, var)                                                     \
  memcpy(&var, rec->pos, sizeof(var));                                               \
  rec->pos += sizeof(var);                                                           \
  if ((rec->byte_order == ELG_LITTLE_ENDIAN && ELG_BYTE_ORDER == ELG_BIG_ENDIAN) ||  \
      (rec->byte_order == ELG_BIG_ENDIAN && ELG_BYTE_ORDER == ELG_LITTLE_ENDIAN))    \
      elg_swap(&var, sizeof(var))

#define ELGREC_WRITE_NUM(rec, var, offset)                                           \
  if ((rec->byte_order == ELG_LITTLE_ENDIAN && ELG_BYTE_ORDER == ELG_BIG_ENDIAN) ||  \
      (rec->byte_order == ELG_BIG_ENDIAN && ELG_BYTE_ORDER == ELG_LITTLE_ENDIAN))    \
      elg_swap(&var, sizeof(var));                                                   \
  memcpy(rec->body + offset, &var, sizeof(var))  
     
/*
 *-----------------------------------------------------------------------------
 * ElgRec
 *-----------------------------------------------------------------------------
 */

/* Data type */
struct ElgRec_struct
{
  elg_ui1   type;
  elg_ui1   length;
  elg_ui1   byte_order;
  buffer_t  pos;
  buffer_t  body;
};


ElgRec* ElgRec_create_empty(elg_ui1 length,
			    elg_ui1 type)
{
  ElgRec* rec;
  
  rec = malloc(sizeof(ElgRec));
  if (rec == NULL)
    elg_error();
 
  rec->length = length;
  rec->type   = type;

  if (rec->length)
    {
      rec->body = malloc(rec->length);
      if (rec->body == NULL)
	elg_error();
    }
  else
    {
      rec->body = NULL;
    }
  
  rec->byte_order = ELG_BYTE_ORDER;
  rec->pos        = rec->body;

  return rec;
}

ElgRec* ElgRec_create_NODE(elg_ui4 nid,
			   elg_ui4 mid,
			   elg_ui4 cpuc,
			   elg_ui4 nnid, 
			   elg_d8  cr)
{
  ElgRec* rec;

  rec = ElgRec_create_empty(sizeof(nid) + 
			    sizeof(mid) + 
			    sizeof(cpuc) + 
			    sizeof(nnid) + 
			    sizeof(cr), 
			    ELG_NODE);
  ElgRec_write_ui4(rec, nid, 0);  
  ElgRec_write_ui4(rec, mid, sizeof(nid));  
  ElgRec_write_ui4(rec, cpuc, sizeof(nid) + sizeof(mid));  
  ElgRec_write_ui4(rec, nnid, sizeof(nid) + sizeof(mid) + sizeof(cpuc));  
  ElgRec_write_d8(rec, cr, sizeof(nid) + sizeof(mid) + sizeof(cpuc) + sizeof(nnid));  
  return rec;
}

ElgRec* ElgRec_create_NUM_EVENTS(elg_ui4 eventc)
{
  ElgRec* rec;

  rec = ElgRec_create_empty(sizeof(eventc), ELG_NUM_EVENTS);
  ElgRec_write_ui4(rec, eventc, 0);  
  return rec;
}

ElgRec* ElgRec_create_LAST_DEF()
{
  ElgRec* rec;

  rec = ElgRec_create_empty(0, ELG_LAST_DEF);
  return rec;
}

void ElgRec_free(ElgRec* rec)
{
  free(rec->body);
  free(rec);
}

/* Reading record header */

elg_ui1 ElgRec_get_type(ElgRec* rec) 
{
  return rec->type;
}

int ElgRec_is_event(ElgRec* rec)
{
  elg_ui1 type = ElgRec_get_type(rec);
  
  return (type >= 100);
} 

elg_ui1 ElgRec_get_length(ElgRec* rec) 
{
  return rec->length;
}

/* Reading record data */

elg_ui1 ElgRec_read_ui1(ElgRec* rec) 
{
  elg_ui1 ui1;

  ELGREC_GET_NUM(rec, ui1);  
  return ui1;
}

elg_ui4 ElgRec_read_ui4(ElgRec* rec) 
{
  elg_ui4 ui4;

  ELGREC_GET_NUM(rec, ui4);  
  return ui4;
}

elg_ui8 ElgRec_read_ui8(ElgRec* rec) 
{
  elg_ui8 ui8;

  ELGREC_GET_NUM(rec, ui8);  
  return ui8;
}

elg_d8 ElgRec_read_d8(ElgRec* rec)
{
  elg_d8 d8;

  ELGREC_GET_NUM(rec, d8);  
  return d8;
}

char* ElgRec_read_string(ElgRec* rec)
{
  return (char*) rec->pos;                   
}

int ElgRec_seek(ElgRec* rec, size_t offset)
{
  if (offset >= rec->length)
    return 0;
    
  rec->pos = rec->body + offset;
  return offset;
}

/* Writing record data -*/

ElgRec* ElgRec_write_ui1(ElgRec* rec,
			 elg_ui1 ui1,
			 size_t offset)
{
  if (offset > rec->length)                                                          
      elg_error_msg("Write beyond record boundaries.");                              
  ELGREC_WRITE_NUM(rec, ui1, offset);
  return rec;
} 

ElgRec* ElgRec_write_ui4(ElgRec* rec,
			 elg_ui4 ui4,
			 size_t offset)
{
  if (offset > rec->length)                                                          
      elg_error_msg("Write beyond record boundaries.");                              
  ELGREC_WRITE_NUM(rec, ui4, offset);
  return rec;
} 

ElgRec* ElgRec_write_ui8(ElgRec* rec,
			 elg_ui8 ui8,
			 size_t offset)
{
  if (offset > rec->length)                                                          
      elg_error_msg("Write beyond record boundaries.");                              
  ELGREC_WRITE_NUM(rec, ui8, offset);
  return rec;
} 

ElgRec* ElgRec_write_d8(ElgRec* rec,
			elg_d8 d8,
			size_t offset)
{
  if (offset > rec->length)                                                          
      elg_error_msg("Write beyond record boundaries.");                              
  ELGREC_WRITE_NUM(rec, d8, offset);
  return rec;
}

/*
 *-----------------------------------------------------------------------------
 * ElgIn
 *-----------------------------------------------------------------------------
 */

/* Data type */

struct ElgIn_struct  
{
  FILE*     file;
  char      name[PATH_MAX];
  elg_ui1   minor_vnr; 
  elg_ui1   major_vnr; 
  elg_ui1   byte_order; 
};

/* Open and close trace file */

ElgIn* ElgIn_open(const char* path_name)
{
  ElgIn*   in;
  char     label[7];

  /* allocate ElgIn record */

  in = malloc(sizeof(ElgIn));
  if (in == NULL)
    {
      elg_warning("Cannot open trace file %s", path_name);
      return NULL;
    }

  /* open file */

  if ((in->file = fopen(path_name, "r")) == NULL)
    return NULL;
  else
    elg_cntl_msg("Opened trace file %s for reading", path_name);

  /* store file name */
  strcpy(in->name, path_name);

  /* read header and check file format */ 

  if (fread(label, strlen(ELG_HEADER) + 1, 1, in->file)           != 1 ||
      fread(&in->major_vnr, sizeof(in->major_vnr), 1, in->file)   != 1 ||
      fread(&in->minor_vnr, sizeof(in->minor_vnr), 1, in->file)   != 1 ||
      fread(&in->byte_order, sizeof(in->byte_order), 1, in->file) != 1 ||
      strcmp(label, ELG_HEADER) != 0)
    {
      fclose(in->file);
      free(in);
      return NULL;
    }

  return in;  
}

int ElgIn_close(ElgIn* in)
{
  int result;

  FILE* file = in->file;

  result = fclose(file);

  if (result == 0)
    elg_cntl_msg("Closed trace file %s", in->name);

  free(in);

  return result;
}

/* EPILOG version */

double ElgIn_get_version(ElgIn* in)
{
  return (double) in->major_vnr + (double) in->minor_vnr / 10;
}

/* File position and error management */

int ElgIn_setpos(ElgIn* in, 
		 const fpos_t* pos)
{
  return fsetpos(in->file, pos);
}

int ElgIn_getpos(ElgIn* in, 
		 fpos_t* pos)
{
  return fgetpos(in->file, pos);
}

int ElgIn_rewind(ElgIn* in)
{
  /* sets position to begin of first record */
  return fseek(in->file, 10,  SEEK_SET);
}

int ElgIn_ferror(ElgIn* in)
{
  return ferror(in->file);
}

int ElgIn_feof(ElgIn* in)
{
  return feof(in->file);
}

/* Reading trace records */

ElgRec* ElgIn_read_record(ElgIn* in) 
{
  ElgRec* rec;
  
  /* allocate record */
  rec = malloc(sizeof(ElgRec));
  if (rec == NULL)
    elg_error();
  
  if (fread(&rec->length, sizeof(rec->length), 1, in->file) != 1 ||
      fread(&rec->type, sizeof(rec->type), 1, in->file)     != 1 )
    {
      free(rec);
      return NULL;
    }
  /* allocate record body */

  if (rec->length)
    {
      rec->body = malloc(rec->length);
      if (rec->body == NULL)
	elg_error();

      /* read record body */
      
      if (fread(rec->body, rec->length, 1, in->file) != 1)
	{
	  ElgRec_free(rec);
	  return NULL;
	}
    }
  else
    {
      rec->body = NULL;
    }

  rec->byte_order = in->byte_order;
  rec->pos        = rec->body;

  return rec;
}

int ElgIn_skip_record(ElgIn* in)
{
  int length;

  if ((length = fgetc(in->file)) == EOF)
    return EOF;

  return fseek(in->file, length + 1, SEEK_CUR); 
}

/*
 *-----------------------------------------------------------------------------
 * ElgOut
 *-----------------------------------------------------------------------------
 */

/* Data type */

struct ElgOut_struct
{
  FILE*     file;
  char      name[PATH_MAX];
};

/* Open and close trace file */

ElgOut* ElgOut_open(const char* path_name)
{
  ElgOut* out;
  elg_ui1 major_vnr  = ELG_MAJOR_VNR;
  elg_ui1 minor_vnr  = ELG_MINOR_VNR;
  elg_ui1 byte_order = ELG_BYTE_ORDER;
  
  /* allocate ElgOut file rec */

  out = malloc(sizeof(ElgOut));
  if (out == NULL) 
    return NULL;
  
  /* open file */
  
  out->file = fopen(path_name, "w");
  if (out->file == NULL)
    {
      free(out);
      elg_warning("Cannot open trace file %s", path_name);
      return NULL;
    }
  else
    elg_cntl_msg("Opened trace file %s for writing", path_name);
  
  /* store file name */
  strcpy(out->name, path_name);

  /* write file header */

  if (fwrite(ELG_HEADER, strlen(ELG_HEADER) + 1, 1, out->file) != 1 ||
      fwrite(&major_vnr, sizeof(major_vnr), 1, out->file)      != 1 ||
      fwrite(&minor_vnr, sizeof(minor_vnr), 1, out->file)      != 1 ||
      fwrite(&byte_order, sizeof(byte_order), 1, out->file)    != 1 )
    {
      fclose(out->file);
      free(out);
      return NULL;
    }

  /* return */

  return out;
}

int ElgOut_close(ElgOut* out)
{
  int result;

  FILE* file = out->file;

  result = fclose(file);

  if (result == 0)
    elg_cntl_msg("Closed trace file %s", out->name);

  free(out);

  return result;
}

/* Write recs */

int ElgOut_localize_record(ElgOut* out, 
			   ElgRec* rec,
			   elg_ui4 lid)
{
  elg_ui1 new_length = rec->length + sizeof(lid);

  if (fwrite(&new_length, sizeof(new_length), 1, out->file) != 1 ||    
      fwrite(&rec->type, sizeof(rec->type), 1, out->file)   != 1 ||    
      fwrite(&lid, sizeof(lid), 1, out->file)             != 1 ||   
      fwrite(rec->body, rec->length , 1, out->file) != 1)    
    return EOF;
  
  return sizeof(new_length) + sizeof(rec->type) + sizeof(lid) + new_length;  
}

int ElgOut_write_record(ElgOut* out, 
			ElgRec* rec)
{
  if (fwrite(&rec->length, sizeof(rec->length), 1, out->file) != 1 ||    
      fwrite(&rec->type, sizeof(rec->type), 1, out->file)     != 1 ||    
      fwrite(rec->body, rec->length, 1, out->file) != 1)    
    return EOF;
  
  return sizeof(rec->length) + sizeof(rec->type) + rec->length;
} 
