/****************************************************************************
**  KOJAK                                                                  **
**  http://www.fz-juelich.de/zam/kojak/                                    **
*****************************************************************************
**  Copyright (c) 1998-2004                                                **
**  Forschungszentrum Juelich, Zentralinstitut fuer Angewandte Mathematik  **
**                                                                         **
**  Copyright (c) 2003-2004                                                **
**  University of Tennessee                                                **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#ifndef _ELG_TRC_H
#define _ELG_TRC_H

#ifdef __cplusplus
#   define EXTERN extern "C" 
#else
#   define EXTERN extern 
#endif

#include "elg_defs.h"
#include "elg_error.h"

/*
 *-----------------------------------------------------------------------------
 *
 *  EPILOG Library (Event Processing, Investigating, and Logging)
 *
 *  - Interface for event trace generation
 *  - Generation of definition records is NOT thread safe
 *
 *-----------------------------------------------------------------------------
 */

EXTERN void elg_open               ( );

EXTERN void elg_flush              ( );

EXTERN void elg_close              ( );

/*
  The next two functions have to be called immediately after
  initializing and immediately before finalizing the communication
  middle-ware, e.g. after MPI_Init() and before MPI_Finalize().  
*/

EXTERN void elg_mpi_init           ( );
EXTERN void elg_mpi_finalize       ( );

/*
 *-----------------------------------------------------------------------------
 * Definition records
 *-----------------------------------------------------------------------------
 */

/*
  The records needed for describing the machine are generated
  automatically.

  The control of hardware performance counters to be monitored is not
  supported by this interface. A way of controlling performance
  counters will be provided in a later version.  
*/
 
/* 
   This function can be used to register a source file. Generates an
   ELG_FILE record as well as appropriate ELG_STRING and
   ELG_STRING_CNT records. The file identifier is assigned
   automatically and returned.  
*/

EXTERN elg_ui4  elg_def_file       ( const char* fname );

/* 
   This function can be used to register a source region. Generates an
   ELG_REGION record as well as appropriate ELG_STRING and
   ELG_STRING_CNT records. The file identifier fid has to be defined
   in advance. The region identifier is assigned automatically and
   returned.  */

EXTERN elg_ui4  elg_def_region     ( const char* rname,
				     elg_ui4 fid,
				     elg_ui4 begln,
				     elg_ui4 endln,
				     const char* rdesc,
				     elg_ui1 rtype );

/* 
   This function can be used to register a call site. Generates an
   ELG_CALL_SITE record. The file identifier fid as well as the
   identifiers of the regions to be left and entered have to be
   defined in advance. The call-site identifier is assigned
   automatically and returned.  */

EXTERN elg_ui4  elg_def_csite     ( elg_ui4 fid,
				    elg_ui4 lno,
				    elg_ui4 erid, 
				    elg_ui4 lrid );
/* 
   Generates an ELG_MPI_COMM record.  
*/

EXTERN void elg_mpi_com            ( elg_ui4 cid,
				     elg_ui4 grpc,
				     elg_ui1 grpv[] );


/*
  This function can be used to register a cartesian topology. Generates an
  ELG_CART_TOPOLOGY record. The number of cartesian dimensions as well as 
  the arrays storing information about the number of processes in each 
  dimension and periodicity in each dimension have to be defined in 
  advance. The topology identifier is assigned automatically and 
  returned */

EXTERN void  elg_def_cart           ( elg_ui4 topid,
				      elg_ui1 ndims,
				      elg_ui4 dimv[],
				      elg_ui1 periodv[] );

/* 
   Generates an ELG_CART_COORDS record 
*/

EXTERN void elg_def_coords          ( elg_ui4 topid,
				      elg_ui1 ndims,
				      elg_ui4 coordv[] );


/*
 *-----------------------------------------------------------------------------
 * Event records
 *-----------------------------------------------------------------------------
 */

EXTERN void elg_enter              ( elg_ui4 rid );  

EXTERN void elg_enter_cs           ( elg_ui4 csid );

EXTERN void elg_exit               ( );

/* MPI */

/* Process ids of message sources and destinations are mapped onto
   location ids by the merging process. */

EXTERN void elg_mpi_send           ( elg_ui4 dpid,
				     elg_ui4 cid,
				     elg_ui4 tag,      
				     elg_ui4 sent );

EXTERN void elg_mpi_recv           ( elg_ui4 spid,
				     elg_ui4 cid,
				     elg_ui4 tag );

EXTERN void elg_mpi_collexit       ( elg_ui4 rpid,
				     elg_ui4 cid,
				     elg_ui4 sent,
				     elg_ui4 recvd );

/* OpenMP */

EXTERN void elg_omp_fork           ( );

EXTERN void elg_omp_join           ( );

EXTERN void elg_omp_alock          ( elg_ui4 lkid );

EXTERN void elg_omp_rlock          ( elg_ui4 lkid );

EXTERN void elg_omp_collexit       ( );
	
/* Tracing events */

EXTERN void elg_log_off            ( );

EXTERN void elg_log_on             ( );

EXTERN void elg_enter_dump         ( );  /* for internal use only */

EXTERN void elg_exit_dump          ( );  /* for internal use only */
				
#endif


























