import copy
import string

from earl import *
from Analyzer import *
from Severity import *

class Pattern:

    def __init__(self, trace):
        self._trace = trace
        self._severity = Severity(self._trace.get_nlocs())

    def descr(self):
        return ""

    def parent(self):
        return None

    def isenabled(self):
        return 1
        
    def register(self, analyzer):
        pass

    def end(self, profile, ctree):        
        pass

    def confidence(self):
        return 'max'

    def severity(self):
        return self._severity


################################################################################
# Total                                                        
################################################################################

class Total(Pattern):

    "Total"

    def descr(self):
        return "Total CPU reservation time (includes time allocated for idle threads)"

    def end(self, profile, ctree):        
        for cnode_id in ctree.keys():
            for loc_id in range(self._trace.get_nlocs()):
                # if slave thread outside a parallel region 
                if ( not profile.is_multhrd(cnode_id, loc_id) ) and self._trace.get_loc(loc_id)['thrd_id'] != 0:
                    master_id = self._trace.get_thrd(self._trace.get_loc(loc_id)['proc_id'] , 0)['loc_id']
                    master_val = profile.get_time(cnode_id, master_id)
                    self._severity.set(cnode_id, loc_id, master_val)
                else:
                    self._severity.set(cnode_id, loc_id, profile.get_time(cnode_id, loc_id))
                                        
################################################################################
# Execution                                                        
################################################################################

class Execution(Pattern):

    "Execution"

    def descr(self):
        return "Execution time (does not includes time allocated for idle threads)"

    def parent(self):
        return "Total"

    def end(self, profile, ctree):        
        for cnode_id in ctree.keys():
            for loc_id in range(self._trace.get_nlocs()):
                self._severity.set(cnode_id, loc_id, profile.get_time(cnode_id, loc_id))

################################################################################
# Idle Threads                                                    
################################################################################

class IdleThreads(Pattern):

    "Idle Threads"

    def descr(self):
        return "Unused CPU reservation time during single-threaded execution"

    def parent(self):
        return "Total"

    def isenabled(self):
        return is_omp_appl(self._trace)

    def end(self, profile, ctree):        
        for cnode_id in ctree.keys():
            for loc_id in range(self._trace.get_nlocs()):
                # if slave thread outside a parallel region 
                if ( not profile.is_multhrd(cnode_id, loc_id) ) and self._trace.get_loc(loc_id)['thrd_id'] != 0:
                    master_id = self._trace.get_thrd(self._trace.get_loc(loc_id)['proc_id'] , 0)['loc_id']
                    master_val = profile.get_time(cnode_id, master_id)
                    self._severity.set(cnode_id, loc_id, master_val)

################################################################################
# Low Floating Point                                                         
################################################################################

class FloatingPoint(Pattern):

    "Floating Point"

    def descr(self):
        return "Floating point rate below average"

    def parent(self):
        return "Execution"

    def isenabled(self):
        return is_mets_avail(['FLOATING_POINT'], self._trace)

    def end(self, profile, ctree):
        tot_time = profile.get_time()
        tot_flop = profile.get_metr('FLOATING_POINT')
        flops    = Severity(self._trace.get_nlocs())  
        
        # determine FLOP rate
        avg_flops = tot_flop / tot_time
        for l in range(self._trace.get_nlocs()):
            for c in ctree.keys():
                if profile.get_time(c,l) != 0.0:
                    flops.set(c,l, profile.get_metr('FLOATING_POINT', c,l) / profile.get_time(c,l))
                else:
                    flops.set(c,l, 0.0)
                    
        # leave all matrix elements zero whose rate < avg_flops  
        for l in range(self._trace.get_nlocs()):
            for c in ctree.keys():
                if flops.get(c,l) < avg_flops:
                    self._severity.set(c, l, profile.get_time(c,l))
                else:
                    self._severity.set(c, l, 0.0)

################################################################################
# L1 Data Cache                                                       
################################################################################

class L1DataCache(Pattern):

    "L1 Data Cache"

    def descr(self):
        return "L1 cache miss rate above average"

    def parent(self):
        return "Execution"

    def isenabled(self):        
        return is_mets_avail(['L1_D_MISS'], self._trace)

    def end(self, profile, ctree):
        tot_time = profile.get_time()
        tot_miss = profile.get_metr('L1_D_MISS')
        missr    = Severity(self._trace.get_nlocs())  
        
        # determine miss rate
        avg_missr = tot_miss / tot_time
        for l in range(self._trace.get_nlocs()):
            for c in ctree.keys():
                if profile.get_time(c,l) != 0.0:
                    missr.set(c,l, profile.get_metr('L1_D_MISS', c,l) / profile.get_time(c,l))
                else:
                    missr.set(c,l, 0.0)
                    
        # leave all matrix elements zero whose rate > avg_missr  
        for l in range(self._trace.get_nlocs()):
            for c in ctree.keys():
                if missr.get(c,l) > avg_missr:
                    self._severity.set(c, l, profile.get_time(c,l))
                else:
                    self._severity.set(c, l, 0.0)

################################################################################
#
# MPI
#
################################################################################

################################################################################
# MPI                                                         
################################################################################

class MPI(Pattern):

    "MPI"

    def descr(self):
        return "Time spent in MPI calls"

    def parent(self):
        return "Execution"

    def isenabled(self):
        return is_mpi_appl(self._trace)

    def register(self, analyzer):
        analyzer.subscribe('MPICEXIT', self.mpicexit)

    def mpicexit(self, mpicexit):
            coll = self._trace.mpicoll(mpicexit['pos'])
            # if collective operation complete
            if coll:
                mpi_coll = {}
                mpi_coll['TYPE'] = 'MPI_COLL'
                mpi_coll['EXITS'] = []
                mpi_coll['ENTERS'] = []
                mpi_coll['ROOT_IDX'] = -1
                root_id = mpicexit['root_id']
                for i in range(len(coll)):
                    mpicexit = self._trace.event(coll[i])
                    mpi_coll['EXITS'].append(mpicexit)
                    enter = self._trace.event(mpicexit['enterptr'])
                    mpi_coll['ENTERS'].append(enter)
                    if mpicexit['loc_id'] == root_id:
                        mpi_coll['ROOT_IDX'] = i 
                return mpi_coll

    def end(self, profile, ctree):
        for cnode_id in ctree.keys():
            if is_mpi_api(ctree.callee(cnode_id)):
                for loc_id in range(self._trace.get_nlocs()):
                    self._severity.set(cnode_id, loc_id, profile.get_time(cnode_id, loc_id))

################################################################################
# MPI Communication                                                   
################################################################################

class MPICommunication(Pattern):

    "Communication"
    
    def descr(self):
        return "Time spent in MPI communication calls"

    def parent(self):
        return "MPI"
        
    def isenabled(self):
        return is_mpi_appl(self._trace)

    def end(self, profile, ctree):
        for cnode_id in ctree.keys():
            if is_mpi_p2p(ctree.callee(cnode_id)) or is_mpi_collcom(ctree.callee(cnode_id)):
                for loc_id in range(self._trace.get_nlocs()):
                    self._severity.set(cnode_id, loc_id, profile.get_time(cnode_id, loc_id))


################################################################################
# MPI IO                                                   
################################################################################

class MPIIO(Pattern):

    "IO"
    
    def descr(self):
        return "Time spent in MPI IO calls"

    def parent(self):
        return "MPI"

    def isenabled(self):
        return is_mpi_appl(self._trace)

    def end(self, profile, ctree):
        for cnode_id in ctree.keys():
            if is_mpi_io(ctree.callee(cnode_id)):
                for loc_id in range(self._trace.get_nlocs()):
                    self._severity.set(cnode_id, loc_id, profile.get_time(cnode_id, loc_id))

################################################################################
# MPI Synchronization                                                   
################################################################################

class MPISynchronization(Pattern):

    "Synchronization"
    
    def descr(self):
        return "Time spent in MPI barrier calls"

    def parent(self):
        return "MPI"

    def isenabled(self):
        return is_mpi_appl(self._trace)

    def end(self, profile, ctree):
        for cnode_id in ctree.keys():
            if is_mpi_barrier(ctree.callee(cnode_id)):
                for loc_id in range(self._trace.get_nlocs()):
                    self._severity.set(cnode_id, loc_id, profile.get_time(cnode_id, loc_id))

################################################################################
# Wait at Barrier (MPI)                                                       
################################################################################

class MPIWaitAtBarrier(Pattern):

    "Wait at Barrier"

    def descr(self):
        return "Waiting time in front of MPI barriers"

    def parent(self):
        return "MPISynchronization"

    def isenabled(self):
        return is_mpi_appl(self._trace)

    def register(self, analyzer):
        analyzer.subscribe('MPI_COLL', self.mpi_coll)

    def mpi_coll(self, mpi_coll):
        enter = mpi_coll['ENTERS'][0]
        if is_mpi_barrier(self._trace.get_reg(enter['reg_id'])):
            start_times = []
            for enter in mpi_coll['ENTERS']:
                time = enter['time']
                start_times.append(time)
            max_start_time = max(start_times)
            for enter in mpi_coll['ENTERS']:
                loc_id   = enter['loc_id']
                cnode_id = enter['cnodeptr']
                wait_time = max_start_time - enter['time']          
                self._severity.add(cnode_id, loc_id, wait_time)

################################################################################
# Barrier Completion (MPI)                                                       
################################################################################

class MPIBarrierCompletion(Pattern):

    "Barrier Completion"

    def descr(self):
        return "Time needed to finish an MPI barrier"

    def parent(self):
        return "MPISynchronization"

    def isenabled(self):
        return is_mpi_appl(self._trace)

    def register(self, analyzer):
        analyzer.subscribe('MPI_COLL', self.mpi_coll)

    def mpi_coll(self, mpi_coll):
        enter = mpi_coll['ENTERS'][0]
        if is_mpi_barrier(self._trace.get_reg(enter['reg_id'])):
            end_times = []
            for mpicexit in mpi_coll['EXITS']:
                time = mpicexit['time']
                end_times.append(time)
            min_end_time = min(end_times)
            for i in range(len(end_times)):
                enter = mpi_coll['ENTERS'][i]
                loc_id    = enter['loc_id']
                cnode_id  = enter['cnodeptr']
                comp_time = end_times[i] - min_end_time         
                self._severity.add(cnode_id, loc_id, comp_time)

################################################################################
# P2P                                                       
################################################################################
                    
class P2P(Pattern):

    "P2P"

    def descr(self):
        return "MPI point-to-point communication"

    def parent(self):
        return "MPICommunication"

    def isenabled(self):
        return is_mpi_appl(self._trace)

    def register(self, analyzer):
        analyzer.subscribe('RECV', self.recv)

    def recv(self, recv):
        recv_op = {}
        recv_op['TYPE'] = 'RECV_OP'
        recv_op['RECV'] = recv
        recv_op['ENTER_RECV'] = self._trace.event(recv['enterptr'])
        recv_op['REGION'] = self._trace.get_reg(recv_op['ENTER_RECV']['reg_id'])
        return recv_op

    def end(self, profile, ctree):
        for cnode_id in ctree.keys():
            for loc_id in range(self._trace.get_nlocs()):
                # if callee P2P function
                if is_mpi_p2p(ctree.callee(cnode_id)):
                    self._severity.set(cnode_id, loc_id, profile.get_time(cnode_id, loc_id))
                else:
                    self._severity.set(cnode_id, loc_id, 0)
                    
################################################################################
# Late Sender                                                                 
################################################################################

class LateSender(Pattern):

    "Late Sender"

    def descr(self):
        return "Time a receiving process is waiting for a message"

    def parent(self):
        return "P2P"

    def isenabled(self):
        return is_mpi_appl(self._trace)

    def register(self, analyzer):
        analyzer.subscribe('RECV_OP', self.recv_op)

    def recv_op(self, recv_op):
        if not ( is_mpi_testx(recv_op['REGION']) or
                 is_mpi_waitall(recv_op['REGION']) or
                 is_mpi_waitsome(recv_op['REGION']) ):
            ls = copy.copy(recv_op)
            ls['SEND'] = self._trace.event(ls['RECV']['sendptr'])  
            ls['ENTER_SEND'] = self._trace.event(ls['SEND']['enterptr'])  
            ls['IDLE_TIME'] =  ls['ENTER_SEND']['time'] - ls['ENTER_RECV']['time'] 
            if ls['IDLE_TIME'] > 0 :
                loc_id  = ls['ENTER_RECV']['loc_id']
                cnode_id  = ls['ENTER_RECV']['cnodeptr']
                self._severity.add(cnode_id, loc_id, ls['IDLE_TIME'])
                ls['TYPE'] = 'LATE_SEND'
                return ls
            else:
                return None

################################################################################
# Messages in Wrong Order (Late Sender)
################################################################################

class MsgsWrongOrderLS(Pattern):

    "Messages in Wrong Order"

    def descr(self):
        return "Late Sender situation due to messages received in the wrong oder"

    def parent(self):
        return "LateSender"

    def isenabled(self):
        return is_mpi_appl(self._trace)

    def register(self, analyzer):
        analyzer.subscribe('LATE_SEND', self.late_send)

    def late_send(self, ls):
        pos    = ls['RECV']['pos']
        loc_id = ls['RECV']['loc_id']
        queue  = self._trace.queue(pos, -1, loc_id)
        if queue and queue[0] < ls['SEND']['pos']:
            loc_id   = ls['ENTER_RECV']['loc_id']
            cnode_id = ls['ENTER_RECV']['cnodeptr']
            self._severity.add(cnode_id, loc_id, ls['IDLE_TIME'])
        return None

    def confidence(self):
        return 'medium'

################################################################################
# Late Receiver                                                               
################################################################################

class LateReceiver(Pattern):

    "Late Receiver"

    def descr(self):
        return "Time a sending process is waiting for the receiver to become ready"

    def parent(self):
        return "P2P"

    def isenabled(self):
        return is_mpi_appl(self._trace)

    def register(self, analyzer):
        analyzer.subscribe('RECV_OP', self.recv_op)

    def recv_op(self, recv_op):
    	if is_mpi_recv(recv_op['REGION']):
            lr = copy.copy(recv_op)
            lr['SEND'] = self._trace.event(lr['RECV']['sendptr'])  
            lr['ENTER_SEND'] = self._trace.event(lr['SEND']['enterptr'])
            send_op_name = string.lower(self._trace.get_reg(lr['ENTER_SEND']['reg_id'])['name'])
            if (send_op_name in ["mpi_send", "mpi_ssend"]): 
                lr['IDLE_TIME'] =  lr['ENTER_RECV']['time'] - lr['ENTER_SEND']['time']  
                if (lr['IDLE_TIME'] > 0 and
                    lr['ENTER_SEND']['pos'] in self._trace.stack(lr['ENTER_RECV']['pos'], lr['ENTER_SEND']['loc_id'])):
                    loc_id   = lr['ENTER_SEND']['loc_id']
                    cnode_id = lr['ENTER_SEND']['cnodeptr']
                    self._severity.add(cnode_id, loc_id, lr['IDLE_TIME'])
                    lr['TYPE'] = 'LATE_RECV'
                    return lr
                else:
                    return None

    def confidence(self):
        return 'medium'

################################################################################
# Messages in Wrong Order (Late Receiver)
################################################################################

class MsgsWrongOrderLR(Pattern):

    "Messages in Wrong Order"

    def descr(self):
        return "Late Receiver situation due to messages sent in the wrong oder"

    def parent(self):
        return "LateReceiver"

    def isenabled(self):
        return is_mpi_appl(self._trace)

    def register(self, analyzer):
        analyzer.subscribe('LATE_RECV', self.late_recv)

    def late_recv(self, lr):
        pos    = lr['RECV']['pos']
        src_id = lr['RECV']['src_id']
        queue  = self._trace.queue(pos, src_id, -1)
        if queue and queue[0] < lr['SEND']['pos']:
            loc_id   = lr['ENTER_SEND']['loc_id']
            cnode_id = lr['ENTER_SEND']['cnodeptr']
            self._severity.add(cnode_id, loc_id, lr['IDLE_TIME'])
        return None

    def confidence(self):
        return 'medium'

################################################################################
# Collective                                                       
################################################################################

class Collective(Pattern):

    "Collective"

    def descr(self):
        return "MPI collective communication"

    def parent(self):
        return "MPICommunication"

    def isenabled(self):
        return is_mpi_appl(self._trace)

    def end(self, profile, ctree):
        for cnode_id in ctree.keys():
            for loc_id in range(self._trace.get_nlocs()):
                # if callee MPI collective function
                if is_mpi_collcom(ctree.callee(cnode_id)):
                    self._severity.set(cnode_id, loc_id, profile.get_time(cnode_id, loc_id))
                else:
                    self._severity.set(cnode_id, loc_id, 0)

################################################################################
# Wait at NxN                                                   
################################################################################

class WaitAtNxN(Pattern):

    "Wait at N x N"

    def descr(self):
        return "Waiting time due to inherent synchronization in MPI n-to-n operations"

    def parent(self):
        return "Collective"

    def isenabled(self):
        return is_mpi_appl(self._trace)

    def register(self, analyzer):
        analyzer.subscribe('MPI_COLL', self.mpi_coll)

    def mpi_coll(self, mpi_coll):
        enter = mpi_coll['ENTERS'][0]
        if is_mpi_n2n(self._trace.get_reg(enter['reg_id'])):
            start_times = []
            for enter in mpi_coll['ENTERS']:
                time = enter['time']
                start_times.append(time)
            max_start_time = max(start_times)
            for enter in mpi_coll['ENTERS']:
                loc_id   = enter['loc_id']
                cnode_id = enter['cnodeptr']
                wait_time = max_start_time - enter['time']          
                self._severity.add(cnode_id, loc_id, wait_time)
        
################################################################################
# Late Broadcast                                                   
################################################################################

class LateBroadcast(Pattern):

    "Late Broadcast"

    def descr(self):
        return "Waiting time due to a late sender in MPI 1-to-n operations"

    def parent(self):
        return "Collective"

    def isenabled(self):
        return is_mpi_appl(self._trace)

    def register(self, analyzer):
        analyzer.subscribe('MPI_COLL', self.mpi_coll)

    def mpi_coll(self, mpi_coll):
        enter = mpi_coll['ENTERS'][0]
        if is_mpi_12n(self._trace.get_reg(enter['reg_id'])):
            root_start_time = mpi_coll['ENTERS'][mpi_coll['ROOT_IDX']]['time']
            for enter in mpi_coll['ENTERS']:
                wait_time = root_start_time  - enter['time']
                if wait_time > 0:        
                    loc_id   = enter['loc_id']
                    cnode_id = enter['cnodeptr']
                    self._severity.add(cnode_id, loc_id, wait_time)
        
################################################################################
# Early Reduce                                                  
################################################################################

class EarlyReduce(Pattern):

    "Early Reduce"

    def descr(self):
        return "Waiting time due to inherent synchronization in MPI all-to-all operations"

    def parent(self):
        return "Collective"

    def isenabled(self):
        return is_mpi_appl(self._trace)

    def register(self, analyzer):
        analyzer.subscribe('MPI_COLL', self.mpi_coll)

    def mpi_coll(self, mpi_coll):
        enter = mpi_coll['ENTERS'][0]
        if is_mpi_n21(self._trace.get_reg(enter['reg_id'])):
            root_enter = mpi_coll['ENTERS'][mpi_coll['ROOT_IDX']]
            sender_start_times = []
            for enter in mpi_coll['ENTERS']:
                if enter['loc_id'] != root_enter['loc_id']:
                    sender_start_times.append(enter['time'])
            wait_time = min(sender_start_times) - root_enter['time']
            if wait_time > 0:
                cnode_id = root_enter['cnodeptr']
                loc_id   = root_enter['loc_id']
                self._severity.add(cnode_id, loc_id, wait_time)
        

################################################################################
#
# OpenMP
#
################################################################################

################################################################################
# OMP                                                         
################################################################################

class OMP(Pattern):

    "OMP"

    def descr(self):
        return "Time spent in the OpenMP run-time system and API"

    def parent(self):
        return "Execution"

    def isenabled(self):
        return is_omp_appl(self._trace)
    
    def register(self, analyzer):
        analyzer.subscribe('TEAM_CREATE', self.team_create)
        
    def team_create(self, tc):
        self._severity.add(tc['ENTER']['cedgeptr'], tc['ENTER']['loc_id'], tc['TIME'])
       
    def end(self, profile, ctree):
        for cnode_id in ctree.keys():
            for loc_id in range(self._trace.get_nlocs()):
                callee = ctree.callee(cnode_id)
                # if callee 
                if (is_omp_api(callee) or
                    is_omp_barrier(callee) or
                    is_omp_flush(callee) or
                    is_omp_critical(callee)):
                    self._severity.set(cnode_id, loc_id, profile.get_time(cnode_id, loc_id))

################################################################################
# OMP Fork                                                        
################################################################################

class OMPFork(Pattern):

    "Fork"

    def __init__(self, trace):
        Pattern.__init__(self, trace)
        self._forkv = vector(self._trace.get_nprocs())

    def descr(self):
        return "Time needed by the master thread to create a team"

    def parent(self):
        return "OMP"

    def isenabled(self):
        return is_omp_appl(self._trace)
    
    def register(self, analyzer):
        analyzer.subscribe('ENTER', self.enter)
        analyzer.subscribe('FORK', self.fork)
        
    def enter(self, enter):
        if is_omp_parallel(self._trace.get_reg(enter['reg_id'])):
            loc_id   = enter['loc_id']
            loc = self._trace.get_loc(loc_id)
            proc_id  = loc['proc_id']
            thrd_id  = loc['thrd_id']
            if thrd_id != 0:
                return
            cnode_id = enter['cedgeptr']
            if cnode_id != -1:
                tc = {}
                tc['TYPE']  = 'TEAM_CREATE'
                tc['ENTER'] = enter
                tc['TIME']  = enter['time'] - self._forkv[proc_id]['time']
                self._severity.add(cnode_id, loc_id, tc['TIME'])
                return tc
            
    def fork(self, fork):
        loc_id  = fork['loc_id']
        proc_id = self._trace.get_loc(loc_id)['proc_id']
        self._forkv[proc_id] = fork


################################################################################
# OMP Flush                                                         
################################################################################

class OMPFlush(Pattern):

    "Flush"

    def descr(self):
        return "Time spent in OpenMP flush directives"

    def parent(self):
        return "OMP"
        
    def isenabled(self):
        return is_omp_appl(self._trace)

    def end(self, profile, ctree):
        for cnode_id in ctree.keys():
            if is_omp_flush(ctree.callee(cnode_id)):
                for loc_id in range(self._trace.get_nlocs()):
                    self._severity.set(cnode_id, loc_id, profile.get_time(cnode_id, loc_id))


################################################################################
# OMP Synchronization                                                         
################################################################################

class OMPSynchronization(Pattern):

    "Synchronization"

    def descr(self):
        return "Time spent on OpenMP synchronization"

    def parent(self):
        return "OMP"

    def isenabled(self):
        return is_omp_appl(self._trace)

    def end(self, profile, ctree):
        for cnode_id in ctree.keys():
            callee = ctree.callee(cnode_id)
            if is_omp_lock(callee) or is_omp_barrier(callee) or is_omp_critical(callee):
                for loc_id in range(self._trace.get_nlocs()):
                    self._severity.set(cnode_id, loc_id, profile.get_time(cnode_id, loc_id))

################################################################################
# OMP Barrier                                                         
################################################################################

class OMPBarrier(Pattern):

    "Barrier"


    def descr(self):
        return "OpenMP barrier synchronization"

    def parent(self):
        return "OMPSynchronization"

    def isenabled(self):
        return is_omp_appl(self._trace)

    def register(self, analyzer):
        analyzer.subscribe('OMPCEXIT', self.ompcexit)

    def ompcexit(self, ompcexit):
            coll = self._trace.ompcoll(ompcexit['pos'])
            # if collective operation complete
            if coll:
                omp_coll = {}
                omp_coll['TYPE'] = 'OMP_COLL'
                omp_coll['EXITS'] = []
                omp_coll['ENTERS'] = []
                for i in range(len(coll)):
                    ompcexit = self._trace.event(coll[i])
                    omp_coll['EXITS'].append(exit)
                    enter = self._trace.event(ompcexit['enterptr'])
                    omp_coll['ENTERS'].append(enter)
                return omp_coll
        
    def end(self, profile, ctree):
        for cnode_id in ctree.keys():
            for loc_id in range(self._trace.get_nlocs()):
                # if callee OMP barrier
                if is_omp_barrier(ctree.callee(cnode_id)):
                    self._severity.set(cnode_id, loc_id, profile.get_time(cnode_id, loc_id))

################################################################################
# OMP Explicit Barrier                                                         
################################################################################

class OMPEBarrier(Pattern):

    "Explicit"

    def descr(self):
        return "Time spent in explicit OpenMP barriers"

    def parent(self):
        return "OMPBarrier"

    def isenabled(self):
        return is_omp_appl(self._trace)

    def end(self, profile, ctree):
        for cnode_id in ctree.keys():
            if is_omp_ebarrier(ctree.callee(cnode_id)):
                for loc_id in range(self._trace.get_nlocs()):
                    self._severity.set(cnode_id, loc_id, profile.get_time(cnode_id, loc_id))

################################################################################
# Wait at Barrier (OMP, explicit)                                                       
################################################################################

class OMPWaitAtEBarrier(Pattern):

    "Wait at Barrier"

    def descr(self):
        return "Waiting time in front of explicit OpenMP barriers"

    def parent(self):
        return "OMPEBarrier"

    def isenabled(self):
        return is_omp_appl(self._trace)

    def register(self, analyzer):
        analyzer.subscribe('OMP_COLL', self.omp_coll)

    def omp_coll(self, omp_coll):
        enter = omp_coll['ENTERS'][0]
        if is_omp_ebarrier(self._trace.get_reg(enter['reg_id'])):
            start_times = []
            for enter in omp_coll['ENTERS']:
                time = enter['time']
                start_times.append(time)
            max_start_time = max(start_times)
            for enter in omp_coll['ENTERS']:
                loc_id   = enter['loc_id']
                cnode_id = enter['cnodeptr']
                wait_time = max_start_time - enter['time']          
                self._severity.add(cnode_id, loc_id, wait_time)

################################################################################
# OMP Implict Barrier                                                         
################################################################################

class OMPIBarrier(Pattern):

    "Implict"

    def descr(self):
        return "Time spent in implicit OpenMP barriers"

    def parent(self):
        return "OMPBarrier"

    def isenabled(self):
        return is_omp_appl(self._trace)

    def end(self, profile, ctree):
        for cnode_id in ctree.keys():
            if is_omp_ibarrier(ctree.callee(cnode_id)):
                for loc_id in range(self._trace.get_nlocs()):
                    self._severity.set(cnode_id, loc_id, profile.get_time(cnode_id, loc_id))

################################################################################
# Wait at Barrier (OMP, implicit)                                                       
################################################################################

class OMPWaitAtIBarrier(Pattern):

    "Wait at Barrier"

    def descr(self):
        return "Waiting time in front of implicit OpenMP barriers"

    def parent(self):
        return "OMPIBarrier"

    def isenabled(self):
        return is_omp_appl(self._trace)

    def register(self, analyzer):
        analyzer.subscribe('OMP_COLL', self.omp_coll)

    def omp_coll(self, omp_coll):
        enter = omp_coll['ENTERS'][0]
        if is_omp_ibarrier(self._trace.get_reg(enter['reg_id'])):
            start_times = []
            for enter in omp_coll['ENTERS']:
                time = enter['time']
                start_times.append(time)
            max_start_time = max(start_times)
            for enter in omp_coll['ENTERS']:
                loc_id   = enter['loc_id']
                cnode_id = enter['cnodeptr']
                wait_time = max_start_time - enter['time']          
                self._severity.add(cnode_id, loc_id, wait_time)


################################################################################
# OMP Lock                                                        
################################################################################

class OMPLock(Pattern):

    "Lock Competition"

    def descr(self):
        return "Time spent waiting for a lock which is in possesion of another thread"

    def parent(self):
        return "OMPSynchronization"

    def isenabled(self):
        return is_omp_appl(self._trace)

    def register(self, analyzer):
        analyzer.subscribe('ALOCK', self.alock)

    def alock(self, alock):

        if alock['lockptr'] != -1:
            lc = {}
            lc['ALOCK']       = alock
            lc['ENTER_ALOCK'] = self._trace.event(alock['enterptr']) 
            lc['RLOCK']       = self._trace.event(alock['lockptr'])
            lc['IDLE_TIME']   = lc['RLOCK']['time'] - lc['ENTER_ALOCK']['time'] 
            if lc['IDLE_TIME'] > 0:
                loc_id  = lc['ENTER_ALOCK']['loc_id']
                cnode   = lc['ENTER_ALOCK']['cnodeptr']
                self._severity.add(cnode, loc_id, idle_time)
                return lc

################################################################################
# OMP API Lock                                                        
################################################################################

class OMPAPILock(Pattern):

    "API"

    def descr(self):
        return "Time spent in OpenMP API calls waiting for a lock"

    def parent(self):
        return "OMPLock"

    def isenabled(self):
        return is_omp_appl(self._trace)

    def register(self, analyzer):
        analyzer.subscribe('LOCK_COMP', self.lock_comp)

    def lock_comp(self, lc):
        if is_function(self._trace.get_reg(lc['ENTER_ALOCK']['reg_id'])):
            loc_id  = lc['ENTER_ALOCK']['loc_id']
            cnode   = lc['ENTER_ALOCK']['cnodeptr']
            self._severity.add(cnode, loc_id, idle_time)
            

################################################################################
# OMP Critical Lock                                                        
################################################################################

class OMPCriticalLock(Pattern):

    "Critical"

    def descr(self):
        return "Time spent in front of a critical section waiting for a lock"

    def parent(self):
        return "OMPLock"

    def isenabled(self):
        return is_omp_appl(self._trace)

    def register(self, analyzer):
        analyzer.subscribe('LOCK_COMP', self.lock_comp)

    def lock_comp(self, lc):
        if is_omp_critical(self._trace.get_reg(lc['ENTER_ALOCK']['reg_id'])):
            loc_id  = lc['ENTER_ALOCK']['loc_id']
            cnode   = lc['ENTER_ALOCK']['cnodeptr']
            self._severity.add(cnode, loc_id, idle_time)

################################################################################
#
# Helper functions
#
################################################################################

def vector(length, default=0):
    nl = []
    for i in range(length):
        nl.append(default)
    return nl

def is_mpi_appl(trace):
    return trace.get_nprocs > 1

def is_mets_avail(metname_list, trace):
    avail = 0
    for i in range(trace.get_nmets()):
        metname = trace.get_met(i)['name']
        if metname in metname_list:
            avail += 1
    if avail != len(metname_list):
        return 0
    else:
        return 1

def is_mpi_api(region):
    prefix = string.lower(region['name'][0:3])
    return (prefix == 'mpi')

def is_mpi_barrier(region):
    prefix = string.lower(region['name'][0:11])
    return (prefix == 'mpi_barrier')

def is_mpi_collcom(region):
    if not(is_mpi_api(region)):
        return 0
    suffix = string.lower(region['name'][3:])
    for substr in ('_all', '_op', 'bcast', 'gather', 'reduce', 'scan', 'scatter'):
        if (string.find(suffix, substr) != -1):
            return 1
    return 0

def is_mpi_n2n(region):
    if not is_mpi_api(region):
        return 0
    suffix = string.lower(region['name'][3:])
    if (string.find(suffix, '_all') != -1) or (suffix == '_reduce_scatter'):
        return 1
    return 0

def is_mpi_12n(region):
    if not is_mpi_api(region):
        return 0
    suffix = string.lower(region['name'][4:])
    if suffix in ('bcast', 'scatter', 'scatterv'):
        return 1
    return 0

def is_mpi_n21(region):
    if not is_mpi_api(region) :
        return 0
    suffix = string.lower(region['name'][4:])
    if suffix in ('gather', 'gatherv', 'reduce'):
        return 1
    return 0

def is_mpi_io(region):
    prefix = string.lower(region['name'][0:8])
    return (prefix == 'mpi_file')

def is_mpi_p2p(region):
    if not is_mpi_api(region):
        return 0
    suffix = string.lower(region['name'][2:])
    for substr in ('buffer', 'cancel', 'get_count', 'probe', 'recv', 'request',  'send', 'start', 'i_test', 'wait'):
        if string.find(suffix, substr) != -1:
            return 1
    return 0

def is_mpi_recv(region):
    return string.lower(region['name']) == 'mpi_recv'

def is_mpi_testx(region):
    if not is_mpi_api(region):
        return 0
    suffix = string.lower(region['name'][4:])
    return (len(suffix) >= 4 and suffix[0:4] == 'test') 

def is_mpi_waitall(region):
    if not is_mpi_api(region):
        return 0
    suffix = string.lower(region['name'][4:])
    return (len(suffix) == 7 and suffix[0:7] == 'waitall') 

def is_mpi_waitsome(region):
    if not is_mpi_api(region):
        return 0
    suffix = string.lower(region['name'][4:])
    return (len(suffix) == 8 and suffix[0:8] == 'waitsome') 

def is_omp_appl(trace):
    return trace.get_nthrds() > trace.get_nprocs() 

def is_omp_api(region):
    return (string.lower(region['name'][0:3]) == 'omp') 

def is_omp_lock(region):
    return (string.lower(region['name'][0:3]) == 'omp' and string.lower(region['name'][-4:]) == 'lock')  

def is_omp_barrier(region):
    return (region['rtype'] == 'OMP_BARRIER' or region['rtype'] == 'OMP_IBARRIER') 

def is_omp_ibarrier(region):
    return (region['rtype'] == 'OMP_IBARRIER') 

def is_omp_ebarrier(region):
    return (region['rtype'] == 'OMP_BARRIER') 

def is_omp_flush(region):
    return (region['rtype'] == 'OMP_FLUSH') 

def is_omp_critical(region):
    return (region['rtype'] == 'OMP_CRITICAL') 

def is_omp_critical_sblock(region):
    return (region['rtype'] == 'OMP_CRITICAL_SBLOCK') 

def is_omp_parallel(region):
    return (region['rtype'] == 'OMP_PARALLEL') 

def is_function(region):
    return (region['rtype'] == 'FUNCTION') 
