/****************************************************************************
**  KOJAK                                                                  **
**  http://www.fz-juelich.de/zam/kojak/                                    **
*****************************************************************************
**  Copyright (C) 1998-2004                                                **
**  Forschungszentrum Juelich, Zentralinstitut fuer Angewandte Mathematik  **
**                                                                         **
**  Copyright (c) 2003-2004                                                **
**  University of Tennessee, Innovative Computing Laboratory               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#ifndef EXP_PATTERN_H
#define EXP_PATTERN_H

/*
 *----------------------------------------------------------------------------
 *
 * class Pattern and descendants
 *
 * classes representing execution patterns to be searched for
 *
 *---------------------------------------------------------------------------- 
 */

#define EXP_PAT_NONE                      -1
#define EXP_PAT_TOTAL                      0
#define EXP_PAT_EXECUTION                  1
#define EXP_PAT_IDLE_THREADS               2
#define EXP_PAT_FLOATING_POINT             3            
#define EXP_PAT_L1_DATA_CACHE              4
#define EXP_PAT_MPI                        5
#define EXP_PAT_MPI_COMMUNICATION          6
#define EXP_PAT_MPI_P2P                    7
#define EXP_PAT_MPI_LATE_SENDER            8     
#define EXP_PAT_MPI_WRONG_ORDER_LS         9     
#define EXP_PAT_MPI_LATE_RECEIVER         10     
#define EXP_PAT_MPI_WRONG_ORDER_LR        11
#define EXP_PAT_MPI_IO                    12
#define EXP_PAT_MPI_SYNCHRONIZATION       13
#define EXP_PAT_MPI_COLLECTIVE            14
#define EXP_PAT_MPI_WAIT_AT_N2N           15
#define EXP_PAT_MPI_LATE_BCAST            16
#define EXP_PAT_MPI_EARLY_REDUCE          17
#define EXP_PAT_MPI_WAIT_AT_BARRIER       18
#define EXP_PAT_MPI_BARRIER_COMPLETION    19
#define EXP_PAT_OMP                       20
#define EXP_PAT_OMP_FORK                  21
#define EXP_PAT_OMP_FLUSH                 22
#define EXP_PAT_OMP_SYNCHRONIZATION       23
#define EXP_PAT_OMP_BARRIER               24
#define EXP_PAT_OMP_EXPLICIT_BARRIER      25
#define EXP_PAT_OMP_WAIT_AT_EBARRIER      26
#define EXP_PAT_OMP_IMPLICIT_BARRIER      27
#define EXP_PAT_OMP_WAIT_AT_IBARRIER      28
#define EXP_PAT_OMP_LOCK                  29
#define EXP_PAT_OMP_LOCK_API              30
#define EXP_PAT_OMP_LOCK_CRITICAL         31

     
#include <earl.h>

#include "Callback.h"
#include "Severity.h"

class Analyzer;
class CallTree;
class Pattern;
class Profile; 

/****************************************************************************
* Pattern creation                                                                                     
****************************************************************************/

void get_patternv(std::vector<Pattern*>& pv, earl::EventTrace* trace);

/****************************************************************************
*
* Pattern interface (abstract class)                                                                                     
*
****************************************************************************/

class Pattern : public Callback
{
 public:

  Pattern(earl::EventTrace* trace) : 
    trace(trace),
    severity(trace->get_nlocs()) 
    {}

  virtual long get_id() =0;
  virtual long get_parent_id() =0;
  virtual std::string get_name() =0;
  virtual std::string get_descr() =0;
  
  virtual bool isenabled()                                { return true; }
  virtual void regcb(Analyzer& analyzer)                  { return; }
  virtual void end(Profile& profile, CallTree& ctree)     { return; }

  double get_sev(long cnode_id, long loc_id)              { return severity.get(cnode_id, loc_id); }   

 protected:
  
  earl::EventTrace* trace;
  Severity<double> severity;
};

/****************************************************************************
*
* Derived pattern classes (concrete)                                                                                     
*
****************************************************************************/

/****************************************************************************
* Total                                                                                     
****************************************************************************/

class Total : public Pattern
{
 public:

  Total(earl::EventTrace* trace) : Pattern(trace) {}

  long get_id()                  { return EXP_PAT_TOTAL; }
  long get_parent_id()           { return EXP_PAT_NONE; }
  std::string get_name()         { return "Total"; }
  std::string get_descr()        { return "Total CPU reservation time (includes time allocated for idle threads)"; }
  
  void end(Profile& profile, CallTree& ctree);
};

/****************************************************************************
* Execution                                                                                    
****************************************************************************/

class Execution : public Pattern
{
 public:

  Execution(earl::EventTrace* trace) : Pattern(trace) {}

  long get_id()                  { return EXP_PAT_EXECUTION; } 
  long get_parent_id()           { return EXP_PAT_TOTAL; }
  std::string get_name()         { return "Execution"; }
  std::string get_descr()        { return "Execution time (does not includes time allocated for idle threads)"; }
  
  void end(Profile& profile, CallTree& ctree);
};

/****************************************************************************
* Idle Threads                                                                                    
****************************************************************************/

class IdleThreads : public Pattern
{
 public:

  IdleThreads(earl::EventTrace* trace) : Pattern(trace) {}

  long get_id()                  { return EXP_PAT_IDLE_THREADS; } 
  long get_parent_id()           { return EXP_PAT_TOTAL; }
  std::string get_name()         { return "Idle Threads"; }
  std::string get_descr()        { return "Unused CPU reservation time during single-threaded execution"; }
  
  bool isenabled();                                
  void end(Profile& profile, CallTree& ctree);
};

/****************************************************************************
*
* Hardware Counter                                                                                    
*
****************************************************************************/

/****************************************************************************
* Low Floating Point                                                                                   
****************************************************************************/

class FloatingPoint : public Pattern
{
 public:

  FloatingPoint(earl::EventTrace* trace) : Pattern(trace) {}

  long get_id()                  { return EXP_PAT_FLOATING_POINT; } 
  long get_parent_id()           { return EXP_PAT_EXECUTION; }
  std::string get_name()         { return "Floating Point"; }
  std::string get_descr()        { return "Floating point rate below average"; }
  
  bool isenabled();                                
  void end(Profile& profile, CallTree& ctree);
};

/****************************************************************************
* L1 Data Cache                                                                                   
****************************************************************************/

class L1DataCache : public Pattern
{
 public:

  L1DataCache(earl::EventTrace* trace) : Pattern(trace) {}

  long get_id()                  { return EXP_PAT_L1_DATA_CACHE; } 
  long get_parent_id()           { return EXP_PAT_EXECUTION; }
  std::string get_name()         { return "L1 Data Cache"; }
  std::string get_descr()        { return "L1 cache miss rate above average"; }
  
  bool isenabled();                                
  void end(Profile& profile, CallTree& ctree);
};

/****************************************************************************
*
* MPI                                                                                    
*
****************************************************************************/

/****************************************************************************
* MPI                                                                                    
****************************************************************************/

class MPI : public Pattern
{
 public:

  MPI(earl::EventTrace* trace) : Pattern(trace) {}

  long get_id()                  { return EXP_PAT_MPI; } 
  long get_parent_id()           { return EXP_PAT_EXECUTION; }
  std::string get_name()         { return "MPI"; }
  std::string get_descr()        { return "Time spent in MPI calls"; }

  bool isenabled();                                
  void regcb(Analyzer& analyzer); 
  void end(Profile& profile, CallTree& ctree);

  cmp_event* mpicexit(earl::Event& event);
};

/****************************************************************************
* MPI Communication                                                                                    
****************************************************************************/

class MPICommunication : public Pattern
{
 public:

  MPICommunication(earl::EventTrace* trace) : Pattern(trace) {}

  long get_id()                  { return EXP_PAT_MPI_COMMUNICATION; } 
  long get_parent_id()           { return EXP_PAT_MPI; }
  std::string get_name()         { return "Communication"; }
  std::string get_descr()        { return "Time spent in MPI communication calls"; }

  bool isenabled();                                
  void end(Profile& profile, CallTree& ctree);
};

/****************************************************************************
* MPI IO                                                                                    
****************************************************************************/

class MPIIO : public Pattern
{
 public:

  MPIIO(earl::EventTrace* trace) : Pattern(trace) {}

  long get_id()                  { return EXP_PAT_MPI_IO; } 
  long get_parent_id()           { return EXP_PAT_MPI; }
  std::string get_name()         { return "IO"; }
  std::string get_descr()        { return "Time spent in MPI IO calls"; }

  bool isenabled();                                
  void end(Profile& profile, CallTree& ctree);
};

/****************************************************************************
* MPI Synchronization                                                                                    
****************************************************************************/

class MPISynchronization : public Pattern
{
 public:

  MPISynchronization(earl::EventTrace* trace) : Pattern(trace) {}

  long get_id()                  { return EXP_PAT_MPI_SYNCHRONIZATION; } 
  long get_parent_id()           { return EXP_PAT_MPI; }
  std::string get_name()         { return "Synchronization"; }
  std::string get_descr()        { return "Time spent in MPI barrier calls"; }

  bool isenabled();                                
  void end(Profile& profile, CallTree& ctree);
};

/****************************************************************************
* Wait at Barrier (MPI)                                                                             
****************************************************************************/

class MPIWaitAtBarrier : public Pattern
{
 public:

  MPIWaitAtBarrier(earl::EventTrace* trace) : Pattern(trace) {}

  long get_id()                  { return EXP_PAT_MPI_WAIT_AT_BARRIER; } 
  long get_parent_id()           { return EXP_PAT_MPI_SYNCHRONIZATION; }
  std::string get_name()         { return "Wait at Barrier"; }
  std::string get_descr()        { return "Waiting time in front of MPI barriers"; }

  bool isenabled();                                
  void regcb(Analyzer& analyzer); 

  cmp_event* cmpnd(const cmp_event* ce);   // parameter type: mpi_coll
};

/****************************************************************************
* Barrier Completion (MPI)                                                                             
****************************************************************************/

class MPIBarrierCompletion : public Pattern
{
 public:

  MPIBarrierCompletion(earl::EventTrace* trace) : Pattern(trace) {}

  long get_id()                  { return EXP_PAT_MPI_BARRIER_COMPLETION; } 
  long get_parent_id()           { return EXP_PAT_MPI_SYNCHRONIZATION; }
  std::string get_name()         { return "Barrier Completion"; }
  std::string get_descr()        { return "Time needed to finish an MPI barrier"; }

  bool isenabled();                                
  void regcb(Analyzer& analyzer); 

  cmp_event* cmpnd(const cmp_event* ce);   // parameter type: mpi_coll
};

/****************************************************************************
* MPI Point-to-Point                                                                                   
****************************************************************************/

class PointToPoint : public Pattern
{
 public:

  PointToPoint(earl::EventTrace* trace) : Pattern(trace) {}

  long get_id()                  { return EXP_PAT_MPI_P2P; } 
  long get_parent_id()           { return EXP_PAT_MPI_COMMUNICATION; }
  std::string get_name()         { return "P2P"; }
  std::string get_descr()        { return "MPI point-to-point communication"; }

  bool isenabled();                                
  void regcb(Analyzer& analyzer); 
  void end(Profile& profile, CallTree& ctree);

  cmp_event* recv(earl::Event& event);
};

/****************************************************************************
* MPI Late Sender                                                                                   
****************************************************************************/

class LateSender : public Pattern
{
 public:

  LateSender(earl::EventTrace* trace) : Pattern(trace) {}

  long get_id()                  { return EXP_PAT_MPI_LATE_SENDER; } 
  long get_parent_id()           { return EXP_PAT_MPI_P2P; }
  std::string get_name()         { return "Late Sender"; }
  std::string get_descr()        { return "Time a receiving process is waiting for a message"; }

  bool isenabled();                                
  void regcb(Analyzer& analyzer); 

  cmp_event* cmpnd(const cmp_event* ce);   // parameter type: recv_op
};

/****************************************************************************
* Messages in Wrong Order (Late Sender)                                                                                 
****************************************************************************/

class MsgsWrongOrderLS : public Pattern
{
 public:

  MsgsWrongOrderLS(earl::EventTrace* trace) : Pattern(trace) {}

  long get_id()                  { return EXP_PAT_MPI_WRONG_ORDER_LS; } 
  long get_parent_id()           { return EXP_PAT_MPI_LATE_SENDER; }
  std::string get_name()         { return "Messages in Wrong Order"; }
  std::string get_descr()        { return "Late Sender situation due to messages received in the wrong oder"; }

  bool isenabled();                                
  void regcb(Analyzer& analyzer); 

  cmp_event* cmpnd(const cmp_event* ce);   // parameter type: late_send
};

/****************************************************************************
* MPI Late Receiver                                                                                   
****************************************************************************/

class LateReceiver : public Pattern
{
 public:

  LateReceiver(earl::EventTrace* trace) : Pattern(trace) {}

  long get_id()                  { return EXP_PAT_MPI_LATE_RECEIVER; } 
  long get_parent_id()           { return EXP_PAT_MPI_P2P; }
  std::string get_name()         { return "Late Receiver"; }
  std::string get_descr()        { return "Time a sending process is waiting for the receiver to become ready"; }

  bool isenabled();                                
  void regcb(Analyzer& analyzer); 

  cmp_event* cmpnd(const cmp_event* ce);   // parameter type: recv_op
};

/****************************************************************************
* Messages in Wrong Order (Late Receiver)                                                                                 
****************************************************************************/

class MsgsWrongOrderLR : public Pattern
{
 public:

  MsgsWrongOrderLR(earl::EventTrace* trace) : Pattern(trace) {}

  long get_id()                  { return EXP_PAT_MPI_WRONG_ORDER_LR; } 
  long get_parent_id()           { return EXP_PAT_MPI_LATE_RECEIVER; }
  std::string get_name()         { return "Messages in Wrong Order"; }
  std::string get_descr()        { return "Late Receiver situation due to messages sent in the wrong oder"; }

  bool isenabled();                                
  void regcb(Analyzer& analyzer); 

  cmp_event* cmpnd(const cmp_event* ce);   // parameter type: late_recv
};

/****************************************************************************
* Collective                                                                                
****************************************************************************/

class Collective : public Pattern
{
 public:

  Collective(earl::EventTrace* trace) : Pattern(trace) {}

  long get_id()                  { return EXP_PAT_MPI_COLLECTIVE; } 
  long get_parent_id()           { return EXP_PAT_MPI_COMMUNICATION; }
  std::string get_name()         { return "Collective"; }
  std::string get_descr()        { return "MPI collective communication"; }

  bool isenabled();                                
  void end(Profile& profile, CallTree& ctree);
};

/****************************************************************************
* Early Reduce                                                                               
****************************************************************************/

class EarlyReduce : public Pattern
{
 public:

  EarlyReduce(earl::EventTrace* trace) : Pattern(trace) {}

  long get_id()                  { return EXP_PAT_MPI_EARLY_REDUCE; } 
  long get_parent_id()           { return EXP_PAT_MPI_COLLECTIVE; }
  std::string get_name()         { return "Early Reduce"; }
  std::string get_descr()        { return "Waiting time due to inherent synchronization in MPI all-to-all operations"; }

  bool isenabled();                                
  void regcb(Analyzer& analyzer); 

  cmp_event* cmpnd(const cmp_event* ce);   // parameter type: mpi_coll
};

/****************************************************************************
* Late Broadcast                                                                              
****************************************************************************/

class LateBroadcast : public Pattern
{
 public:

  LateBroadcast(earl::EventTrace* trace) : Pattern(trace) {}

  long get_id()                  { return EXP_PAT_MPI_LATE_BCAST; } 
  long get_parent_id()           { return EXP_PAT_MPI_COLLECTIVE; }
  std::string get_name()         { return "Late Broadcast"; }
  std::string get_descr()        { return "Waiting time due to a late sender in MPI 1-to-n operations"; }

  bool isenabled();                                
  void regcb(Analyzer& analyzer); 

  cmp_event* cmpnd(const cmp_event* ce);   // parameter type: mpi_coll
};

/****************************************************************************
* Wait at N to N                                                                            
****************************************************************************/

class WaitAtN2N : public Pattern
{
 public:

  WaitAtN2N(earl::EventTrace* trace) : Pattern(trace) {}

  long get_id()                  { return EXP_PAT_MPI_WAIT_AT_N2N; } 
  long get_parent_id()           { return EXP_PAT_MPI_COLLECTIVE; }
  std::string get_name()         { return "Wait at N x N"; }
  std::string get_descr()        { return "Waiting time due to inherent synchronization in MPI n-to-n operations"; }

  bool isenabled();                                
  void regcb(Analyzer& analyzer); 

  cmp_event* cmpnd(const cmp_event* ce);   // parameter type: mpi_coll
};

/****************************************************************************
*
* OpenMP                                                                                  
*
****************************************************************************/

/****************************************************************************
* OpenMP                                                                            
****************************************************************************/

class OMP : public Pattern
{
 public:

  OMP(earl::EventTrace* trace) : Pattern(trace) {}

  long get_id()                  { return EXP_PAT_OMP; } 
  long get_parent_id()           { return EXP_PAT_EXECUTION; }
  std::string get_name()         { return "OMP"; }
  std::string get_descr()        { return "Time spent in the OpenMP run-time system and API"; }

  bool isenabled();                                
  void regcb(Analyzer& analyzer); 
  void end(Profile& profile, CallTree& ctree);

  cmp_event* cmpnd(const cmp_event* ce);  // parameter type: team_create
};

/****************************************************************************
* OMP Flush                                                                          
****************************************************************************/

class OMPFlush : public Pattern
{
 public:

  OMPFlush(earl::EventTrace* trace) : Pattern(trace) {}

  long get_id()                  { return EXP_PAT_OMP_FLUSH; } 
  long get_parent_id()           { return EXP_PAT_OMP; }
  std::string get_name()         { return "Flush"; }
  std::string get_descr()        { return "Time spent in OpenMP flush directives"; }

  bool isenabled();                                
  void end(Profile& profile, CallTree& ctree);
};

/****************************************************************************
* OMP Fork                                                                           
****************************************************************************/

class OMPFork : public Pattern
{
 public:

  OMPFork(earl::EventTrace* trace) : Pattern(trace), forkv(trace->get_nprocs()) {}

  long get_id()                  { return EXP_PAT_OMP_FORK; } 
  long get_parent_id()           { return EXP_PAT_OMP; }
  std::string get_name()         { return "Fork"; }
  std::string get_descr()        { return "Time needed by the master thread to create a team"; }

  bool isenabled();                                
  void regcb(Analyzer& analyzer); 

  cmp_event* enter(earl::Event& event);
  cmp_event* fork(earl::Event& event);

 private:

  std::vector<earl::Event> forkv;
};

/****************************************************************************
* OMP Synchronization                                                                          
****************************************************************************/

class OMPSynchronization : public Pattern
{
 public:

  OMPSynchronization(earl::EventTrace* trace) : Pattern(trace) {}

  long get_id()                  { return EXP_PAT_OMP_SYNCHRONIZATION; } 
  long get_parent_id()           { return EXP_PAT_OMP; }
  std::string get_name()         { return "Synchronization"; }
  std::string get_descr()        { return "Time spent on OpenMP synchronization"; }

  bool isenabled();                                
  void end(Profile& profile, CallTree& ctree);
};

/****************************************************************************
* OMP Barrier                                                                         
****************************************************************************/

class OMPBarrier : public Pattern
{
 public:

  OMPBarrier(earl::EventTrace* trace) : Pattern(trace) {}

  long get_id()                  { return EXP_PAT_OMP_BARRIER; } 
  long get_parent_id()           { return EXP_PAT_OMP_SYNCHRONIZATION; }
  std::string get_name()         { return "Barrier"; }
  std::string get_descr()        { return "OpenMP barrier synchronization"; }

  bool isenabled();                                
  void regcb(Analyzer& analyzer); 
  void end(Profile& profile, CallTree& ctree);

  cmp_event* ompcexit(earl::Event& event);
};

/****************************************************************************
* OMP Explicit Barrier                                                                         
****************************************************************************/

class OMPEBarrier : public Pattern
{
 public:

  OMPEBarrier(earl::EventTrace* trace) : Pattern(trace) {}

  long get_id()                  { return EXP_PAT_OMP_EXPLICIT_BARRIER; } 
  long get_parent_id()           { return EXP_PAT_OMP_BARRIER; }
  std::string get_name()         { return "Explicit"; }
  std::string get_descr()        { return "Time spent in explicit OpenMP barriers"; }

  bool isenabled();                                
  void end(Profile& profile, CallTree& ctree);
};

/****************************************************************************
* Wait at Barrier (OMP, explicit)                                                                         
****************************************************************************/

class OMPWaitAtEBarrier : public Pattern
{
 public:

  OMPWaitAtEBarrier(earl::EventTrace* trace) : Pattern(trace) {}

  long get_id()                  { return EXP_PAT_OMP_WAIT_AT_EBARRIER; } 
  long get_parent_id()           { return EXP_PAT_OMP_EXPLICIT_BARRIER; }
  std::string get_name()         { return "Wait at Barrier"; }
  std::string get_descr()        { return "Waiting time in front of explicit OpenMP barriers"; }

  bool isenabled();                                
  void regcb(Analyzer& analyzer); 

  cmp_event* cmpnd(const cmp_event* ce);  // parameter type: omp_coll
};

/****************************************************************************
* OMP Implicit Barrier                                                                         
****************************************************************************/

class OMPIBarrier : public Pattern
{
 public:

  OMPIBarrier(earl::EventTrace* trace) : Pattern(trace) {}

  long get_id()                  { return EXP_PAT_OMP_IMPLICIT_BARRIER; } 
  long get_parent_id()           { return EXP_PAT_OMP_BARRIER; }
  std::string get_name()         { return "Implict"; }
  std::string get_descr()        { return "Time spent in implicit OpenMP barriers"; }

  bool isenabled();                                
  void end(Profile& profile, CallTree& ctree);
};

/****************************************************************************
* Wait at Barrier (OMP, implicit)                                                                         
****************************************************************************/

class OMPWaitAtIBarrier : public Pattern
{
 public:

  OMPWaitAtIBarrier(earl::EventTrace* trace) : Pattern(trace) {}

  long get_id()                  { return EXP_PAT_OMP_WAIT_AT_IBARRIER; } 
  long get_parent_id()           { return EXP_PAT_OMP_IMPLICIT_BARRIER; }
  std::string get_name()         { return "Wait at Barrier"; }
  std::string get_descr()        { return "Waiting time in front of implicit OpenMP barriers"; }

  bool isenabled();                                
  void regcb(Analyzer& analyzer); 

  cmp_event* cmpnd(const cmp_event* ce); // parameter type: omp_coll  
};

/****************************************************************************
* OMP Lock                                                                     
****************************************************************************/

class OMPLock : public Pattern
{
 public:

  OMPLock(earl::EventTrace* trace) : Pattern(trace) {}

  long get_id()                  { return EXP_PAT_OMP_LOCK; } 
  long get_parent_id()           { return EXP_PAT_OMP_SYNCHRONIZATION; }
  std::string get_name()         { return "Lock Competition"; }
  std::string get_descr()        { return "Time spent waiting for a lock which is in possesion of another thread"; }

  bool isenabled();                                
  void regcb(Analyzer& analyzer); 

  cmp_event* alock(earl::Event& event);
};

/****************************************************************************
* OMP Lock (API)                                                                     
****************************************************************************/

class OMPAPILock : public Pattern
{
 public:

  OMPAPILock(earl::EventTrace* trace) : Pattern(trace) {}

  long get_id()                  { return EXP_PAT_OMP_LOCK_API; } 
  long get_parent_id()           { return EXP_PAT_OMP_LOCK; }
  std::string get_name()         { return "API"; }
  std::string get_descr()        { return "Time spent in OpenMP API calls waiting for a lock"; }

  bool isenabled();                                
  void regcb(Analyzer& analyzer); 

  cmp_event* cmpnd(const cmp_event* ce); // paremter type: lock_comp
};

/****************************************************************************
* OMP Lock (Critical Section)                                                                     
****************************************************************************/

class OMPCriticalLock : public Pattern
{
 public:

  OMPCriticalLock(earl::EventTrace* trace) : Pattern(trace) {}

  long get_id()                  { return EXP_PAT_OMP_LOCK_CRITICAL; } 
  long get_parent_id()           { return EXP_PAT_OMP_LOCK; }
  std::string get_name()         { return "Critical"; }
  std::string get_descr()        { return "Time spent in front of a critical section waiting for a lock"; }

  bool isenabled();                                
  void regcb(Analyzer& analyzer); 

  cmp_event* cmpnd(const cmp_event* ce); // paremter type: lock_comp
};

#endif






