/****************************************************************************
**  KOJAK                                                                  **
**  http://www.fz-juelich.de/zam/kojak/                                    **
*****************************************************************************
**  Copyright (C) 1998-2004                                                **
**  Forschungszentrum Juelich, Zentralinstitut fuer Angewandte Mathematik  **
**                                                                         **
**  Copyright (c) 2003-2004                                                **
**  University of Tennessee, Innovative Computing Laboratory               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#include "Report.h"

#include "CallTree.h"
#include "Pattern.h"
#include "PatternTree.h"

using namespace std;

Report::Report(const PatternTree& ptree, 
	       const CallTree&    ctree,
	       const earl::EventTrace& trace)
{
  // mapping of EXPERT IDs onto CUBE IDs

  map<long, long> patm;
  map<string, long> modm;
  map<long, long> regm;
  map<long, long> csitem;
  map<long, long> cnodem;
  map<long, long> machm;
  map<pair<long, long>, long> nodem;
  map<long, long> procm;
  map<pair<long, long>, long> thrdm;
  map<long, long> locm;

  vector<long> keyv;

  // specify mapping of absent information

  patm[-1] = -1;
  modm["UNKNOWN"] = cube.def_module("UNKNOWN");
  regm[-1] = cube.def_region("UNKNOWN", -1, -1, "PSEUDO", modm["UNKNOWN"]);
  cnodem[-1] = -1;

  // define CUBE metrics

  ptree.get_dfkeyv(keyv);
  for ( long i = 0; i < keyv.size(); i++ )
    {
      Pattern* pat = ptree.get_data(keyv[i]);
      patm[pat->get_id()] = cube.def_met(pat->get_name(), "sec", pat->get_descr(), patm[pat->get_parent_id()]);
    }

  // define CUBE modules

  for ( long file_id = 0; file_id < trace.get_nfiles(); file_id++ )
    {
      string file = trace.get_file(file_id);
      modm[file] = cube.def_module(file);
    }

  // define CUBE regions

  for ( long reg_id = 0; reg_id < trace.get_nregs(); reg_id++ )
    {
      const earl::Region* reg = trace.get_reg(reg_id);
      regm[reg_id] = cube.def_region(reg->get_name(), reg->get_begln(), reg->get_endln(), reg->get_descr(), modm[reg->get_file()]);
    }
  
  // define CUBE call sites

  for ( long csite_id = 0; csite_id < trace.get_ncsites(); csite_id++ )
    {
      const earl::Callsite* csite = trace.get_csite(csite_id);
      csitem[csite_id] = cube.def_csite(modm[csite->get_file()], csite->get_line(), regm[csite->get_callee()->get_id()]);
    }

  // define CUBE call paths (i.e., call-tree nodes)

  ctree.get_dfkeyv(keyv);
  for ( long i = 0; i < keyv.size(); i++ )
    {
      const earl::Callsite* csite  = ctree.get_csite(keyv[i]);
      const earl::Region*   callee = ctree.get_callee(keyv[i]);

      long callee_id, cube_csite_id;

      if ( callee == NULL )
	callee_id = -1;
      else
	callee_id = callee->get_id();

      if ( csite == NULL )
	cube_csite_id = cube.def_csite(modm["UNKNOWN"], -1, regm[callee_id]);
      else
	cube_csite_id = csitem[csite->get_id()];

      cnodem[keyv[i]] = cube.def_cnode(cube_csite_id, cnodem[ctree.get_parent(keyv[i])]);
    }

  // define CUBE machines and nodes

  for ( long mach_id = 0; mach_id < trace.get_nmachs(); mach_id++ )
    {
      const earl::Machine* mach = trace.get_mach(mach_id);
      machm[mach_id] = cube.def_mach(mach->get_name(), cube.def_grid("Grid"));
				     
      for ( long node_id = 0; node_id < mach->get_nnodes(); node_id++ )
	{
	  const earl::Node* node = trace.get_node(mach_id, node_id);
	  nodem[make_pair(mach_id, node_id)] = cube.def_node(node->get_name(), machm[mach_id]);
	}				       
    }
  
  // define CUBE processes and threads

  for ( long proc_id = 0; proc_id < trace.get_nprocs(); proc_id++ )
    {
      const earl::Process* proc = trace.get_proc(proc_id);
      procm[proc_id] = cube.def_proc(proc->get_name(), 
				     nodem[make_pair(proc->get_node()->get_mach_id(), proc->get_node()->get_node_id())]);
      
      for ( long thrd_id = 0; thrd_id < proc->get_nthrds(); thrd_id++ )
	{
	  const earl::Thread* thrd = trace.get_thrd(proc_id, thrd_id);
	  thrdm[make_pair(proc_id, thrd_id)] = cube.def_thrd(thrd->get_name(), procm[proc_id]);
	  
	  locm[thrd->get_loc()->get_id()] = thrdm[make_pair(proc_id, thrd_id)];
	}				       
    }

  // define CUBE severity
  
  vector<long> pat_idv, cnode_idv;

  ptree.get_keyv(pat_idv);
  ctree.get_keyv(cnode_idv);
  
  for ( long i = 0; i < pat_idv.size(); i++ )
    {
      long pat_id = pat_idv[i];

      for ( long j = 0; j < cnode_idv.size(); j++ )
	{
	  long cnode_id = cnode_idv[j];

	  for ( long loc_id = 0; loc_id < trace.get_nlocs(); loc_id++ )
	    cube.set_sev(patm[pat_id], cnodem[cnode_id], locm[loc_id], ptree.get_data(pat_id)->get_sev(cnode_id, loc_id));
	}
    }
}

void 
Report::write(std::string path)
{
  cube.write(path);
}


